/* xrmap - vector based global map generating program 
 *
 * Copyright (C) 2001-2002 Jean-Pierre Demailly <demailly@ujf-grenoble.fr>
 * ftp://ftp.ac-grenoble.fr/ge/geosciences
 *
 * loosely derived from "rmap"
 * Copyright (C) 2000 Reza Naima <reza@reza.net>
 * http://www.reza.net/rmap/
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* #define DEBUG */
/* #define TRACE */
/* #define SUPERTRACE */

/* #include <sys/timeb.h> */

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysymdef.h>
#include <X11/keysym.h>
#include <X11/xpm.h>

#ifdef ZLIB
#include <zlib.h>
#endif

#include "xrmap.h"
#include "version.h"
#include "bitmaps.h"
#include "pixmaps.h"
#include "menudefs.h"

/* Extended ascii conversion for search routine ('192 and more) */
/*  */
/* AAAAAAACEEEEIIIIDNOOOOOOUUUUYPSaaaaaaaceeeeiiiitnoooooouuuuyhy */
static char cvs[] =
  "aaaaaaaceeeeiiiidnoooooouuuuypsaaaaaaaceeeeiiiitnoooooouuuuyhy";
static char *continent[] = { "AF", "AS", "EU", "NA", "SA", "OC" };

/* Main static variables */

static short country_table[676];
static char **feature_descr[6];

static int feature_length[6] = { 
  C_ENDCOLOR-C_FG_TEXT, C_FG_TEXT-C_BG_MAP, 3, 2, 11, 12};
static int feature_index[6] = { C_FG_TEXT, C_BG_MAP, 0, 3, 5, 16};

static char hierarchy0[6] = "01234";
static int mark_step0[5] = { 150, 1200, 3000, 6000, 10000};
static int name_step0[5] = { 625, 5000, 12500, 25000, 40000};
static int city_mark0[5]  = { 1, 2, 3, 4, 5 };
static int city_size0[5] = { 3000000, 700000, 300000, 100000, 1};
static int peak_mark0[5]  = { 1, 2, 3, 4, 5 };
static int peak_size0[5] = { 6500, 4000, 1000, 500, 0};
static char city_filter0[4] = "CR*";
static char loc_filter0[16] = "BCIKLORSTV*";
static char cmd_killed_char;

static char language[4];
static char *search_string;
static int *loc_pointer = NULL;
static char *data_text = NULL;
static int search_index = -1;
static int search_by = 1;
static int num_found;
static int num_listed;
static int start_list;
static int datawin_on = 0;
static int exposed = 0;
static int escape_length = 0;
static int ranking = 0;

Display * dpy;
Visual * visual;
Window Root, mainwin, datawin;
Colormap cmap = 0;
GC gc = 0, textgc = 0;
Atom wm_delete_window, wm_protocols;
Pixel black, white;
Pixmap textpix = 0;

XFontStruct * xfont[6];

int scr;
static int num_msg = 0;
int time_count = 0;
int color_depth;
int color_pad;
int bigendian;

int fontasc[6], fonth[6], fontasc[6], char_width;

int numpt, printedpt, numcurves = 0;
int numcountry = 0, numtz = 0, numloc = 0;

int runlevel = 0;
int img_output = 0;
int reverse = 0;
int but_pressed = 0;
int move_notify = 0;
int control_pressed = 0;
int line_clicked;
int overwrite = 1;
int point_shown = 0;
int setup_change = 0;
int caret_pos = 0;
int cmd_killed_pos = 0;
int cmd_killed_length = 0;
int action_width = 0;
int use_memory = 0;
int big_flags = 0;
int start_menu = 0;
int im_compress = 0;

int box_x, box_y, box_width, box_height, box_spacing, box_numitems;
int datawin_x=-1000000, datawin_y;
int pixmap1_width, pixmap1_height;
int pixmap2_width, pixmap2_height, pixmap_height;

unsigned int datawin_width, datawin_height;
int cmd_mode = MODE_RUN, cmd_type = M_RUN_START;
int spacing_generalmenu;
int spacing_filemenu;
int spacing_optionmenu;
int spacing_colormenu;
int spacing_helpmenu;

int line_pos;
int city_shown;
int city_found = -1;
int text_shown;

char * cmd_string;

double earth_radius, coordconv, preview;

double ps_width=150.0, ps_lm=30.0, ps_bm=30.0;
int ps_rot = 0, ps_grayscale = 0;
char * ps_viewer;
char * im_viewer;
char * print_cmd;
char * editor_cmd;
char * list_lang;

Country **countries = NULL;
Timezone **timezones = NULL;
Location **locations = NULL;
Header * header = NULL;
Segment_index * segment_index = NULL;
signed char * segment_buffer = NULL;
signed char * fullmap_buffer = NULL;

/* return a pointer to the color struct that 
   contains the rgb values for a color
   or null if nothing is found that matches
*/
void get_color (char* c, ImageLayout * scene, int i) {
        XColor color, exact;
	Colordata *colors;
        int bool = 0;
        
	if (!c) return;
	colors = &scene->color[i];
	if (*c == '+') {
	   colors->bool = ON;
	   bool = 1;
	   ++c;
	}
	if (*c == '-') {
	   colors->bool = OFF;
	   bool = 1;
	   ++c;
	}
	if (!(*c)) return;

	if (cmap) {
           if (XAllocNamedColor(dpy, cmap, c, &color, &exact)!=(Status)1)
   	      return;
   	   colors->r = exact.red;
   	   colors->g = exact.green;
   	   colors->b = exact.blue;
           colors->pix = color.pixel;
   	   if (reverse && i<C_FG_TEXT) {
   	      if (color_depth <= 8) {
                 color.red = 65535 - exact.red;
                 color.green = 65535 - exact.green;
                 color.blue = 65535 - exact.blue;
                 XAllocColor(dpy, cmap, &color);
   	         colors->pix = color.pixel;
   	      } else
   	         colors->pix = ((1<<color_depth) - 1) - colors->pix;
   	   }
	}

	strncpy(colors->name, c, 39);
        colors->name[39] = '\0';
}

void set_bg_fg(ImageLayout * scene)
{
        if (mainwin)
           XSetWindowBackground(dpy, mainwin, scene->color[C_BG_MAP].pix);
	if (gc) {
           XSetBackground(dpy, gc, scene->color[C_BG_TEXT].pix);
           XSetForeground(dpy, gc, scene->color[C_FG_TEXT].pix);
	}
}

void get_all_colors (ImageLayout * scene)
{
        int i;

	if (color_depth<=8) {
	   if (cmap) XFreeColormap(dpy, cmap);
	   cmap = XCreateColormap(dpy, Root, visual, AllocNone);
	}

	for (i=0; i<C_ENDCOLOR; i++)
	       if (scene->color[i].bool != INACTIVE) {
		  if (i>=C_FG_TEXT) scene->color[i].bool = 1;
	          get_color(scene->color[i].name, scene, i);
	       }

	if (color_depth<=8 && mainwin)
           XSetWindowColormap(dpy, mainwin, cmap);

	set_bg_fg(scene);
}

char * parse_item(char *str, int *num)
{
        char *ptr0, *ptr, *ptrp;
	char c;
	int i;

	if (!str) return NULL;
	if (!*str) return NULL;
	ptr0 = str;
	while (*ptr0 && isspace(*ptr0)) ++ptr0;

        ptr = index(ptr0, ',');
	if (!ptr) ptr = index(ptr0, ' ');
	if (!ptr) return NULL;
        c = *ptr;
	*ptr = '\0';
	if (*ptr0>='0' && *ptr0<='9') 
           *num = atoi(ptr0);
	else {
	   i = 0;
           if (!strncasecmp(ptr0, "int", 3)) i = 0; else
           if (!strncasecmp(ptr0, "nat", 3)) i = 3; else
	   if (!strncasecmp(ptr0, "riv", 3)) i = 5; else
	   if (!strncasecmp(ptr0, "cil", 3)) i = 16; else
	   if (!strncasecmp(ptr0, "mrk", 3)) i = C_BG_MAP; else 
	   if (!strncasecmp(ptr0, "wid", 3)) i = C_FG_TEXT; else {
               *ptr = c;
	       return NULL;
	   }
	   ptrp = ptr0+3;
	   while (*ptrp>='A') ++ptrp;
	   *num = i+atoi(ptrp);
	}
        *ptr = c;
	++ptr;
	while (*ptr && isspace(*ptr)) ++ptr;
        return ptr;
}

void set_opt_color(ImageLayout * scene, char * string, int force)
{
int i;
char *ptr;
        if (!(ptr=parse_item(string, &i))) return;
	if (*ptr && i>=0 && i<C_ENDCOLOR) {
	    if (scene->color[i].bool != INACTIVE || force) {
	        if (i>=C_FG_TEXT) scene->color[i].bool = 1;
	        get_color(ptr, scene, i);
	    }
	}
}

void get_output_type(ImageLayout *scene) 
{
    int l;

    l = strlen(scene->outfile);
    if (l==0) {
        img_output = OUT_NONE;
	return;
    }

    if (!strcmp(scene->outfile+l-3, ".gz")) l-= 3;

    if (!strncmp(scene->outfile+l-4, ".ppm", 4))
        img_output = OUT_PPM;
    else
    if (!strncmp(scene->outfile+l-4, ".xpm", 4))
        img_output = OUT_XPM;
    else
    if (!strncmp(scene->outfile+l-3, ".ps", 3))
        img_output = OUT_PS;
    else
        /* otherwise set default format = XPM */
        img_output = OUT_XPM;
}

int getWindowPlacement (win, x, y, w, h)
Window win;
int *x, *y;
unsigned int *w, *h;
{
   unsigned int b, d;
   Window root, junk; 

   XFlush(dpy);
   /*
   Window parent, *children;

   XQueryTree(dpy, win, &root, &parent, &children, &n);

   if (!parent) {
      fprintf(stderr, "%s\n", msg[CANNOT_QUERY_WINDOW_TREE]);
      return 0;
   }
   */

   XGetGeometry(dpy, win, &root, x, y, w, h, &b, &d);
   XTranslateCoordinates(dpy, win, RootWindow(dpy, scr), 0, 0, x, y, &junk);
   return 1;
}

/* This tells us whether user should enter text or not */
int text_input_mode()
{
   return
       cmd_mode==MODE_CMD || 
      (cmd_mode==MODE_MENU && cmd_type==M_MENU_SEARCH) || 
      (cmd_mode==MODE_FILE && cmd_type>M_FILE_START && cmd_type<M_FILE_SAVE)|| 
      (cmd_mode==MODE_FILE && cmd_type==M_FILE_SAVE && line_clicked==1) || 
      (cmd_mode==MODE_OPTION && cmd_type>M_OPTION_START) ||
      (cmd_mode==MODE_COLOR && cmd_type>M_COLOR_START);
}

/* This tells us whether user should click on a box */
int click_input_mode()
{
   return
      (cmd_mode==MODE_FILE && cmd_type==M_FILE_DATA) || 
      (cmd_mode==MODE_FILE && cmd_type==M_FILE_PRINTCFG) || 
      (cmd_mode==MODE_FILE && cmd_type==M_FILE_SAVE) || 
      (cmd_mode==MODE_MENU && cmd_type==M_MENU_SEARCH) ||
      (cmd_mode==MODE_MENU && cmd_type==M_MENU_LIST) ||
       cmd_mode==MODE_OPTION || cmd_mode==MODE_COLOR;
}

void draw_separator(int x, int position)
{
int y;

	y = (position-TOP)*(fonth[0]+16);
	XDrawLine(dpy, mainwin, gc, x, y, x, y+fonth[0]+5);
}

void draw_mark(ImageLayout *scene, int mode, int num, int position)
{
int i, a, d, x1, x2, y, spacing = 0;
char *str = "";

        if (num<=0) return;
	y = (position-TOP)*(fonth[0]+16)+fonth[0]+6;

        if (mode==MODE_MENU) {
           str = labels_generalmenu;
           spacing = spacing_generalmenu;
	}
        if (mode==MODE_OPTION) {
           str = labels_optionmenu;
           spacing = spacing_optionmenu;
	}
        if (mode==MODE_COLOR) {
           str = labels_colormenu;
           spacing = spacing_colormenu;
	}
        if (mode==MODE_HELP) {
           str = labels_helpmenu;
           spacing = spacing_helpmenu;
	}
        if (mode==MODE_FILE) {
           str = labels_filemenu;
           spacing = spacing_filemenu;
	}

        x1 = XTextWidth(xfont[0], str, spacing*(num-1)-1);
        x2 = XTextWidth(xfont[0], str, spacing*num-1);

        XSetWindowBackground(dpy, mainwin, scene->color[C_BG_MAP].pix);
        XClearArea(dpy, mainwin, 0, y, scene->width, 10, False);

        XSetForeground(dpy, gc, scene->color[C_FG_TICKS].pix);
	a = (x1+x2)/2;
	++y;
	for (i=0; i<=7; i++) {
	   d = 3-i/2;
	   XDrawLine(dpy, mainwin, gc, a-d, y+i, a+d, y+i);
	}
        XSetForeground(dpy, gc, scene->color[C_FG_TEXT].pix);
	/*
        XSetForeground(dpy, gc, scene->color[C_BG_TEXT].pix);
        d = (x2-x1)/2;
	for (i=0; i<=7; i++) {
	   a = (2*d*i)/15;
	   XDrawLine(dpy, mainwin, gc, x1+a, y+i, x2-a, y+i);
	}
        XSetForeground(dpy, gc, scene->color[C_FG_TEXT].pix);
	for (i=0; i<=7; i++) {
	   a = (2*d*i)/15;
	   ap = (2*d*(i+1))/15;
	   if (ap>d) ap=d;
	   XDrawLine(dpy, mainwin, gc, x1+a, y+i, x1+ap, y+i);
	   XDrawLine(dpy, mainwin, gc, x2-ap, y+i, x2-a, y+i);
	}
	*/
}

void drawWinString (ImageLayout *scene, char *s, int position)
{
        int i, y0=0, yt=0, tw;

        if (position<BOTTOM/2) {
	   yt = (position-TOP)*(fonth[0]+16);
	   y0 = yt + fontasc[0] + 2;
	} else {
	   yt = scene->height - (BOTTOM-position)*(fonth[0]+4);
	   y0 = yt - xfont[0]->max_bounds.descent - 4;
	   yt = yt - fonth[0] - 6;
	}
        XSetWindowBackground(dpy, mainwin, scene->color[C_BG_TEXT].pix);
        XClearArea(dpy, mainwin, 0, yt, scene->width, fonth[0]+6, False);
        XDrawImageString(dpy, mainwin, gc, 4, y0, s, strlen(s));
	if (text_input_mode() && position==BOTTOM) {
	   tw = XTextWidth(xfont[0], s, caret_pos);
           XSetWindowBackground(dpy, mainwin, scene->color[C_FG_CARET].pix);
           XClearArea(dpy, mainwin, tw+4, yt+fonth[0]+3, char_width, 2, False);
           XSetWindowBackground(dpy, mainwin, scene->color[C_BG_TEXT].pix);
	}

	if (cmd_mode>MODE_RUN && position==TOP) {
           XSetWindowBackground(dpy, mainwin, scene->color[C_BG_ESCAPE].pix);
           XClearArea(dpy, mainwin, scene->width-3*action_width, 0, 
                                    3*action_width, fonth[0]+6, False);

	   XSetBackground(dpy, gc, scene->color[C_BG_ESCAPE].pix);
	   for (i=ACTIVATE; i<=QUIT; i++) {
	      draw_separator(scene->width-(3-i)*action_width, TOP);
	      XDrawImageString(dpy, mainwin, gc, 
                   scene->width-(3-i)*action_width+8, fontasc[0]+2,
                   msg[i], strlen(msg[i]));
	   }
           XSetBackground(dpy, gc, scene->color[C_BG_TEXT].pix);
	}

        XSetWindowBackground(dpy, mainwin, scene->color[C_BG_MAP].pix);
	XFlush(dpy);
	text_shown = 1;
}

int parse_cmd_line(ImageLayout * scene, char *string);

void parse_rcfile (ImageLayout *s) {
	FILE *f;
	int i, j, pos=0, commands=0, numtype;
	int ret, line_number=0;
	char string[512];
	char type[256];
	char item_color[256];
	char active;
	char *ptr;

	/* initialize stuff */
	bzero(string,256);

	/* read in the color config file and allocate colors */
	f = fopen(s->rcfile, "r");
	if (f == NULL) {
	   sprintf(string, msg[COULDNT_OPEN_RCFILE], s->rcfile);
	   if (runlevel==0) {
	      fprintf(stderr, "%s\n", string);
	      exit(-2);
	   } else {
	      drawWinString(s, string, BOTTOM);
	      XFlush(dpy);
	      usleep(WARNING_TIME);
	      return;
	   }
	}

	/* now parse the file */
	while ((ret = fread(string, 1, 511, f)) != 0) {
		line_number++;
		pos = (int) ftell(f) - ret;
		for (i=0; i<511; i++) {
			if (string[i] == '\n') {
				string[i] = '\0';
				fseek(f,pos+i+1, SEEK_SET);
				break;
			}
		}
		/* remove comments */
		for (j=0; j<i; j++) {
			if (string[j] == '%')  {
				string[j] = '\0';
                                i = j;
				break;
			}
		} 
		
		/* remove trailing blank characters */
		while (i>0 && isspace(string[i-1])) string[--i] = '\0';

		/* don't consider blank lines */
		if (i==0) continue;

		/* remove initial blank characters */
		i = 0;
		while (string[i] && isspace(string[i])) ++i;
		if (i>0) {
		   j = i;
		   while (string[j]) {
		      string[j-i] = string[j];
                      ++j;
		   }
		   string[j-i] = '\0';
		}

		if (!strncasecmp(string, "<commandlines>", 9)) {
		   commands = 1;
		   continue;
		}
                if (commands==1) {
		   parse_cmd_line(s, string);
		   continue;
		}

		ret = sscanf(string, "%s %c%s", type, &active, item_color);
		strcat(type, " /");
		(void)parse_item(type, &numtype);
		ptr = item_color;
		if (ptr) while (*ptr && isspace(*ptr)) ++ptr;

		if (ret != 3 || (active!='+' && active!='-')) {
			fprintf(stderr, msg[UNABLE_PARSE_STRING], 
				string,line_number,s->rcfile);	
                        fprintf(stderr, "\n");
			continue;
		}
		if (numtype < 0 || numtype >= C_ENDCOLOR) {
			fprintf(stderr, msg[VALUE_OUT_OF_RANGE],
			       C_ENDCOLOR-1, line_number, s->rcfile);
                        fprintf(stderr, "\n");
		} else {
			if (active == '+')
				s->color[numtype].bool = ON;
			else
			if (active == '-')
				s->color[numtype].bool = OFF;
                  	if (s->color[numtype].bool != INACTIVE) {
				strncpy(s->color[numtype].name, ptr, 39);
				s->color[numtype].name[39] = '\0';
			}
  		}
	}
	fclose(f);
}

double angle_inc(double z)
{
        double norm[] = { 1.0, 0.75, 0.5, 0.3, 0.2, 0.15 };
        double val, valn, fact;
	int i;

	if (z<=1.0) return 15.0;

	val = 15.0/z;
	valn = 10.0;
	fact = 10.0;
	i=0;
	while (valn>val) {
	   ++i;
	   if (i>5) {
	      fact *= 0.1;
	      i = 0;
	   }
	   valn = norm[i]*fact;
	}

        return valn;
}

void check_spacing(ImageLayout * scene)
{
        if (scene->ropt.spacing_lat< 1.0/scene->gopt.zoom)
	    scene->ropt.spacing_lat = angle_inc(scene->gopt.zoom);

        if (scene->ropt.spacing_lon< 1.0/scene->gopt.zoom)
	    scene->ropt.spacing_lon = angle_inc(scene->gopt.zoom);

}

void show_title(ImageLayout * scene, Window w)
{
        char title[80];
	int i;

	if (cmd_mode<MODE_MENU)
	   i = cmd_mode;
	else
	   i = MODE_MENU;
	mode_descr[0] = projtype[scene->projection];
	sprintf(title, "%s / %s", PACKAGE, mode_descr[i]);
        XStoreName(dpy, w, title);
}

void set_radius(ImageLayout * scene) 
{
        earth_radius = (double)(scene->width/2);

        if (scene->projection==SPHERICAL && scene->width>scene->height)
	   earth_radius = (double)(scene->height/2);
        coordconv = 180.0/earth_radius;
}

void set_font_params(int which)
{
int i, w;
        if (which<0 || which>5) return;
	fontasc[which] = xfont[which]->max_bounds.ascent;
	fonth[which] = fontasc[which] + xfont[which]->max_bounds.descent;
	if (which==0) {
	   char_width = XTextWidth(xfont[0], "w", 1);
	   action_width = 0;
	   for (i=ACTIVATE; i<=QUIT; i++) {
	      w = XTextWidth(xfont[0], msg[i], strlen(msg[i]));
	      if (w>action_width) action_width = w;
	   }
	   action_width += 16;
	}
}

int strpcmp(char *s, char *t, int n)
{
        int mode;

        if (n<0) {
	   mode = 0;
	   n = -n;
	} else
	   mode = 1;
        if (!s || !t) return -1;
	if (runlevel==0 && *s!='-' && *s!='+') return -1;

        if (*t!='+' && *t!='-') { 
           while(*s=='-') ++s; 
	}
	if (strlen(s)<n || strlen(s)>strlen(t)) return -1;
        n = strlen(s);

	if (mode) 
           return strncmp(s, t, n);
	else
           return strncasecmp(s, t, n);
}

int colornumber(int mode, char * str)
{
       int i, run;
       char *endptr;

       if (!str) return -1;
       run = runlevel;
       runlevel = 1;
       if (mode) 
	  for (i = 0; i < NUM_CONTINENTS; i++) {
             if (!strpcmp(str, continent_name[i], -2)) {
	        runlevel = run;
	        return i+1;
	     }
	  }
       else
	  for (i = 0; i < NUM_ITEMS; i++) {
             if (!strpcmp(str, item_name[i], -2)) {
	        runlevel = run;
	        return i;
	     }
	  }
       runlevel = run;
       i = strtol(str, &endptr, 10);
       if (!endptr) return -1;
       return i;
}

int set_flag(ImageLayout * scene, char *c, int mode)
{
    char * number, *ptr1, *ptr2;
    char str[256];
    int i, tot;

    if (!strcasecmp(c, "all")) {
       if (mode) 
          scene->gopt.desired_continents = 0xFFFFFFFF;
       else
          scene->gopt.desired_categories = 0xFFFFFFFF;
       return 0;
    }
    
    number = (char *) malloc(strlen(c)+2);
    ptr1 = c;
    tot = 0;

    while (ptr1 && *ptr1) {
       ptr2 = index(ptr1, ',');
       strcpy(number, ptr1);
       if (ptr2) number[ptr2-ptr1] = '\0';
       i = colornumber(mode, number);
       if (i < 1 || i > 4+2*mode) {
	  sprintf(str, msg[PARAMETER_RANGE], itemtype[mode], 4+2*mode, "\n");
	  if (runlevel)
             drawWinString(scene, str, BOTTOM);
	  free(number);
	  return 1;
       }
       tot |= (0x00000001 << (i-1));
       if (ptr2)
          ptr1 = ptr2+1;
       else
	  ptr1 = NULL;
    }
    if (mode) 
       scene->gopt.desired_continents = tot;
    else
       scene->gopt.desired_categories = tot;
    free(number);
    return 0;
}

void usage()
{
int i;

    printf("%s\n", msg[BEGIN_USAGE]);
    for (i=BEGIN_USAGE+1; i<END_USAGE; i++)
        printf("             %s\n", msg[i]);
    printf("\n");
    for (i=BEGIN_LIST_PARAMETERS; i<END_LIST_PARAMETERS; i++)
        printf("  %s\n", msg[i]);
}

short set_proj(char *s)
{
        int run;
	short projection;

	run = runlevel;
	runlevel = 1;
	if (!strpcmp(s, "spherical", -3))
           projection = SPHERICAL;
	else
	if (!strpcmp(s, "rectangular", -3))
	   projection = RECTANGULAR;
	else
	if (!strpcmp(s, "cylindrical", -3))
	   projection = CYLINDRICAL;
	else
	if (!strpcmp(s, "mercator", -3))
	   projection = MERCATOR;
	else
	if (!strpcmp(s, "miller", -3))
	   projection = MILLER;
	else
	if (!strpcmp(s, "sinusoidal", -3))
	   projection = SINUSOIDAL;
	else
	if (!strpcmp(s, "elliptic", -3))
	   projection = ELLIPTIC;
	else
	if (!strpcmp(s, "mollweide", -3))
	   projection = MOLLWEIDE;
	else {
           projection = atoi(s); 
           if (projection<SPHERICAL || projection>=NUMPROJ) 
              projection = SPHERICAL;
	}
	runlevel = run;
	return projection;
}

void create_textpix(int width)
{
        int i, mfonth;
        if (textpix)
	   XFreePixmap(dpy, textpix);
	mfonth = 0;
	for (i=1; i<=5; i++)
	    if (fonth[i]>mfonth) mfonth = fonth[i];
	textpix = XCreatePixmap(dpy, Root, width, mfonth, 1);
}

int set_font(char *s, int which)
{
	XFontStruct * nfont;
	nfont = XLoadQueryFont(dpy, s);
	if (nfont == (XFontStruct *)NULL) {
	   fprintf(stderr, msg[CANT_OPEN_FONT], s);
	   fprintf(stderr, "\n");
	   return 1;
	}
        if (xfont[which]) XFreeFont(dpy, xfont[which]);
	xfont[which] = nfont;
        if (which==0)
           XSetFont(dpy, gc, nfont->fid);
        if (textpix)
	   XFreePixmap(dpy, textpix);
	textpix = 0;
	set_font_params(which);
	setup_change |= FONT_CHANGE;
	return 0;
}

void putPicture (ImageLayout *s) {
     XPutImage(dpy, mainwin, gc,
                    s->image, 0, 0, 0, 0,
                    s->image->width, s->image->height);
     line_pos = -1;
     city_shown = 0;
     point_shown = 0;
     text_shown = 0;
     XFlush(dpy);
}

void putPicturePiece (ImageLayout *s, int pos)
{
     if (pos==TOP)
        XPutImage(dpy, mainwin, gc,
                    s->image, 0, 0, 0, 0,
                    s->image->width, fonth[0]+6);
     if (pos==BOTTOM)
        XPutImage(dpy, mainwin, gc, s->image,
                    0, s->image->height-fonth[0]-6,
                    0, s->image->height-fonth[0]-6,
                    s->image->width, fonth[0]+6);
}

void freeCountries()
{
int i;
   for (i=0; i<numcountry; i++) {
      free(countries[i]->name);
      free(countries[i]->flags);
      free(countries[i]->curr_code);
      free(countries[i]->currency);
      free(countries[i]->telephone);
      free(countries[i]->capital);
      free(countries[i]);
   }
   free(countries);
   countries = NULL;
   numcountry = 0;
}

void freeTZ()
{
int i;
   if (timezones) { 
      for (i=0; i<numtz; i++) {
         free(timezones[i]->code);
         free(timezones[i]->flags);
         free(timezones[i]->name);
         free(timezones[i]->posix);
	 free(timezones[i]);
      }
      free(timezones);
      timezones = NULL;
   }
   numtz = 0;
}

void freeLocations()
{
int i;
   if (locations) {
      for (i=0; i<numloc; i++) {
         if (locations[i]->name) free(locations[i]->name);
         if (locations[i]->region) free(locations[i]->region);
	 free(locations[i]);
      }
      free(locations);
      locations = NULL;
   }
   numloc = 0;
}

void freepeaks()
{
int i;
   if (locations) {
      for (i=0; i<numloc; i++) {
         free(locations[i]->name);
         free(locations[i]->region);
	 free(locations[i]);
      }
      free(locations);
      locations = NULL;
   }
   numloc = 0;
}

int round2int (double x)
{
        if (x>0) 
           return (int)(x+0.5) ;
	else
	   return -(int)(-x+0.5);
}

void error_tz(char * sect, char * str)
{
     fprintf(stderr, "Incorrect timezone label in section <%s> : %s\n", 
                     sect, str);
}

#define BIGNUM 100000000
void reordering(char mode)
{
long int v, *value;
int i, j, k, n, step;
int *order, *rank, *minrank;

   value = malloc(numloc*sizeof(long int));
   order = malloc(numloc*sizeof(int));

   rank = malloc(numloc*sizeof(int));
   minrank = malloc(numcountry*sizeof(int));

   for (step=0; step<=2; step++) {
      bzero(value, numloc*sizeof(long int));
      bzero(order, numloc*sizeof(int));
      n = 0;
      for (i=0; i<numloc; i++) {
         if (locations[i]->obj==mode) {
	    if (i%500==0)
	       fprintf(stderr, "Step %d, %d \n", step, i);
	    if (mode=='p')
	       v = locations[i]->alt;
	    else
	       v = locations[i]->population;
	    if (step==1)
	       v += BIGNUM*timezones[locations[i]->tz]->continent;
	    else
	    if (step==2)
	       v += BIGNUM*timezones[locations[i]->tz]->country;
	    value[n] = v;
	    order[n] = i;
	    for (j=0; j<=n-1; j++) {
	       if (v>value[j]) {
	          for (k=n; k>j; k--) {
		     value[k] = value[k-1]; 
		     order[k] = order[k-1]; 
	          }
	          value[j] = v;
	          order[j] = i;
                  break;
	       }
	    }
            ++n;
         }
      }
      for (i=0; i<n; i++)
         rank[order[i]] = i;
      for (i=0; i<numcountry; i++)
         minrank[i] = numloc;

      if (step==0 && mode!='p') {
         for (i=0; i<numloc; i++)
	    if (locations[i]->obj==mode) {
	       locations[i]->worldrank = rank[i]+1;
	    }
      }

      if (step==1) {
         for (i=0; i<numloc; i++) {
            if (locations[i]->obj==mode) {
	       j = timezones[locations[i]->tz]->continent;
	       if (rank[i]<minrank[j]) minrank[j] = rank[i];
            }
         }
         for (i=0; i<numloc; i++) {
	    if (locations[i]->obj==mode) {
	       j = timezones[locations[i]->tz]->continent;
	       locations[i]->contrank = rank[i]-minrank[j]+1;
	    }
	 }
      }

      if (step==2) {
         for (i=0; i<numloc; i++) {
            if (locations[i]->obj==mode) {
	       j = timezones[locations[i]->tz]->country;
	       if (rank[i]<minrank[j]) minrank[j] = rank[i];
            }
         }
         for (i=0; i<numloc; i++) {
	    if (locations[i]->obj==mode) {
	       j = timezones[locations[i]->tz]->country;
	       locations[i]->rank = rank[i]-minrank[j]+1;
	    }
	 }
      }
   }

   free(minrank);   
   free(value);   
   free(order);
   free(rank);
}

void parse_locfile (ImageLayout *s) {
	FILE *f;	
	char buf[512];
        char field[11][128];
	char *sect[] = { "", "timezones", "cities" };
        char *str, *ptr, *ptrp;
	int bool, i, j, itz, numfields, datatype;

	if (!s->locfile || !*s->locfile) return;

	f = fopen(s->locfile, "r");
	if (f == NULL) {
	   sprintf(buf, msg[COULDNT_OPEN_LOCFILE], s->locfile);
	   if (runlevel==0)
	      fprintf(stderr, "%s\n", buf);
	   else {
	      drawWinString(s, buf, BOTTOM);
	      XFlush(dpy);
	      usleep(WARNING_TIME);
	   }
	   return;
	}

	numfields = 0;
	itz = 0;
	datatype = T_NONE;
        while((str=fgets(buf, 510, f))) {
	   while (*str && isspace(*str)) ++str;
	   if (!*str || *str=='%') continue;
	   if (*str=='<') {
	      if (!strncmp(str, "<reset_countries>", 17)) {
                 freeCountries();
	         continue;
	      }
	      if (!strncmp(str, "<reset_TZ>", 10)) {
                 freeTZ();
	         continue;
	      }
	      if (!strncmp(str, "<reset_locations>", 17)) {
                 freeLocations();
	         continue;
              }
	      if (!strncmp(str, "<countries>", 11)) {
	         datatype = T_COUNTRIES;
	         numfields = 11;
	         continue;
	      }
	      if (!strncmp(str, "<timezones>", 11)) {
	         datatype = T_TIMEZONES;
	         numfields = 6;
	         continue;
	      }
	      if (!strncmp(str, "<cities>", 8)) {
	         datatype = T_CITIES;
	         numfields = 11;
	         continue;
	      }
	      if (!strncmp(str, "<peaks>", 7)) {
	         datatype = T_PEAKS;
	         numfields = 9;
	         continue;
	      }
	      if (!strncmp(str, "<locations>", 11)) {
	         datatype = T_LOCATIONS;
	         numfields = 7;
	         continue;
	      }
	      if (datatype==T_COUNTRIES && !strncmp(str, "</countries>", 12)) {
	         datatype = T_NONE;
	         numfields = 0;
	         continue;
	      }
	      if (datatype==T_TIMEZONES && !strncmp(str, "</timezones>", 12)) {
	         datatype = T_NONE;
	         numfields = 0;
	         continue;
	      }
	      if (datatype==T_CITIES && !strncmp(str, "</cities>", 9)) {
	         datatype = T_NONE;
	         numfields = 0;
	         continue;
	      }
	      if (datatype==T_PEAKS && !strncmp(str, "</peaks>", 8)) {
	         datatype = T_NONE;
	         numfields = 0;
	         continue;
	      }
	      if (datatype==T_LOCATIONS && !strncmp(str, "</locations>", 12)) {
	         datatype = T_NONE;
	         numfields = 0;
	         continue;
	      }
	   }
           bool = 0;
           for (i=0; i<numfields; i++) {
              while (*str && isspace(*str)) ++str;
	      ptr = str;
              while (*ptr && *ptr!='|') ++ptr;
	      if (!*ptr && i<numfields-1) bool=1;
              *ptr = '\0';
              ptrp = ptr+1;
	      --ptr;
              while(isspace(*ptr)) { *ptr = '\0'; --ptr; }
              strcpy(field[i], str);
	      str = ptrp;
           }
           if (bool) {
              fprintf(stderr, msg[FIELD_MISSING], field[0], sect[datatype]);
              fprintf(stderr, "\n");
	   }
	   if (datatype==T_COUNTRIES) {
	      i = (field[0][0]-'A')*26+(field[0][1]-'A');
              country_table[i] = numcountry;
	      countries = realloc(countries,
                                    (numcountry+1)*sizeof(Country *));
	      countries[numcountry] = malloc(sizeof(Country));
	      strcpy(countries[numcountry]->code2, field[0]);
	      strcpy(countries[numcountry]->code3, field[1]);
	      strcpy(countries[numcountry]->iana, field[3]);
	      countries[numcountry]->name = strdup(field[2]);
	      countries[numcountry]->flags = strdup(field[4]);
	      countries[numcountry]->curr_code = strdup(field[5]);
	      countries[numcountry]->currency = strdup(field[6]);
	      countries[numcountry]->telephone = strdup(field[7]);
	      countries[numcountry]->capital = strdup(field[8]);
	      countries[numcountry]->area = atoi(field[9]);
	      countries[numcountry]->population = atoi(field[10]);
#ifdef DEBUG
	      fprintf(stderr, "%d : %s\n", numcountry, 
                                           countries[numcountry]->name);
#endif
	      ++numcountry;
	   }
           else
	   if (datatype==T_TIMEZONES) {
              timezones = (Timezone **)
                  realloc(timezones, (numtz+1)*sizeof(Timezone *));
	      timezones[numtz] = malloc(sizeof(Timezone));
	      timezones[numtz]->code = strdup(field[0]);
              i = (field[1][0]-'A')*26+(field[1][1]-'A');
	      timezones[numtz]->country = country_table[i];
	      timezones[numtz]->flags = strdup(field[3]);
	      timezones[numtz]->name = strdup(field[4]);
	      timezones[numtz]->posix = strdup(field[5]);
	      for (i=0; i<=5; i++) {
	          if (!strcasecmp(field[2], continent[i])) {
		     timezones[numtz]->continent = i;
		     break;
		  }
	      }
	      ++numtz;
	   }
	   else
           if (datatype==T_CITIES) {
              locations = (Location **)
                  realloc(locations, (numloc+1)*sizeof(Location *));
	      locations[numloc] = malloc(sizeof(Location));
	      locations[numloc]->obj = 'c';
	      locations[numloc]->name = strdup(field[1]);
	      locations[numloc]->status = field[2][0];
	      locations[numloc]->region = strdup(field[3]);
	      locations[numloc]->population = atoi(field[4]);
	      locations[numloc]->rank = atoi(field[5]);
	      locations[numloc]->contrank = atoi(field[6]);
	      locations[numloc]->worldrank = atoi(field[7]);
	      locations[numloc]->lat = atof(field[8]);
	      locations[numloc]->lon = atof(field[9]);
	      if (index(field[10],'?'))
                 locations[numloc]->alt = -30000;
	      else
	         locations[numloc]->alt = round2int(atof(field[10]));
	      locations[numloc]->tz = -1;
	      bool = 1;
              for (i=0; i<numtz; i++) {
                 j = (i+itz)%numtz;
		 if (!strcmp(field[0], timezones[j]->code)) {
		    locations[numloc]->tz = j;
		    itz = j;
		    bool = 0;
                    break;
		 }
	      }
	      if (bool) error_tz("locations", field[0]);
	      ++numloc;
	   }
	   else
           if (datatype==T_PEAKS) {
              locations = (Location **)
                  realloc(locations, (numloc+1)*sizeof(Location *));
	      locations[numloc] = malloc(sizeof(Location));
	      locations[numloc]->obj = 'p';
	      locations[numloc]->status = '*';
	      locations[numloc]->tz = -1;
              bool = 1;
              for (i=0; i<numtz; i++) {
                 j = (i+itz)%numtz;
		 if (!strcmp(field[0], timezones[j]->code)) {
		    locations[numloc]->tz = j;
		    itz = j;
		    bool = 0;
                    break;
		 }
	      }
	      if (bool) error_tz("peaks", field[0]);
	      locations[numloc]->name = strdup(field[1]);
	      locations[numloc]->region = strdup(field[2]);
	      locations[numloc]->rank = atoi(field[3]);
	      locations[numloc]->contrank = atoi(field[4]);
	      locations[numloc]->worldrank = atoi(field[5]);
	      locations[numloc]->lat = atof(field[6]);
	      locations[numloc]->lon = atof(field[7]);
	      locations[numloc]->alt = atoi(field[8]);
	      ++numloc;
	   }
	   else
           if (datatype==T_LOCATIONS) {
              locations = (Location **)
                  realloc(locations, (numloc+1)*sizeof(Location *));
	      locations[numloc] = malloc(sizeof(Location));
	      locations[numloc]->obj = 'l';
	      locations[numloc]->status = field[2][0];
	      locations[numloc]->name = strdup(field[0]);
	      locations[numloc]->region = NULL;
	      locations[numloc]->rank = (int)(field[4][1]-'0');
	      locations[numloc]->worldrank = 0;
	      locations[numloc]->population = atoi(field[3]);
	      locations[numloc]->lat = atof(field[5]);
	      locations[numloc]->lon = atof(field[6]);
	      locations[numloc]->alt = 0;
              locations[numloc]->tz = -1;
	      for (i=0; i<=5; i++) {
	          if (!strcasecmp(field[1], continent[i])) {
		     locations[numloc]->tz = i;
		     break;
		  }
	      }
	      ++numloc;
	   }
	}
	      
	fclose(f);

	if (ranking) {
	   char alt[20];
	   reordering((char)ranking);
	   for (i=0; i<numloc; i++) {
	      if (locations[i]->alt<=-30000)
		 strcpy(alt, "?");
	      else
		 sprintf(alt, "%d", locations[i]->alt);
	      if (locations[i]->obj=='c' && ranking==(int)'c')
		 printf("%s|%s|%c|%s|%d|%d|%d|%d|%.3f|%.3f|%s\n",
		    timezones[locations[i]->tz]->code,
		    locations[i]->name,
		    locations[i]->status,
		    locations[i]->region,
		    locations[i]->population,
		    locations[i]->rank,
		    locations[i]->contrank,
		    locations[i]->worldrank,
		    locations[i]->lat,
		    locations[i]->lon,
		    alt);
	      else
	      if (locations[i]->obj=='p' && ranking==(int)'p')
		 printf("%s|%s|%s|%d|%d|%d|%.3f|%.3f|%d\n",
		    timezones[locations[i]->tz]->code,
		    locations[i]->name,
		    locations[i]->region,
		    locations[i]->rank,
		    locations[i]->contrank,
		    locations[i]->worldrank,
		    locations[i]->lat,
		    locations[i]->lon,
		    locations[i]->alt);
	      else
	      if (locations[i]->obj=='l' && ranking==(int)'l')
		 printf("%s|%s|%c|%d|p%d|%.3f|%.3f\n",
		    locations[i]->name,
		    continent[locations[i]->tz],
		    locations[i]->status,
		    locations[i]->population,
		    locations[i]->rank,
		    locations[i]->lat,
		    locations[i]->lon);
	   }
	}
}

void load_city_filter(ImageLayout * scene, char *str)
{
     char ch;
     int j, k;

     *scene->ropt.city_filter ='\0';
     k = 0;
     for (j=0;j<strlen(str);j++) {
        ch = toupper(str[j]);
        if (index(scene->ropt.city_filter, ch)) continue;
	if (index(city_filter0, ch)) {
	   scene->ropt.city_filter[k] = ch;
	   ++k;
	   scene->ropt.city_filter[k] = '\0';
	}
     }
}

void load_loc_filter(ImageLayout * scene, char *str)
{
     char ch;
     int j, k;

     *scene->ropt.loc_filter ='\0';
     k = 0;
     for (j=0;j<strlen(str);j++) {
        ch = toupper(str[j]);
        if (index(scene->ropt.loc_filter, ch)) continue;
	if (index(loc_filter0, ch)) {
	   scene->ropt.loc_filter[k] = ch;
	   ++k;
	   scene->ropt.loc_filter[k] = '\0';
	}
     }
}

char * list_languages()
{
     int i;
     char dirname[256];
     char *ptr, *str;
     struct dirent *dirent;
     DIR *dir;

     sprintf(dirname, "%s/i18n", SHAREDIR);
     dir = opendir(dirname);
     if (dir == NULL) return "";
     
     i = 0;
     str = (char *)malloc((3+i)*sizeof(char));
     str[0] = '\0';

     for (dirent = readdir(dir); dirent != NULL; dirent = readdir(dir)) {
         ptr = dirent->d_name;
	 if (!strncmp(ptr, "xrmap_msg.", 10)) {
	    ptr = ptr+10;
	    if (i!=0) {
	       str[i] = ',';
	       str[++i] = '\0';
	    }
	    i += strlen(ptr);
	    str = realloc(str, (3+i)*sizeof(char));
	    strcat(str, ptr);
	 }
     }
     return str;
}

void read_i18n_file(ImageLayout * scene)
{
FILE *fd;
char *ptr;
char buf[512];
char i18n_file[256];
int i, j, l, n;
 
    if (!*language) {
       if (getenv("LANG")) {
          strncpy(language, getenv("LANG"), 2);
	  language[2] = '\0';
       } else
          strcpy(language, "en"); 
    }
    sprintf(i18n_file, "%s/xrmap_msg.%s", I18NPATH, language);

    fd = fopen(i18n_file, "r");
    if (!fd) {
       fprintf(stderr, "Couldn't find %s !!\n", i18n_file);
     retry:
       strcpy(language, "en");
       sprintf(i18n_file, "%s/xrmap_msg.en", I18NPATH);
       fprintf(stderr, "Using instead default %s\n", i18n_file);
       fd = fopen(i18n_file, "r");      
    }
    if (!fd) {
       fprintf(stderr, "Cannot open language file %s !!\n", i18n_file);
       exit(-1);
    }

    if (msg) {
       for (i=0; i<num_msg; i++) free(msg[i]);
       free(msg);
       msg = NULL;
    }

    num_msg = 0;
    setup_change |= LANG_CHANGE;
    while ((ptr = fgets(buf, 510, fd))) {
       if (!strncmp(ptr, "#END_OF_I18N_FILE", 17)) break;
       if (*ptr=='#') continue;
       msg = (char **) realloc(msg, (num_msg+1)*sizeof(char *));
       l = strlen(ptr)-1;
       if (ptr[l]=='\n') {
          ptr[l] = '\0';
	  --l;
       }
       while(l>=0 && isspace(ptr[l])) {
	  ptr[l] = '\0';
	  --l;
       }
       if (l<0) continue;
       msg[num_msg] = strdup(ptr);
       ++num_msg;
    }
    fclose(fd);
    if (num_msg<END_OF_I18N_FILE) {
       fprintf(stderr, "Language dialog file %s incomplete !!\n", i18n_file);
       if (!strstr(i18n_file, "xrmap_msg.en")) goto retry;
       exit(-1);
    }

    n = END_LABELS_GENERALMENU-BEGIN_LABELS_GENERALMENU;
    spacing_generalmenu = atoi(msg[SPACING_GENERALMENU]);
    l = n * spacing_generalmenu + 20;
    if (labels_generalmenu) free(labels_generalmenu);
    labels_generalmenu = 
       (char *)malloc((l+1) * sizeof(char));
    for (i=0; i<l; i++) labels_generalmenu[i] = ' ';
    labels_generalmenu[l] = '\0';

    for (i=0; i<n; i++) {
        ptr = msg[BEGIN_LABELS_GENERALMENU+i];
        for (j=0; j<strlen(ptr); j++)
	    labels_generalmenu[i*spacing_generalmenu+j] = ptr[j];
    }

    n = END_LABELS_FILEMENU-BEGIN_LABELS_FILEMENU;
    spacing_filemenu = atoi(msg[SPACING_FILEMENU]);
    l = n * spacing_filemenu + 20;
    if (labels_filemenu) free(labels_filemenu);
    labels_filemenu = 
       (char *)malloc((l+1) * sizeof(char));
    for (i=0; i<l; i++) labels_filemenu[i] = ' ';
    labels_filemenu[l] = '\0';

    for (i=0; i<n; i++) {
        ptr = msg[BEGIN_LABELS_FILEMENU+i];
        for (j=0; j<strlen(ptr); j++)
	    labels_filemenu[i*spacing_filemenu+j] = ptr[j];
    }

    n = END_LABELS_OPTIONMENU-BEGIN_LABELS_OPTIONMENU;
    spacing_optionmenu = atoi(msg[SPACING_OPTIONMENU]);
    l = n * spacing_optionmenu + 20;
    if (labels_optionmenu) free(labels_optionmenu);
    labels_optionmenu = 
       (char *)malloc((l+1) * sizeof(char));
    for (i=0; i<l; i++) labels_optionmenu[i] = ' ';
    labels_optionmenu[l] = '\0';

    for (i=0; i<n; i++) {
        ptr = msg[BEGIN_LABELS_OPTIONMENU+i];
        for (j=0; j<strlen(ptr); j++)
	    labels_optionmenu[i*spacing_optionmenu+j] = ptr[j];
    }

    n = END_LABELS_COLORMENU-BEGIN_LABELS_COLORMENU;
    spacing_colormenu = atoi(msg[SPACING_COLORMENU]);
    l = n * spacing_colormenu + 20;
    if (labels_colormenu) free(labels_colormenu);
    labels_colormenu = 
       (char *)malloc((l+1) * sizeof(char));
    for (i=0; i<l; i++) labels_colormenu[i] = ' ';
    labels_colormenu[l] = '\0';

    for (i=0; i<n; i++) {
        ptr = msg[BEGIN_LABELS_COLORMENU+i];
        for (j=0; j<strlen(ptr); j++)
	    labels_colormenu[i*spacing_colormenu+j] = ptr[j];
    }

    n = END_LABELS_HELPMENU-BEGIN_LABELS_HELPMENU;
    spacing_helpmenu = atoi(msg[SPACING_HELPMENU]);
    l = n * spacing_helpmenu + 20;
    if (labels_helpmenu) free(labels_helpmenu);
    labels_helpmenu = 
       (char *)malloc((l+1) * sizeof(char));
    for (i=0; i<l; i++) labels_helpmenu[i] = ' ';
    labels_helpmenu[l] = '\0';

    for (i=0; i<n; i++) {
        ptr = msg[BEGIN_LABELS_HELPMENU+i];
        for (j=0; j<strlen(ptr); j++)
	    labels_helpmenu[i*spacing_helpmenu+j] = ptr[j];
    }

    l = END_PROJTYPES - BEGIN_PROJTYPES;
    if (projtype) free(projtype);
    projtype = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) projtype[i] = msg[BEGIN_PROJTYPES+i];

    l = END_ITEMTYPES - BEGIN_ITEMTYPES;
    if (itemtype) free(itemtype);
    itemtype = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) itemtype[i] = msg[BEGIN_ITEMTYPES+i];

    l = END_MODE_DESCRIPTION - BEGIN_MODE_DESCRIPTION;
    if (mode_descr) free(mode_descr);
    mode_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) mode_descr[i] = msg[BEGIN_MODE_DESCRIPTION+i];

    l = END_DATAFILE_DESCRIPTION - BEGIN_DATAFILE_DESCRIPTION;
    if (datafile_descr) free(datafile_descr);
    datafile_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) datafile_descr[i] = msg[BEGIN_DATAFILE_DESCRIPTION+i];

    l = END_PRINTCFG_DESCRIPTION - BEGIN_PRINTCFG_DESCRIPTION;
    if (printcfg_descr) free(printcfg_descr);
    printcfg_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) printcfg_descr[i] = msg[BEGIN_PRINTCFG_DESCRIPTION+i];

    l = END_SAVE_DESCRIPTION - BEGIN_SAVE_DESCRIPTION;
    if (save_descr) free(save_descr);
    save_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) save_descr[i] = msg[BEGIN_SAVE_DESCRIPTION+i];

    l = END_SEARCHBY_DESCRIPTION - BEGIN_SEARCHBY_DESCRIPTION;
    if (searchby_descr) free(searchby_descr);
    searchby_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) searchby_descr[i] = msg[BEGIN_SEARCHBY_DESCRIPTION+i];

    l = END_PARAMETER_DESCRIPTION - BEGIN_PARAMETER_DESCRIPTION;
    if (parameter_descr) free(parameter_descr);
    parameter_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) parameter_descr[i] = msg[BEGIN_PARAMETER_DESCRIPTION+i];

    l = END_DISPLAY_DESCRIPTION - BEGIN_DISPLAY_DESCRIPTION;
    if (display_descr) free(display_descr);
    display_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) display_descr[i] = msg[BEGIN_DISPLAY_DESCRIPTION+i];

    l = END_FONT_DESCRIPTION - BEGIN_FONT_DESCRIPTION;
    if (font_descr) free(font_descr);
    font_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) font_descr[i] = msg[BEGIN_FONT_DESCRIPTION+i];

    l = END_MARKS_DESCRIPTION - BEGIN_MARKS_DESCRIPTION;
    if (marks_descr) free(marks_descr);
    marks_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) marks_descr[i] = msg[BEGIN_MARKS_DESCRIPTION+i];

    l = END_CONTINENT_DESCRIPTION - BEGIN_CONTINENT_DESCRIPTION;
    if (continent_descr) free(continent_descr);
    continent_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) continent_descr[i] = msg[BEGIN_CONTINENT_DESCRIPTION+i];

    l = END_CATEGORY_DESCRIPTION - BEGIN_CATEGORY_DESCRIPTION;
    if (category_descr) free(category_descr);
    category_descr = (char **)malloc(l*sizeof(char *));
    for (i=0; i<l; i++) category_descr[i] = msg[BEGIN_CATEGORY_DESCRIPTION+i];

    for (i=BEGIN_COLOR_DESCRIPTION; i<END_COLOR_DESCRIPTION; i++) {
       ptr = msg[i];
       ptr[2] = '\0';
       l = atoi(ptr);
       if (l>=0 && l<C_ENDCOLOR)
          scene->color[l].comment = ptr+3;
    }
}

void fix_features_description(ImageLayout * scene)
{
int i, j, n;

   for (i=0; i<6; i++) {
      if (feature_descr[i]) free(feature_descr[i]);
      feature_descr[i] = (char **) 
         malloc(feature_length[i]*sizeof(char **));
      for (n=0; n<feature_length[i]; n++) {
	 j = feature_index[i]+n;
         if (scene->color[j].bool != INACTIVE) {
            feature_descr[i][n] = scene->color[j].comment;
	 }
      }
   }
}

int wild_strcmp(char * string, char * needle)
{
int i, j, l;
    i = 0;
    while (string[i]!='\0' && needle[i]!='\0' && 
           string[i]==needle[i] && needle[i]!='*') ++i;
    if (needle[i]=='\0')
       return (string[i]!='\0');
    if (needle[i]!='*') 
       return 1;
    l = strlen(string+i);
    for (j=0; j<=l; j++)
       if (!wild_strcmp(string+i+j, needle+i+1)) return 0;
    return 1;
}

int string_cmp(char * string, char * needle)
{
int i, l;
char *ptr;
unsigned char c;

    l = strlen(string); 
    ptr = NULL;

    for (i=0; i<l; i++) {
       c = string[i];
       if (c<='Z') {
	 if (c>='A' && c<='Z') c += 32;
	 else
	 if (c=='-') c=' ';
	 else
	 if (c=='(') {
	    c='\0';
	    ptr = string+i+1;
	 }
         else
	 if (c==')') c='\0';
	 string[i] = c;
       } else
       if (c>=192)
	 string[i] = cvs[c-192];
    }

    if (!wild_strcmp(string, needle)) return 0;
    if (ptr && !wild_strcmp(ptr, needle)) return 0;
    return 1;
}

void simplify_string(char * str)
{
    int i, j=0, l;
    l = strlen(str);
    for (i=0; i<l; i++) {
       if (str[i]=='\\') {
	  ++i;
	  if (str[i]=='n') {
	     str[j] = ' ';
	     ++j;
	  }
       } else {
	  str[j] = str[i];
	  ++j;
       }
    }
    str[j] = '\0';
}

void search_city(ImageLayout * scene)
{
int i, l;
char *needle;
char fieldname[256];
unsigned char c;
 
    l = strlen(search_string); 
    needle = (char *)malloc(sizeof(char)*(l+2));

    for (i=0; i<=l; i++) {
       c = search_string[i];
       if (c<='Z') {
	 if (c=='-') c=' ';
	 if (c>='A' && c<='Z') c += 32;
	 needle[i] = c;
       } else
       if (c>=192)
	 needle[i] = cvs[c-192];
       else
         needle[i] = c;
    }

    num_found = 0;
    start_list = 0;
    for (i=0; i<numloc; i++) {
       if (search_by==1) {
          strcpy(fieldname, locations[i]->name);
	  if (locations[i]->obj!='c') simplify_string(fieldname);
       } else
       if (search_by==2) {
	  if (locations[i]->obj!='c') continue;
          strcpy(fieldname, 
                 countries[timezones[locations[i]->tz]->country]->name);
       }
       else
       if (search_by==3) {
	  if (locations[i]->obj!='c') continue;
          strcpy(fieldname, locations[i]->region);
       }
       if (!string_cmp(fieldname, needle)) {
	  loc_pointer = (int *)realloc(loc_pointer, (num_found+1)*sizeof(int));
	  loc_pointer[num_found] = i;
          ++num_found;
       }
    }
    if (num_found==1) 
       search_index = *loc_pointer;
    else
       search_index = -1;
    free(needle);
}

void search_output(ImageLayout * scene, int force);

void check_hierarchy(ImageLayout * scene)
{
char str[6];
int busy[5];
int i, j;

    i=0;
    j=0;
    bzero(busy, sizeof(busy));

    for (i=0; i<strlen(scene->gopt.hierarchy); i++) {
      if (scene->gopt.hierarchy[i]>='0' && 
          scene->gopt.hierarchy[i]<='4' && 
          !busy[scene->gopt.hierarchy[i]-'0']) {
	 busy[scene->gopt.hierarchy[i]-'0'] = 1;
	 str[j] = scene->gopt.hierarchy[i];
	 ++j;
      }
    }
    str[j] = '\0';
    strcpy(scene->gopt.hierarchy, str);
}

void free_data_buffers(ImageLayout * scene)
{
    if (segment_index) {
       free(segment_index);
       segment_index = NULL;
    }
    if (fullmap_buffer) {
       free(fullmap_buffer);
       fullmap_buffer = NULL;
       segment_buffer = NULL;
    } else {
       free (segment_buffer);
       segment_buffer = NULL;
    }
}

int parse_cmd_line(ImageLayout * scene, char *str)
{
        char *string, *ptr, *ptr_err;
	int w0, h0;

	if (!str || !*str) return -1;
	string = strdup(str);

	w0 = scene->width;
	h0 = scene->height;

	while ((ptr = (char*) strtok(string, " ")) != NULL) {
	   ptr_err = ptr;
	   string = NULL;
	   if (!runlevel && !strpcmp(ptr, "help", 1)) {
	      usage();
	      exit(0);
	   } else
	   if (!runlevel && !strpcmp(ptr, "version", 1)) {
              printf(msg[VERSION_FORMAT], 
                     PACKAGE, VERSION, DATEPKG, "\n", COPYRIGHT);
	      printf("\n");
              exit(0);
	   }
           else
	   if (!runlevel && !strpcmp(ptr, "menu", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              start_menu = (int)*ptr;
	   }
           else
	   if (!strpcmp(ptr, "language", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if (strcmp(ptr, language)) {
	         strncpy(language, ptr, 2);
	         language[2] = '\0';
	         read_i18n_file(scene);
	         fix_features_description(scene);
	      }
	   }
	   else
	   if (!strpcmp(ptr, "mapfile", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
	      if (!strcasecmp(ptr, "DEFAULT"))
                 strncpy(scene->mapfile, MAPFILE, 255);
	      else
                 strncpy(scene->mapfile, ptr, 255);
	      scene->mapfile[255] = '\0';
	      free_data_buffers(scene);
	   } 
           else 
	   if (!strpcmp(ptr, "rcfile", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr || (ptr && !strcasecmp(ptr, "NULL")))
		 *scene->rcfile = '\0';
	      else {
	         if (!strcasecmp(ptr, "DEFAULT"))
                    strncpy(scene->rcfile, RCFILE, 255);
	         else
                    strncpy(scene->rcfile, ptr, 255);
		 scene->rcfile[255] = '\0';
	    	 if (runlevel) parse_rcfile(scene);
	      }
	   } 
           else
	   if (!strpcmp(ptr, "locfile", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr || (ptr && !strcasecmp(ptr, "NULL")))
		 *scene->locfile = '\0';
	      else {
	         if (!strcasecmp(ptr, "DEFAULT"))
                    strncpy(scene->locfile, LOCFILE, 255);
	         else
                    strncpy(scene->locfile, ptr, 255);
		 scene->locfile[255] = '\0';
	         parse_locfile(scene);
	      }
	   }
           else
	   if (!strpcmp(ptr, "outfile", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr || (ptr && !strcasecmp(ptr, "NULL")))
		 *scene->outfile = '\0';
	      else {
                 strncpy(scene->outfile, ptr, 255);
		 scene->outfile[255] = '\0';
	      }
	   }
	   else
	   if (!strpcmp(ptr, "+memory", 4)) {
              use_memory = 1;
	   } 
	   else
	   if (!strpcmp(ptr, "-memory", 4)) {
              use_memory = 0;
	      if (fullmap_buffer) {
                 free(fullmap_buffer);
	         fullmap_buffer = NULL;
		 segment_buffer = NULL;
	      }
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "ppmoutput", 3)) {
              img_output = OUT_PPM;
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "xpmoutput", 3)) {
              img_output = OUT_XPM;
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "psoutput", 3)) {
              img_output = OUT_PS;
	   } 
	   else
	   if (!runlevel && !strpcmp(ptr, "preview", 7)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) preview = atof(ptr); else goto error;
              if (preview<0.0) goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "psviewer", 3)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) {
		 free(ps_viewer);
		 ps_viewer = strdup(ptr); 
	      } else goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "imviewer", 3)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) {
		 free(im_viewer);
		 im_viewer = strdup(ptr); 
	      } else goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "printcmd", 3)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) {
		 free(print_cmd);
		 print_cmd = strdup(ptr); 
	      } else goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "editor", 4)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) {
		 free(editor_cmd);
		 editor_cmd = strdup(ptr); 
	      } else goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "pswidth", 3)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) ps_width = atof(ptr); else goto error;
	      if (ps_width<=0) goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "pslmargin", 4)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) ps_lm = atof(ptr); else goto error;
	      if (ps_lm<0) goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "psbmargin", 4)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) ps_bm = atof(ptr); else goto error;
	      if (ps_bm<0) goto error;
	   } 
	   else
	   if (!strpcmp(ptr, "psrot", 3)) {
              ptr = (char*) strtok(NULL, " ");
              if (ptr) ps_rot = atoi(ptr); else goto error;
	      if (ps_rot<0) ps_rot = -90;
	      if (ps_rot>0) ps_rot = 90;
	   } 
	   else
	   if (!strpcmp(ptr, "pscolor", 3)) {
              ps_grayscale = 0;
	   } 
	   else
	   if (!strpcmp(ptr, "psgrayscale", 3)) {
              ps_grayscale = 1;
	   } 
	   else
	   if (!strpcmp(ptr, "longitude", 2)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
                 scene->gopt.yrot = atof(ptr); 
              else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "latitude", 2)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
                 scene->gopt.xrot = atof(ptr); 
              else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "inclination", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
                 scene->gopt.zrot = atof(ptr); 
              else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "search", 4)) {
    	      ptr = (char*) strtok(NULL, " ");
	      if (ptr) {
		 free(search_string);
		 search_string = strdup(ptr);
	         if (runlevel) {
		    search_city(scene);
		    search_output(scene, 1);
		 }
	      }
	   } 
	   else
	   if (!strpcmp(ptr, "field", 4)) {
    	      ptr = (char*) strtok(NULL, " ");
	      if (ptr) {
		 if (!strcasecmp(ptr, "name"))
		    search_by = 1;
		 else
		 if (!strcasecmp(ptr, "country"))
		    search_by = 2;
		 else
		 if (!strcasecmp(ptr, "region"))
		    search_by = 3;
		 else {
                    search_by = atoi(ptr);
		    if (search_by<1 || search_by>3) goto error;
		 }
	      } else
		 goto error;
	   }
	   else
	   if (!strpcmp(ptr, "reverse", 1)) {
    	      reverse = 1 - reverse;
	      if (runlevel) get_all_colors(scene);
	   } 
	   else
	   if (!strpcmp(ptr, "resetcolors", 3)) {
	      if (runlevel && color_depth<=8) get_all_colors(scene);
	   } 
           else
	   if (!strpcmp(ptr, "font", 2)) {
	      int num;
	      char *ptr2;
              ptr = (char*) strtok(NULL, " ");
      	      if (!ptr) goto error;
	      if (!(ptr2=index(ptr,','))) goto error;
	      *ptr2 = '\0';
              num = atoi(ptr);
	      *ptr2 = ',';
	      if (num<0 || num>5) goto error;
	      if (!runlevel || !set_font(ptr2+1, num)) {
	         strncpy(scene->ropt.font[num], ptr2+1, 255);
	         scene->ropt.font[num][255] = '\0';
              } else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "width", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->width = atoi(ptr); else goto error;
	      if (scene->width < 16) goto error;
              if (runlevel && textpix) {
	         XFreePixmap(dpy, textpix);
		 textpix = 0;
	      }
	   }
	   else
	   if (!strpcmp(ptr, "height", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->height = atoi(ptr); else goto error;
	      if (scene->height < 16) goto error;
	   }
	   else
	   if (!strpcmp(ptr, "color", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) set_opt_color(scene, ptr, 0); else goto error;
	      set_bg_fg(scene);
	   }
	   else
	   if (!strpcmp(ptr, "category", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) {
                 if (set_flag(scene, ptr, 0))
		    goto error;
	      }
              else 
                 goto error;
	   }
	   else
	   if (!strpcmp(ptr, "continent", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) {
		 if (set_flag(scene, ptr, 1))
		    goto error;
	      }
              else 
                 goto error;
	   }
	   else
	   if (!strpcmp(ptr, "projection", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->projection = set_proj(ptr); else goto error;
              if (runlevel) show_title(scene, mainwin);
	   }
	   else
	   if (!strpcmp(ptr, "aspect", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->aspect = atof(ptr); else goto error;
	      if (scene->aspect<0.01) goto error;
	   }
	   else
	   if (!strpcmp(ptr, "zoom", 1)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->gopt.zoom = atof(ptr); else goto error;
	      if (scene->gopt.zoom<0.01) goto error;
	   }
	   else
	   if (!strpcmp(ptr, "gslat", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->ropt.spacing_lat = atof(ptr); else goto error;
	      check_spacing(scene);
	   }
	   else
	   if (!strpcmp(ptr,"gslon", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->ropt.spacing_lon = atof(ptr); else goto error;
	      check_spacing(scene);
	   }
	   else
	   if (!strpcmp(ptr, "gspacing", 3)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) scene->ropt.spacing_lat = atof(ptr); else goto error;
	      scene->ropt.spacing_lon = scene->ropt.spacing_lat;
	      check_spacing(scene);
	   } 
	   else
	   if (!strpcmp(ptr, "hierarchy", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr) {
		 strncpy(scene->gopt.hierarchy, ptr, 5);
		 scene->gopt.hierarchy[5] = '\0';
                 check_hierarchy(scene);
	      } else
		 goto error;
	   }
	   else
	   if (!strpcmp(ptr, "ranking", 4)) {
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              ranking = (int)*ptr;
	   }
	   else
	   if (!strpcmp(ptr, "marksteps", 3)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &scene->ropt.mark_step[0], &scene->ropt.mark_step[1], 
                            &scene->ropt.mark_step[2], &scene->ropt.mark_step[3], 
                            &scene->ropt.mark_step[4]))<5)
		 for (j=i; j<5; j++) 
                    scene->ropt.mark_step[j] = scene->ropt.mark_step[i-1];
	   }
           else
	   if (!strpcmp(ptr, "namesteps", 3)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &scene->ropt.name_step[0], &scene->ropt.name_step[1], 
                            &scene->ropt.name_step[2], &scene->ropt.name_step[3], 
                            &scene->ropt.name_step[4]))<5)
		 for (j=i; j<5; j++) 
                    scene->ropt.name_step[j] = scene->ropt.name_step[i-1];
	   }                          
	   else
	   if (!strpcmp(ptr, "peakmarks", 7)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &scene->ropt.peak_mark[0], &scene->ropt.peak_mark[1], 
                            &scene->ropt.peak_mark[2], &scene->ropt.peak_mark[3], 
                            &scene->ropt.peak_mark[4]))<5)
		 for (j=i; j<5; j++) 
                     scene->ropt.peak_mark[j] = scene->ropt.peak_mark[i-1];
	   }                          
	   else
	   if (!strpcmp(ptr, "peaksizes", 7)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &scene->ropt.peak_size[0], &scene->ropt.peak_size[1], 
                            &scene->ropt.peak_size[2], &scene->ropt.peak_size[3], 
                            &scene->ropt.peak_size[4]))<5)
		 for (j=i; j<5; j++) 
                    scene->ropt.peak_size[j] = scene->ropt.peak_size[i-1];
	   }                          
	   else
	   if (!strpcmp(ptr, "citymarks", 7)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &scene->ropt.city_mark[0], &scene->ropt.city_mark[1], 
                            &scene->ropt.city_mark[2], &scene->ropt.city_mark[3], 
                            &scene->ropt.city_mark[4]))<5)
		 for (j=i; j<5; j++) 
                     scene->ropt.city_mark[j] = scene->ropt.city_mark[i-1];
	   }                          
	   else
	   if (!strpcmp(ptr, "citysizes", 7)) {
	      int i, j;
              ptr = (char*) strtok(NULL, " ");
	      if (!ptr) goto error;
              if ((i=sscanf(ptr, "%d,%d,%d,%d,%d",
                            &scene->ropt.city_size[0], &scene->ropt.city_size[1], 
                            &scene->ropt.city_size[2], &scene->ropt.city_size[3], 
                            &scene->ropt.city_size[4]))<5)
		 for (j=i; j<5; j++) 
                    scene->ropt.city_size[j] = scene->ropt.city_size[i-1];
	   }                          
           else
	   if (!strpcmp(ptr, "cityfilter", 7)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
		 load_city_filter(scene, ptr);
              else goto error;
	   }
           else
	   if (!strpcmp(ptr, "locfilter", 6)) {
              ptr = (char*) strtok(NULL, " ");
	      if (ptr)
		 load_loc_filter(scene, ptr);
              else goto error;
	   }
	   else
	   if (!strpcmp(ptr, "+placename", 3)) 
	      scene->ropt.placetext = 1;
           else
	   if (!strpcmp(ptr, "-placename", 3)) 
	      scene->ropt.placetext = 0;
           else
	   if (!strpcmp(ptr, "+compress", 5))
	      im_compress = 1;
           else
	   if (!strpcmp(ptr, "-compress", 5))
	      im_compress = 0;
           else
	   if (!strpcmp(ptr, "+cities", 3)) {
	      scene->color[C_CITY_CIRCLE].bool = ON;
	      scene->color[C_CITY_TEXT].bool = ON;
	   }
	   else
	   if (!strpcmp(ptr, "-cities", 3)) {
	      scene->color[C_CITY_CIRCLE].bool = OFF;
	      scene->color[C_CITY_TEXT].bool = OFF;
	   }
	   else
	   if (!strpcmp(ptr, "+peaks", 3)) {
              scene->color[C_PEAK_TRIANGLE].bool = ON;
              scene->color[C_PEAK_TEXT].bool = ON;
	   }
	   else
	   if (!strpcmp(ptr, "-peaks", 3)) {
              scene->color[C_PEAK_TRIANGLE].bool = OFF;
              scene->color[C_PEAK_TEXT].bool = OFF;
	   }
	   else
	   if (!strpcmp(ptr, "+locations", 2)) {
              scene->color[C_LOC_LAND].bool = ON;
              scene->color[C_LOC_WATER].bool = ON;
	   } 
           else
	   if (!strpcmp(ptr, "-locations", 2)) {
              scene->color[C_LOC_LAND].bool = OFF;
              scene->color[C_LOC_WATER].bool = OFF;
	   } 
	   else
	   if (!strpcmp(ptr, "+earthcontour", 2)) 
              scene->color[C_EARTH_CONTOUR].bool = ON;
	   else
	   if (!strpcmp(ptr, "-earthcontour", 2)) 
              scene->color[C_EARTH_CONTOUR].bool = OFF;
	   else
	   if (!strpcmp(ptr, "+gridlines", 2)) {
              scene->color[C_LAT_GRID].bool = ON;
              scene->color[C_LON_GRID].bool = ON;
	   }
	   else
	   if (!strpcmp(ptr, "-gridlines", 2)) {
              scene->color[C_LAT_GRID].bool = OFF;
              scene->color[C_LON_GRID].bool = OFF;
	   }
	   else
	   if (!strpcmp(ptr, "+tropics", 5))
              scene->color[C_MAIN_LINES].bool = ON;
	   else
	   if (!strpcmp(ptr, "-tropics", 5))
              scene->color[C_MAIN_LINES].bool = OFF;
	   else
	   if (!strpcmp(ptr, "+transparency", 5)) 
              scene->gopt.transparent = ON;
	   else
	   if (!strpcmp(ptr, "-transparency", 5)) 
              scene->gopt.transparent = OFF;
	   else {
	      char *errstr;
	      ptr_err = ptr;
	      free(string);
	   error:
	      errstr = (char *)malloc(strlen(ptr_err)+100);
              sprintf(errstr, msg[INVALID_OPTION], ptr_err);
	      *cmd_string = '\0';
	      caret_pos = 0;
	      if (runlevel) {
	         drawWinString(scene, errstr, BOTTOM);
	         usleep(WARNING_TIME);
	      } else
		 fprintf(stderr, "%s\n", errstr);
	      free(errstr);
	      return 1;
	   }
	}
	free(string);
	if (runlevel) {
	   if (scene->width<=0) scene->width = w0;
	   if (scene->height<=0) scene->height = h0;
	   if (scene->width!=w0 || scene->height!=h0) { 
	      set_radius(scene);
              XResizeWindow(dpy, mainwin, scene->width, scene->height);
	      XFlush(dpy);
	   }
           if (textpix) {
               XFreePixmap(dpy, textpix);
	       textpix = 0;
	   }
	}
	return 0;
}

void value2str(ImageLayout * scene, int i, char * str, int mode)
{
        int j;
	char c;

        *str = '\0';   

	if (cmd_mode==MODE_FILE) {
	   if (cmd_type==M_FILE_DATA) {
	      switch(i) {
		 case 1:  strcpy(str, scene->rcfile); break;
		 case 2:  strcpy(str, scene->mapfile); break;
		 case 3:  strcpy(str, scene->locfile); break;
                 default: break;
	      }
	   }
	   if (cmd_type==M_FILE_PRINTCFG) {
	      switch(i) {
		 case 1:  sprintf(str, "%.1f", ps_width); break;
		 case 2:  sprintf(str, "%.1f", ps_lm); break;
		 case 3:  sprintf(str, "%.1f", ps_bm); break;
		 case 4:  sprintf(str, "%d", ps_rot); break;
		 case 5:  str[0] = (ps_grayscale)? '+':'-'; str[1] = '\0'; 
                          break;
		 case 6:  str[0] = (im_compress)? '+':'-'; str[1] = '\0'; 
                          break;
	         case 7:  strcpy(str, ps_viewer); break;
	         case 8:  strcpy(str, im_viewer); break;
	         case 9:  strcpy(str, print_cmd); break;
	         case 10:  strcpy(str, editor_cmd); break;
                 default: break;
	      }
	   }
	   if (cmd_type==M_FILE_SAVE && i==1)
	      strcpy(str, scene->outfile);
        }
	if (cmd_mode==MODE_OPTION) {
	   if (cmd_type==M_OPTION_PARAM) {
	      switch(i) {
		 case 1:  if (mode) 
                             sprintf(str, "%d = %s", scene->projection,
                                  projtype[scene->projection]); 
                          else
                             sprintf(str, "%d", scene->projection);
                          break;
		 case 2:  sprintf(str, "%d", scene->width); break;
		 case 3:  sprintf(str, "%d", scene->height); break;
		 case 4:  sprintf(str, "%.3f", scene->gopt.zoom); break;
		 case 5:  sprintf(str, "%.3f", scene->aspect); break;
		 case 6:  sprintf(str, "%.3f", scene->gopt.xrot); break;
		 case 7:  sprintf(str, "%.3f", scene->gopt.yrot); break;
		 case 8:  sprintf(str, "%.3f", scene->gopt.zrot); break;
		 case 9:  sprintf(str, "%.3f", scene->ropt.spacing_lat); break;
		 case 10: sprintf(str, "%.3f", scene->ropt.spacing_lon); break;
	         default: break;
	      }
	   }
	   if (cmd_type==M_OPTION_DISPLAY) {
	      switch(i) {
		 case 1: sprintf(str, "%s (mrk=0, int=1, nat=2, riv=3, cil=4)",
                                 scene->gopt.hierarchy);
		         break;
		 case 2: sprintf(str, "%d,%d,%d,%d,%d", 
			    scene->ropt.mark_step[0], scene->ropt.mark_step[1], 
                            scene->ropt.mark_step[2], scene->ropt.mark_step[3], 
                            scene->ropt.mark_step[4]); 
                         break;
		 case 3: sprintf(str, "%d,%d,%d,%d,%d", 
			    scene->ropt.name_step[0], scene->ropt.name_step[1], 
                            scene->ropt.name_step[2], scene->ropt.name_step[3], 
                            scene->ropt.name_step[4]); 
                         break;
		 case 4: sprintf(str, "%d,%d,%d,%d,%d", 
			    scene->ropt.peak_mark[0], scene->ropt.peak_mark[1], 
                            scene->ropt.peak_mark[2], scene->ropt.peak_mark[3], 
                            scene->ropt.peak_mark[4]); 
                         break;
		 case 5: sprintf(str, "%d,%d,%d,%d,%d", 
			    scene->ropt.peak_size[0], scene->ropt.peak_size[1], 
                            scene->ropt.peak_size[2], scene->ropt.peak_size[3], 
                            scene->ropt.peak_size[4]); 
                         break;
		 case 6: sprintf(str, "%d,%d,%d,%d,%d", 
			    scene->ropt.city_mark[0], scene->ropt.city_mark[1], 
                            scene->ropt.city_mark[2], scene->ropt.city_mark[3], 
                            scene->ropt.city_mark[4]); 
                         break;
		 case 7: sprintf(str, "%d,%d,%d,%d,%d", 
			    scene->ropt.city_size[0], scene->ropt.city_size[1], 
                            scene->ropt.city_size[2], scene->ropt.city_size[3], 
                            scene->ropt.city_size[4]); 
                         break;
	         case 8: strcpy(str, scene->ropt.city_filter);
		         break;
	         case 9: strcpy(str, scene->ropt.loc_filter);
		         break;
	         case 10: if (scene->ropt.placetext) 
                             strcpy(str, "+"); 
                         else 
                             strcpy(str, "-");
		         break;
	         case 11: if (scene->gopt.transparent)
                            strcpy(str, "+"); 
                         else 
                            strcpy(str, "-");
		         break;
	         case 12: sprintf(str, "%s (%s)", language, list_lang);
		         break;
	         default: break;
	      }
	   }

	   if (cmd_type==M_OPTION_FONT) {
              if (i>=1 && i<=6)
	         strcpy(str, scene->ropt.font[i-1]);
	   }

	   if (cmd_type==M_OPTION_MARKS) {
	      switch(i) {
              case 1: if (scene->color[C_CITY_CIRCLE].bool == OFF)
		         strcpy(str, "-");
                      if (scene->color[C_CITY_CIRCLE].bool == ON &&
			  scene->color[C_CITY_TEXT].bool == OFF) 
			 strcpy(str, "+");
                      if (scene->color[C_CITY_CIRCLE].bool == ON &&
			  scene->color[C_CITY_TEXT].bool == ON)
			 strcpy(str, "++");
                      break;

              case 2: if (scene->color[C_PEAK_TRIANGLE].bool == OFF)
			 strcpy(str, "-");
                      if (scene->color[C_PEAK_TRIANGLE].bool == ON &&
			  scene->color[C_PEAK_TEXT].bool == OFF)
			 strcpy(str, "+");
                      if (scene->color[C_PEAK_TRIANGLE].bool == ON &&
			  scene->color[C_PEAK_TEXT].bool == ON)
			 strcpy(str, "++");
                      break;

              case 3: j = scene->color[C_LOC_LAND].bool;
                      strcpy(str, (j)? "+":"-"); break;

              case 4: j = scene->color[C_LOC_WATER].bool;
                      strcpy(str, (j)? "+":"-"); break;

              case 5: j = scene->color[C_LON_GRID].bool;
                      strcpy(str, (j)? "+":"-"); break;

              case 6: j = scene->color[C_LAT_GRID].bool;
                      strcpy(str, (j)? "+":"-"); break;

              case 7: j = scene->color[C_MAIN_LINES].bool;
                      strcpy(str, (j)? "+":"-"); break;

              case 8: j = scene->color[C_EARTH_CONTOUR].bool;
                      strcpy(str, (j)? "+":"-"); break;

              case 9: j = scene->color[C_BG_SKY].bool;
                      strcpy(str, (j)? "+":"-"); break;

              case 10: j = scene->color[C_FG_STARS].bool;
                       strcpy(str, (j)? "+":"-"); break;

	      default: break;
	      }
	   }

	   if (cmd_type==M_OPTION_CONTIN) {
	      if ((1<<(i-1)) & scene->gopt.desired_continents)
		 strcpy(str, "+");
	      else
		 strcpy(str, "-");
	   }
	   if (cmd_type==M_OPTION_CAT) {
	      if ((1<<(i-1)) & scene->gopt.desired_categories)
		 strcpy(str, "+");
	      else
		 strcpy(str, "-");
	   }
	}
        if (cmd_mode==MODE_COLOR) {
	   j = feature_index[cmd_type-1]+i-1;
	   c = (scene->color[j].bool)? '+' : '-';
	   sprintf(str, "%c%s", c, scene->color[j].name);
        }
}

void str2value(ImageLayout * scene, int i, char * str)
{
        int j;
	double f;

	if (cmd_mode==MODE_FILE) {
	   if (cmd_type==M_FILE_DATA) {
	      switch(i) {
		 case 1:  strcpy(scene->rcfile, str); break;
		 case 2:  strcpy(scene->mapfile, str); break;
		 case 3:  strcpy(scene->locfile, str); break;
                 default: break;
	      }
	   }
	   if (cmd_type==M_FILE_PRINTCFG) {
	      switch(i) {
		 case 1:  if (atof(str)>0) ps_width = atof(str); break;
		 case 2:  if (atof(str)>0) ps_lm = atof(str); break;
		 case 3:  if (atof(str)>0) ps_bm = atof(str); break;
		 case 4:  ps_rot = (int)atof(str); 
                          if (ps_rot>0) ps_rot = 90; 
                          if (ps_rot<0) ps_rot = -90; 
                          break;
		 case 5:  ps_grayscale = (*str=='+'); break;
		 case 6:  im_compress = (*str=='+'); break;
		 case 7:  if (*str) {
                             free(ps_viewer); 
			     ps_viewer = strdup(str);
		          }
                          break;
		 case 8:  if (*str) {
                             free(im_viewer); 
			     im_viewer = strdup(str);
		          }
                          break;
		 case 9:  if (*str) {
                             free(print_cmd); 
			     print_cmd = strdup(str);
		          }
                          break;
	         case 10: if (*str) {
                             free(editor_cmd);
                             editor_cmd = strdup(str);
	                  }
                          break;
                 default: break;
	      }
	      return;
	   }
	   if (cmd_type==M_FILE_SAVE && i==1)
	      strcpy(scene->outfile, str);
	}
	if (cmd_mode==MODE_OPTION) {
	   if (cmd_type==M_OPTION_PARAM) {
	      switch(i) {
		 case 1:  scene->projection = set_proj(str); break;
		 case 2:  j = atoi(str); 
		          if (j<16) break;
			  scene->width = j;
			  XResizeWindow(dpy, mainwin, 
                                        scene->width, scene->height);
			  XFlush(dpy);
                          if (textpix) {
                             XFreePixmap(dpy, textpix);
	                     textpix = 0;
	                  }
		          break;
		 case 3:  j = atoi(str); 
		          if (j<16) break;
			  scene->height = j; 
			  XResizeWindow(dpy, mainwin, 
                                        scene->width, scene->height);
			  XFlush(dpy);
                          break;
		 case 4:  f = atof(str); 
		          if (f<0.01) break;
			  scene->gopt.zoom = f;
		          break;
		 case 5:  f = atof(str); 
		          if (f<0.01) break;
                          scene->aspect = f; 
                          break;
		 case 6:  scene->gopt.xrot = atof(str); break;
		 case 7:  scene->gopt.yrot = atof(str); break;
		 case 8:  scene->gopt.zrot = atof(str); break;
		 case 9:  scene->ropt.spacing_lat = atof(str); break;
		 case 10: scene->ropt.spacing_lon = atof(str); break;
		 default: break;
	      }
	   }
	   if (cmd_type==M_OPTION_DISPLAY) {
	      switch(i) {
                 case 1: strncpy(scene->gopt.hierarchy, str, 5);
		         scene->gopt.hierarchy[5] = '\0';
			 check_hierarchy(scene);
                         break;
		 case 2: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &scene->ropt.mark_step[0], &scene->ropt.mark_step[1], &scene->ropt.mark_step[2],
                            &scene->ropt.mark_step[3], &scene->ropt.mark_step[4]))<5)
		             while(j<5) { scene->ropt.mark_step[j]=scene->ropt.mark_step[j-1]; ++j; }
                         break;
		 case 3: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &scene->ropt.name_step[0], &scene->ropt.name_step[1], &scene->ropt.name_step[2],
                            &scene->ropt.name_step[3], &scene->ropt.name_step[4]))<5)
		             while(j<5) { scene->ropt.name_step[j]=scene->ropt.name_step[j-1]; ++j; }
                         break;
		 case 4: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &scene->ropt.city_mark[0], &scene->ropt.city_mark[1], 
                            &scene->ropt.city_mark[2], &scene->ropt.city_mark[3], 
                            &scene->ropt.city_mark[4]))<5)
		         while(j<5) { 
                            scene->ropt.city_mark[j] = scene->ropt.city_mark[j-1]; 
                            ++j; 
                         }
                         break;
		 case 5: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &scene->ropt.peak_size[0], &scene->ropt.peak_size[1], &scene->ropt.peak_size[2],
                            &scene->ropt.peak_size[3], &scene->ropt.peak_size[4]))<5)
		             while(j<5) { scene->ropt.peak_size[j]=scene->ropt.peak_size[j-1]; ++j; }
                         break;
		 case 6: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &scene->ropt.peak_mark[0], &scene->ropt.peak_mark[1], 
                            &scene->ropt.peak_mark[2], &scene->ropt.peak_mark[3], 
                            &scene->ropt.peak_mark[4]))<5)
		         while(j<5) { 
                            scene->ropt.peak_mark[j] = scene->ropt.peak_mark[j-1]; 
                            ++j; 
                         }
                         break;
		 case 7: if ((j=sscanf(str, "%d,%d,%d,%d,%d",
                            &scene->ropt.city_size[0], &scene->ropt.city_size[1], &scene->ropt.city_size[2],
                            &scene->ropt.city_size[3], &scene->ropt.city_size[4]))<5)
		             while(j<5) { scene->ropt.city_size[j]=scene->ropt.city_size[j-1]; ++j; }
                         break;
	         case 8: load_city_filter(scene, str);
			 break;
	         case 9: load_loc_filter(scene, str);
			 break;
	         case 10: if (*str == '+') 
		             scene->ropt.placetext = 1;
		         else
		             scene->ropt.placetext = 0;
                         break;
	         case 11: if (*str == '+') 
		            scene->gopt.transparent = 1;
		         else
		            scene->gopt.transparent = 0;
                         break;
	         case 12: strncpy(language, str, 2);
		         language[2] = '\0';
	                 read_i18n_file(scene);
     	                 fix_features_description(scene);
			 break;
		 default: break;
	      }
	   }
	   if (cmd_type==M_OPTION_FONT) { 
              if (i>=1 && i<=6) {
		 i -= 1;
		 str[255] = '\0';
	         if (!set_font(str, i)) {
	            strcpy(scene->ropt.font[i], str);
		    putPicture(scene);
		 }
	      }
	   }

	   if (cmd_type==M_OPTION_MARKS) {
	      switch(i) {
              case 1: if (!strcmp(str, "-"))
                         scene->color[C_CITY_CIRCLE].bool = OFF;
		      if (!strcmp(str, "+")) {
                         scene->color[C_CITY_CIRCLE].bool = ON;
			 scene->color[C_CITY_TEXT].bool = OFF; 
		      }
		      if (!strcmp(str, "++")) {
                         scene->color[C_CITY_CIRCLE].bool = ON;
			 scene->color[C_CITY_TEXT].bool = ON; 
		      }
                      break;

              case 2: if (!strcmp(str, "-"))
                         scene->color[C_PEAK_TRIANGLE].bool = OFF;
	              if (!strcmp(str, "+")) {
                         scene->color[C_PEAK_TRIANGLE].bool = ON;
                         scene->color[C_PEAK_TEXT].bool = OFF;
	              }
	              if (!strcmp(str, "++")) {
                         scene->color[C_PEAK_TRIANGLE].bool = ON;
                         scene->color[C_PEAK_TEXT].bool = ON;
	              }
                      break;

              case 3: scene->color[C_LOC_LAND].bool = !strcmp(str, "+");
                      break;

              case 4: scene->color[C_LOC_WATER].bool = !strcmp(str, "+");
                      break;

              case 5: scene->color[C_LON_GRID].bool = !strcmp(str, "+");
                      break;

              case 6: scene->color[C_LAT_GRID].bool = !strcmp(str, "+");
                      break;

              case 7: scene->color[C_MAIN_LINES].bool = !strcmp(str, "+");
                      break;

              case 8: scene->color[C_EARTH_CONTOUR].bool = !strcmp(str, "+");
                      break;

              case 9: scene->color[C_BG_SKY].bool = !strcmp(str, "+");
                      break;

              case 10: scene->color[C_FG_STARS].bool = !strcmp(str, "+");
                       break;

	      default: break;
	      }
	   }

	   if (cmd_type==M_OPTION_CONTIN) {
              j = 1<<(i-1);
	      if (!strcmp(str, "+")) scene->gopt.desired_continents |= j;
	      if (!strcmp(str, "-")) scene->gopt.desired_continents &= ~j;
	   }
	   if (cmd_type==M_OPTION_CAT) {
              j = 1<<(i-1);
	      if (!strcmp(str, "+")) scene->gopt.desired_categories |= j;
	      if (!strcmp(str, "-")) scene->gopt.desired_categories &= ~j;
	   }
	}

        if (cmd_mode==MODE_COLOR) {
	   j = feature_index[cmd_type-1]+i-1;
	   if (cmd_type==M_COLOR_WIDGETS || j==C_BG_MAP) {
	      scene->color[j].bool = 1;	      
	      if (*str=='-') *str = '+';
	   }
           get_color(str, scene, j);
	   if (j==C_BG_TEXT || j==C_FG_TEXT) set_bg_fg(scene);
	   setup_change |= COLOR_CHANGE;
        }
}

void drawRectCmd(ImageLayout * scene, char **paramstr, 
                 int n, int position, int level)
{
        int i, l, m, y;
	char str[256];
	char *ptr;

	box_width = 0;
	box_height = fonth[0]+6;
	box_spacing = fonth[0]+((level<0)?5:10);
	box_numitems = n;
	box_x = 6;
	box_y = (position-TOP+1)*(fonth[0]+16)-box_spacing/2+1;
        XSetWindowBackground(dpy, mainwin, scene->color[C_BG_TEXT].pix);
        y = box_y+box_spacing/2-1; 
	if (level<=1) {
           XClearArea(dpy, mainwin, 0, y, 
                   scene->width, (n+1)*box_spacing, False);
           y += (n+1)*box_spacing;
           XPutImage(dpy, mainwin, gc,
                  scene->image, 0, y, 0, y, 
                  scene->image->width, scene->image->height-y);
	}
        for (i=1; i<=n; i++) {
	   ptr = paramstr[i-1];
	   if (!ptr) ptr = "(null)";
	   l = strlen(ptr);
	   XDrawImageString(dpy, mainwin, gc, 
                            box_x+6, box_y+i*box_spacing+fontasc[0]+3,
	                    ptr, l);
           m = XTextWidth(xfont[0], ptr, l) + 12;
	   if (m>box_width) box_width = m;
	}
        for (i=1; i<=n; i++) {
	   if (level>=0)
	      XDrawRectangle(dpy, mainwin, gc, box_x, box_y+i*box_spacing,
		               box_width, box_height);
	   if (line_clicked==i)	   
	      XClearArea(dpy, mainwin, box_x+box_width+1, box_y+i*box_spacing,
                   scene->width, box_height, False);
	   else
	      XSetForeground(dpy, gc, scene->color[C_BG_TEXT].pix);
	   XDrawRectangle(dpy, mainwin, gc, box_x+1, box_y+i*box_spacing+1,
		               box_width-2, box_height-2);
	   XSetForeground(dpy, gc, scene->color[C_FG_TEXT].pix);
	   value2str(scene, i, str, 1);
	   if (*str) {
	      XDrawImageString(dpy, mainwin, gc, 
                               box_x+box_width+5*char_width, 
                               box_y+i*box_spacing+fontasc[0]+3,
	                       str, strlen(str));

	   }
	}
        XSetWindowBackground(dpy, mainwin, scene->color[C_BG_MAP].pix);
	XFlush(dpy);
}

void show_manual()
{
     system("xterm -e man xrmap &");
}

void show_version(ImageLayout * scene)
{
char str[512];

     sprintf(str, msg[VERSION_FORMAT],
	          PACKAGE, VERSION, DATEPKG, " ", COPYRIGHT);
     drawWinString(scene, str, BOTTOM);
}
		
void drawWinCmd(ImageLayout * scene, int level)
{
        int i, j;
	char str[256], name[128];

	if (cmd_mode>MODE_RUN && level==0) {
	   strcpy(str, labels_generalmenu);
	   drawWinString(scene, str, TOP);
	   for (i=1;i<END_LABELS_GENERALMENU-BEGIN_LABELS_GENERALMENU;i++) {
	      j = XTextWidth(xfont[0], str, spacing_generalmenu*i-1);
	      draw_separator(j, TOP);
	   }
	   j = 2*fonth[0]+22;
           XPutImage(dpy, mainwin, gc,
                  scene->image, 0, j, 0, j, 
                  scene->image->width, scene->image->height-j);
	   if (cmd_mode==MODE_MENU && cmd_type==M_MENU_START)
	      drawWinString(scene, msg[MENU_SHORTCUTS], BOTTOM);
	}

        if (cmd_mode==MODE_CMD) {
	   if (level<=1)
	      draw_mark(scene, MODE_MENU, M_MENU_CMD, TOP);
           for (j=0; j<=1; j++)
              drawWinString(scene, msg[LIST_OF_MAIN_COMMANDS+j], TOP+j+1);
	}

        if (cmd_mode==MODE_FILE) {
	   if (level<=1)
	      draw_mark(scene, MODE_MENU, M_MENU_FILE, TOP);
	   strcpy(str, labels_filemenu);
	   drawWinString(scene, str, TOP+1);
	   for (i=1; i<END_LABELS_FILEMENU-BEGIN_LABELS_FILEMENU; i++) {
	      j = XTextWidth(xfont[0], str, spacing_filemenu*i-1);
	      draw_separator(j, TOP+1);
	   }
	   if (level<=1) {
	      draw_mark(scene, MODE_FILE, cmd_type, TOP+1);
	   }
           if (cmd_type==M_FILE_DATA)
              drawRectCmd(scene, datafile_descr, NUM_DATAFILE, TOP+1, level);
	   else
           if (cmd_type==M_FILE_PRINTCFG)
              drawRectCmd(scene, printcfg_descr, NUM_PRINTCFG, TOP+1, level);
	   else
           if (cmd_type==M_FILE_SAVE) {
	      drawRectCmd(scene, save_descr, NUM_SAVE, TOP+1, level);
	      if (line_clicked!=1) {
	         j = scene->image->height-fonth[0]-7;
                 XPutImage(dpy, mainwin, gc,
                    scene->image, 0, j, 0, j, 
                    scene->image->width, scene->image->height-j);
	      }
	   }
	   else {
	      j = 3*fonth[0]+38;
              XPutImage(dpy, mainwin, gc,
                  scene->image, 0, j, 0, j, 
                  scene->image->width, scene->image->height-j);
	   }
	   if (cmd_type==M_FILE_SAVE) {  
	      if (overwrite==1) {
   	         sprintf(str, msg[OUTPUT_FILE_ALREADY_EXISTS], 
                              scene->outfile);
	         drawWinString(scene, str, BOTTOM);  
	      }
	   }
	}

        if (cmd_mode==MODE_MENU) {
	   if (level<=1)
              draw_mark(scene, MODE_MENU, cmd_type, TOP);
           if (cmd_type==M_MENU_SEARCH) {
	      line_clicked = search_by;
              drawWinString(scene, msg[SPECIFY_STRING_TO_SEARCH], TOP+1);
              drawRectCmd(scene, searchby_descr, 
                          END_SEARCHBY_DESCRIPTION-BEGIN_SEARCHBY_DESCRIPTION,
                          TOP+1, level);
	   } else
           if (cmd_type==M_MENU_LIST) {
	      char ** cityname_descr;
              strcpy(str, msg[LIST_OF_MATCHING_LOCATIONS]);
              drawWinString(scene, str, TOP+1);
	      num_listed = scene->height/(fonth[0]+10)-4;
	      if (num_listed>num_found) num_listed = num_found;
	      cityname_descr = (char **)malloc((num_listed+1)*sizeof(char *));
	      line_clicked = -1;
	      for (i=0; i<num_listed; i++) {
		  j = loc_pointer[i+start_list];
		  if (i==0 && start_list>0)
		     strcpy(str, "<<<<<<");
		  else
		  if (i==num_listed-1 && start_list+num_listed<num_found)
		     strcpy(str, ">>>>>>");
		  else {
		     if (j==search_index) line_clicked = i+1;
		     if (locations[j]->obj=='c') {
		        sprintf(str, "%s, %s, %s  (lat = %.3f  lon = %.3f)", 
                           locations[j]->name, 
                       countries[timezones[locations[j]->tz]->country]->name,
                           locations[j]->region,
                           locations[j]->lat, locations[j]->lon);
		     } else 
		     if (locations[j]->obj=='p') {
		        sprintf(str, "%s, %s (lat = %.3f  lon = %.3f)", 
                           locations[j]->name, 
                      countries[timezones[locations[j]->tz]->country]->name,
                           locations[j]->lat, locations[j]->lon);
		     } else 
		     if (locations[j]->obj=='l') {
                        strcpy(name, locations[j]->name);
			simplify_string(name);
		        sprintf(str, "%s, %s (lat = %.3f  lon = %.3f)", name, 
                           msg[BEGIN_CONTINENT_DESCRIPTION+locations[j]->tz],
                           locations[j]->lat, locations[j]->lon);
		     }
		  }
		  cityname_descr[i] = strdup(str);
	      }
              drawRectCmd(scene, cityname_descr, num_listed, TOP+1, level);
	      for (i=0; i<num_listed; i++) free(cityname_descr[i]);
	      free(cityname_descr);
	   }
	}

        if (cmd_mode==MODE_OPTION) {
	   if (level<=1)
	      draw_mark(scene, MODE_MENU, M_MENU_OPTION, TOP);
	   strcpy(str, labels_optionmenu);
	   drawWinString(scene, str, TOP+1);
	   for (i=1;i<END_LABELS_OPTIONMENU-BEGIN_LABELS_OPTIONMENU;i++) {
	       j = XTextWidth(xfont[0], str, spacing_optionmenu*i-1);
	       draw_separator(j, TOP+1);
	   }
	   if (level<=1)
	      draw_mark(scene, cmd_mode, cmd_type, TOP+1);
	   if (cmd_type==M_OPTION_PARAM)
              drawRectCmd(scene, parameter_descr, NUM_PARAM, TOP+1, level);
	   if (cmd_type==M_OPTION_DISPLAY)
              drawRectCmd(scene, display_descr, NUM_DISPLAY, TOP+1, level);
	   if (cmd_type==M_OPTION_FONT)
              drawRectCmd(scene, font_descr, NUM_FONTS, TOP+1, level);
	   if (cmd_type==M_OPTION_MARKS)
              drawRectCmd(scene, marks_descr, NUM_MARKS, TOP+1, level);
           if (cmd_type==M_OPTION_CONTIN)
              drawRectCmd(scene, continent_descr, NUM_CONTIN, TOP+1, level);
           if (cmd_type==M_OPTION_CAT)
              drawRectCmd(scene, category_descr, NUM_CAT, TOP+1, level);
	}

        if (cmd_mode==MODE_COLOR) {
	   if (level<=1)
	      draw_mark(scene, MODE_MENU, M_MENU_COLOR, TOP);
	   strcpy(str, labels_colormenu);             
	   drawWinString(scene, str, TOP+1);
	   for (i=1; i<END_LABELS_COLORMENU-BEGIN_LABELS_COLORMENU; i++) {
	      j = XTextWidth(xfont[0], str, spacing_colormenu*i-1);
	      draw_separator(j, TOP+1);
	   }
	   if (level<=1)
	      draw_mark(scene, cmd_mode, cmd_type, TOP+1);
	   if (cmd_type>M_COLOR_START)
	      drawRectCmd(scene, feature_descr[cmd_type-1], 
                       feature_length[cmd_type-1], TOP+1, level);
	}

        if (cmd_mode==MODE_HELP) {
	   if (level<=1)
	      draw_mark(scene, MODE_MENU, M_MENU_HELP, TOP);
	   strcpy(str, labels_helpmenu);             
	   drawWinString(scene, str, TOP+1);
	   for (i=1; i<END_LABELS_HELPMENU-BEGIN_LABELS_HELPMENU; i++) {
	      j = XTextWidth(xfont[0], str, spacing_helpmenu*i-1);
	      draw_separator(j, TOP+1);
	   }
	   if (cmd_type==M_HELP_SHORTCUTS)
	      drawRectCmd(scene, msg+BEGIN_LIST_SHORTCUTS, 
                      END_LIST_SHORTCUTS-BEGIN_LIST_SHORTCUTS-1, TOP+1, -1);
	   else {
	      j = 2*fonth[0]+22;
	      XPutImage(dpy, mainwin, gc,
                  scene->image, 0, j, 0, j, 
                  scene->image->width, scene->image->height-j);
	   }
	   if (level<=1)
	      draw_mark(scene, cmd_mode, cmd_type, TOP+1);

	   if (cmd_type==M_HELP_ABOUT) 
              show_version(scene);

	   if (cmd_type==M_HELP_MANUAL) {
	      show_manual();
	      cmd_type = M_HELP_START;
	   }
	}

	/* Draw bottom text bar only when needed, i.e. when
	   user has to enter text input */
        if (text_input_mode())
           drawWinString(scene, cmd_string, BOTTOM);
	XFlush(dpy);
}

void image_init (ImageLayout *s) {
	char str[256], compl[128];

	XFlush(dpy);
	usleep(20000);
	if (s->projection==SPHERICAL) 
           *compl = '\0';
	else
	   sprintf(compl, "aspect %.2f, ", s->aspect);

	sprintf(str, "%s %dx%d (%s %s, "
                    "zoom %.2f, %slat %.2f, lon %.2f, inc %.2f) ...",
                msg[BUILDING_MAP], s->width, s->height, 
		msg[PROJECTION_TYPE], projtype[s->projection], 
                s->gopt.zoom, compl, s->gopt.xrot, s->gopt.yrot, s->gopt.zrot);
        drawWinString(s, str, TOP);

	if (s->image) XDestroyImage(s->image);
	if (s->preimage) {
	   free(s->preimage);
	   s->preimage = NULL;
	}

	s->image = XCreateImage(dpy, visual,
                DefaultDepth(dpy, scr), ZPixmap, 0, NULL, 
                s->width, s->height, color_pad, 0);
	s->image->data = (char *) malloc(s->image->bytes_per_line * s->image->height);
}

void printString (void *fo, char * s)
{
    if (fo) {
#ifdef ZLIB
       if (im_compress)
          gzprintf((gzFile *)fo, "%s", s);
       else
	  fprintf((FILE *)fo, "%s", s);
#else
       fprintf((FILE *)fo, "%s", s);
#endif
    } else
       printf("%s", s);
}

void printImage (void *fo, ImageLayout * scene)
{
   char str[1024];
   int i, j, n, k, x, y, test, delta, step;
   Pixel p, bg;
   Pixel * pix;
   unsigned char *u, *v, *w, *code;
   short *inv;

   n = 0;
   for (i=0; i<C_ENDCOLOR; i++)
        if (scene->color[i].bool==ON) ++n;
   u = (unsigned char *) malloc(n*sizeof(unsigned char));
   v = (unsigned char *) malloc(n*sizeof(unsigned char));
   w = (unsigned char *) malloc(n*sizeof(unsigned char));
   pix = (Pixel *)malloc(n*sizeof(Pixel));

   n = 0;
   for (i=0; i<C_ENDCOLOR; i++)
        if (scene->color[i].bool==ON) {
	   u[n] = scene->color[i].r>>8;
	   v[n] = scene->color[i].g>>8;
	   w[n] = scene->color[i].b>>8;
	   if (reverse) {
	      u[n] = ~u[n];
	      v[n] = ~v[n];
	      w[n] = ~w[n];
	   }
	   pix[n] = scene->color[i].pix;
	   ++n;
	}
   bg = scene->color[C_BG_MAP].pix;

   k = 99;
   test = 1;
   while(test) {
   restart:
     k += 2;
     for (i=0; i<n-1; i++)
        for (j=i+1; j<n; j++) {
	   delta = pix[j]-pix[i];
	   if (delta!=0 && delta%k==0) goto restart;
        }
     test = 0;
   }

   inv = (short *)malloc(k*sizeof(short));
   code = (char *)malloc(n*sizeof(char));
   step = 0;
   for (i=0; i<n; i++) {
      j = ((unsigned long int)pix[i])%k;
      if (pix[i] == bg) {
	 code[i] = ' ';
	 step = 1;
      } else
         code[i] = 'A'+i-step;
      inv[j] = i;
   }

   if (img_output==OUT_PPM) {
      sprintf(str, "P6\n#Creator: xrmap (version %s)\n%d %d\n%d\n",
	   VERSION, scene->image->width, scene->image->height, 255);
      printString(fo, str);   
      for (y=0; y<scene->image->height; y++)
      for (x=0; x<scene->image->width; x++) {
	 p = XGetPixel(scene->image, x, y);
	 i = inv[((int)p)%k];
	 if (fo) {
#ifdef ZLIB
	    if (im_compress) {
               gzputc((gzFile *)fo, u[i]);
               gzputc((gzFile *)fo, v[i]);
               gzputc((gzFile *)fo, w[i]);
	    } else {
               fputc(u[i], (FILE *)fo);
               fputc(v[i], (FILE *)fo);
               fputc(w[i], (FILE *)fo);
	    }
#else
            fputc(u[i], (FILE *)fo);
            fputc(v[i], (FILE *)fo);
            fputc(w[i], (FILE *)fo);
#endif
	 } else {
	     fputc(u[i], stdout);
	     fputc(v[i], stdout);
	     fputc(w[i], stdout);
	 }
      }
   }

   if (img_output==OUT_XPM) {
      sprintf(str, "/* XPM */\n/* Creator: xrmap (version %s) */\n"
	          "static char * rmap_xpm[] = {\n"
	          "\"%d %d %d 1\",\n",
	   VERSION, scene->image->width, scene->image->height, n);
      printString(fo, str);   
      for (i=0; i<n; i++) if (code[i]==' ') {
	 sprintf(str, "\"%c\tc #%02X%02X%02X\",\n", code[i], u[i], v[i], w[i]);
         printString(fo, str);
      }
      for (i=0; i<n; i++) if (code[i]!=' ') {
	 sprintf(str, "\"%c\tc #%02X%02X%02X\",\n", code[i], u[i], v[i], w[i]);
         printString(fo, str);
      }
      for (y=0; y<scene->image->height; y++){
	 if (fo) {
#ifdef ZLIB
	    if (im_compress) 
               gzputc((gzFile *)fo, '\"');
	    else
#endif
               fputc('\"', (FILE *)fo);
	 } else
	    fputc('\"', stdout);
         for (x=0; x<scene->image->width; x++) {
	    p = XGetPixel(scene->image, x, y);
	    i = inv[((int)p)%k];
	    if (fo) {
#ifdef ZLIB
	    if (im_compress) 
               gzputc((gzFile *)fo, code[i]);
	    else
#endif
               fputc(code[i], (FILE *)fo);
	    } else
	       fputc(code[i], stdout);
	 }
	 sprintf(str, "\"%c\n", (y==scene->image->height-1)? ';' : ',');
         printString(fo, str);
      }
   }

   if (img_output==OUT_PS) {
      double fact, dx, dy, fx, fy;
      char compl[50];
      * compl = '\0';
      fact = ps_width*PTPERMM/((double)scene->image->width);
      fx = fact*(double)scene->image->width;
      fy = fact*(double)scene->image->height;
      if (ps_rot) {
	 dx = ps_bm * PTPERMM;
	 dy = ps_lm * PTPERMM;
      } else {
         dx = ps_lm * PTPERMM;
         dy = ps_bm * PTPERMM;
      }
      i = (int)(fx+dx+1.0);
      j = (int)(fy+dy+1.0);
      if (ps_rot) {
	 x = i;
	 i = j;
	 j = x;
      }
      if (ps_rot>0) {
	 dy = 0.0;
	 sprintf(compl, "%d 0 translate 90 rotate\n", i);
      }
      if (ps_rot<0) {
	 dx = 0.0;
	 sprintf(compl, "0 %d translate -90 rotate\n", j);
      }
      /*  Real bounding box would be 0 0 i j. Take A4 instead */
      sprintf(str, "%%!PS-Adobe-2.0 EPSF-2.0\n"
                   "%%%%Creator: xrmap (version %s)\n"
                   "%%%%Title: %s\n"
                   "%%%%Pages: 1\n"
                   "%%%%BoundingBox: 0 0 "PAGEFORMAT"\n"
                   "%%%%EndComments\n"
                   "%%EndProlog\n"
                   "%%Page: 1 1\n"
                   "gsave\n%s"
                   "%.3f %.3f translate\n"
                   "%.3f %.3f scale\n"
                   "/line %d string def\n"
                   "%d %d 8 [ %d 0 0 %d 0 %d ]\n"
	           "{currentfile line readhexstring pop}\n"
                   "false 3 colorimage\n",
	   VERSION, scene->outfile, compl,
	   dx, dy, fx, fy,    
           3*scene->image->width,
           scene->image->width, scene->image->height,
           scene->image->width, -scene->image->height,
           scene->image->height);
      printString(fo, str);   
      if (ps_grayscale) {
	 for (i=0; i<n; i++) {
            unsigned char c = (unsigned char)((50*u[i]+32*v[i]+18*w[i])/100);
            u[i] = v[i] = w[i] = c;
	 }
      }
      j = 0;
      for (y=0; y<scene->image->height; y++)
         for (x=0; x<scene->image->width; x++) {
	   p = XGetPixel(scene->image, x, y);
	   i = inv[((int)p)%k];
           sprintf(str, "%02x%02x%02x", u[i], v[i], w[i]);
           printString(fo, str);   
	   j += 6;
	   if (j>70 || x==scene->image->width-1) {
	       j = 0;
#ifdef ZLIB
	       if (im_compress)
                  gzputc((gzFile *)fo, '\n');
	       else
                  fputc('\n', (FILE *)fo);
#else
               fputc('\n', (FILE *)fo);
#endif
	   }
	 }
      sprintf(str,"\n%%\ngrestore\nshowpage\n%%%%Trailer\n");
      printString(fo, str);
      if (runlevel) {
#ifdef ZLIB
         if (im_compress)
            gzflush((gzFile *)fo, Z_FINISH);
	 else
#endif
            fflush((FILE *)fo);
      }
   } 
   free(u); free(v); free(w);
   free(pix); free(inv);
}

void viewOutfile (ImageLayout * scene)
{
   struct stat buf;
   char compl[50], str[512];

   if (!*scene->outfile) return;
   *compl = '\0';
   
   #ifdef ZLIB
   if (im_compress && strcmp(scene->outfile+strlen(scene->outfile)-3,".gz"))
      strcpy(compl, ".gz");
   #endif

   sprintf(str, "%s%s", scene->outfile, compl);
   stat(str, &buf);
   if (!S_ISREG(buf.st_mode)) return;

   get_output_type(scene);
   if (img_output==OUT_PS && *ps_viewer) {
      sprintf(str, "%s %s%s &", ps_viewer, scene->outfile, compl);
      system(str);
   }
   if (img_output!=OUT_PS && *im_viewer) {
      sprintf(str, "%s %s%s &", im_viewer, scene->outfile, compl);
      system(str);
   }
   img_output = OUT_NONE;
}

void createGCs(ImageLayout *s) {
        XGCValues               gcv;
        int                     mask;
        mask = GCForeground | GCBackground | GCFont;
        gcv.background = s->color[C_BG_TEXT].pix;
	gcv.foreground = s->color[C_FG_TEXT].pix;
	gcv.font = xfont[0]->fid;
        gc = XCreateGC(dpy, Root, mask, &gcv);

	create_textpix(s->width);
        gcv.background = black;
        gcv.foreground = white;
        mask = GCForeground | GCBackground;
        textgc = XCreateGC(dpy, textpix, mask, &gcv);
}

void putPixel(XImage *im, int x, int y, Pixel pix)
{
        if (x>=0 && x<im->width && y>=0 && y<im->height)
	         XPutPixel(im, x, y, pix);
}

void drawPoint (ImageLayout *s, Vector a, Pixel color) {
	int x,y;

	x = round2int(a.coord[X]);
	x += s->width/2;
	y = round2int(a.coord[Y]);
	y += s->height/2;

        putPixel(s->image, x, y, color);
}

void drawString(XImage *im, int text_x, int text_y, char * text, 
		int which, Pixel pix)
{
        int i, j, l, textw;
	char u=0, test;
	char *ptr;
        XImage *xim;
      
	if (!text || !*text) return;
        XSetFont(dpy, textgc, xfont[which]->fid);
	ptr = text;

      redo:
	text = ptr;
	l = strlen(text);
	ptr = NULL;
	for (i=0; i<l-1; i++) if (text[i]=='\\' && text[i+1]=='n') {
	    text[i] = '\0';
	    ptr = text+i+2;
	    break;
	}
        XDrawImageString(dpy, textpix, textgc, 0, fontasc[which], 
                         text, strlen(text));
	textw = XTextWidth(xfont[which], text, strlen(text));
	if (textw>im->width) textw=im->width;
	
        xim = XGetImage(dpy, textpix, 0, 0, textw, fonth[which], 1, XYPixmap);
        if (!xim) return;
        test = (bigendian)? 128 : 1;
        for (j=0; j<fonth[which]; ++j) {
            for (i=0; i<textw; ++i) {
                if ((i&7) == 0) u = xim->data[j*xim->bytes_per_line+i/8];
                if (u&test) 
                   putPixel(im, text_x+i+1, text_y-fontasc[which]+j, pix);
                u = (bigendian)? u<<1 : u>>1;
            }
        }
	if (ptr) {
	   *(ptr-2) = '\\';
	   if (*ptr) {
	      text_y += fonth[which];
	      goto redo;
	   }
	}
        XDestroyImage(xim);
}

void drawSegment (ImageLayout *s, Vector a, Vector b, Pixel color) {
  	int x, y, xp, yp, dx, dy,
            m, sum, min, max, addxmin, addxmax, addymin, addymax;
		
        /* lets try it without perspective */
        x = round2int(a.coord[X]);
        xp = round2int(b.coord[X]);
	dx = xp-x;
	min = s->width/2;
        x += min;

        y = round2int(a.coord[Y]);
        yp = round2int(b.coord[Y]);
        dy = yp-y;
	y += s->height/2;

        /* draw the line */
        if (dx>0)
	   addxmin = 1;
	else {
	   dx = -dx;
           addxmin = -1;
	}
	if (dx>escape_length) return;

        if (dy>0)
	   addymin = 1;
	else {
	   dy = -dy;
           addymin = -1;
	}
	if (dx>dy) {
	  max = dx;
	  min = dy;
	  addxmax = addxmin;
	  addymax = 0;
	} else {
	  max = dy;
	  min = dx;
	  addxmax = 0;
	  addymax = addymin;
	}
        sum = (max/2) - max;
	for (m=0; m<max; m++) {
	   sum = sum+min;
	   if (sum>=max/2) {
              sum -= max;
              x += addxmin;
              y += addymin;
           } else {
              x += addxmax;
              y += addymax;
           }
	   putPixel(s->image, x, y, color);
	}
}   

Window makeWindow (int width, int height) {
        Window w;
        black = BlackPixel(dpy, scr);
        white = WhitePixel(dpy, scr);

	w = XCreateSimpleWindow(dpy, Root,
                      10, 10,
                      width, height, 0,
                      white, black);
        XSetWindowColormap(dpy, w, cmap);

        XSelectInput(dpy, w, 
            ButtonPressMask | ButtonReleaseMask | 
            PointerMotionMask | EnterWindowMask | LeaveWindowMask | 
            KeyPressMask | KeyReleaseMask | 
            ExposureMask | StructureNotifyMask);

        wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", False);
        wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
        XSetWMProtocols(dpy, w, &wm_delete_window, 1);
        return w;
}

static Vector project (ImageLayout * s, Vector a) {
	Vector v;
        Matrix *b;
	int i;
	
        b = &s->rotation;
	if (s->projection==SPHERICAL) {
           v.coord[0] = a.coord[0]*b->comp[0] + a.coord[1]*b->comp[3] + a.coord[2]*b->comp[6];
	   v.coord[1] = a.coord[0]*b->comp[1] + a.coord[1]*b->comp[4] + a.coord[2]*b->comp[7];
	   v.coord[2] = a.coord[0]*b->comp[2] + a.coord[1]*b->comp[5] + a.coord[2]*b->comp[8];
	} else {
           v.coord[0] = a.coord[0] - b->comp[0];
	   if (v.coord[0]>earth_radius) {
	      i = (int)(v.coord[0]/earth_radius);
	      i = (i+1)&(-2);
              v.coord[0] -= ((double)i)*earth_radius;
	   } else
	   if (v.coord[0]<-earth_radius) {
	      i = (int)(-v.coord[0]/earth_radius);
	      i = (i+1)&(-2);
              v.coord[0] += ((double)i)*earth_radius;
	   }
           v.coord[0] *= b->comp[2];
           v.coord[1] = (a.coord[1] - b->comp[1]) * b->comp[2];
           v.coord[2] = a.coord[2] * b->comp[2];	  
	}
	return v;
}

void elliptic (double *x)
{
int sign=0;
double t;
    if (*x<0) {
       t = -(*x);
       sign = 1;
    } else
       t = *x;
    if (t>=1.0) return; 
    if (t==0) return;
    /* with STRETCHING=4/3, this closely imitates the Mollweide projection,
       while being much simpler to calculate */
    *x = 1.0 - exp(STRETCHING*log(1-t));
    if (sign) *x = -(*x);
}

void mollweide (double *x)
{
int k, sign=0;
double d, t;

    if (*x<0) {
       t = -(*x);
       sign = 1;
    } else
       t = *x;
    if (t>=1.0) return; 
    if (t==0) return;
    d = PI_OVER_2 * sin(PI_OVER_2 * t);
    if (t<=0.625) 
       *x = PI_SQUARE_OVER_8 * t * (1.0 - 0.09*t);
    else
       *x = 1.0 - exp((PI_SQUARE_OVER_8+MOLLWEIDE_COEFF*t)*log(1-t));
    k = 0;
 iter:
    ++k;
    t = *x;
    *x = 0.5 * (t + (d-asin(t))/sqrt(1-t*t));
    if (fabs(*x-t)>1E-9 && k<=10) goto iter;
    if (sign) *x = - (*x);
}

void operate(ImageLayout * s, double *y)
{
        if (s->projection==CYLINDRICAL)
           *y = sin(PI_OVER_2 * (*y));
	else
        if (s->projection==MERCATOR) {
	   if (*y>0.0) {
              if (*y>0.99) 
                 *y = 3.10;
	      else
                 *y = TWO_OVER_PI*log(tan(PI_OVER_4*(*y+1.0)));
	   } else {
              if (*y<-0.99) 
                 *y = -3.10;
	      else
                 *y = -TWO_OVER_PI*log(tan(PI_OVER_4*(-*y+1.0)));
	   }
	} 
        else
	if (s->projection==MILLER)
           *y = FIVE_OVER_2PI*log(tan(PI_OVER_4*(0.8*(*y)+1.0)));
	else
	if (s->projection==ELLIPTIC)
	   elliptic(y);
	else
	if (s->projection==MOLLWEIDE)
	   mollweide(y);
}

void inverse_operate(ImageLayout * s, double *y)
{
double d;
int sign;
       if (s->projection==CYLINDRICAL) {
	  if (fabs(*y)>=1.0) return;
          *y = TWO_OVER_PI * asin(*y);
       } else
       if (s->projection==MERCATOR)
	  *y = atan(exp((*y)*PI_OVER_2))/PI_OVER_4-1.0;
       else
       if (s->projection==MILLER)
	  *y = 1.25*(atan(exp(*y/FIVE_OVER_2PI))/PI_OVER_4-1.0);
       else
       if (s->projection==ELLIPTIC) {
	  if (*y==0) return;
	  if (*y>0) 
	     sign = 0;
          else {
	     *y = -(*y);
	     sign = 1;
	  }
	  if (*y<1)
             *y = 1.0 - exp(log(1-(*y))/STRETCHING);
	  if (sign) *y = -(*y);
       }
       else
       if (s->projection==MOLLWEIDE) {
	  if (fabs(*y)>=1.0) return;
	  d = 2.0 * asin(*y);
	  *y = asin((d + sin(d))*ONE_OVER_PI)*TWO_OVER_PI;
       }
}

static Vector spher2cart (ImageLayout * scene, SpherePt s) {
        Vector v;
	double latrad, lonrad, y;

	if (scene->projection==SPHERICAL) {
	   latrad = s.sph[LATITUDE] * CONV;
	   lonrad = s.sph[LONGITUDE] * CONV;
	   y = earth_radius * cos(latrad);

           v.coord[X] = y * sin(lonrad);
           v.coord[Y] = earth_radius * sin(latrad);
           v.coord[Z] = y * cos(lonrad);
           return v;
	} else {
	   y = -s.sph[LATITUDE] / 90.0;
	   if (scene->projection==SINUSOIDAL) {
	      lonrad = cos(PI_OVER_2*y);
	      if (lonrad>0.1)
	         escape_length = (int)(scene->width * lonrad * 0.5);
	      v.coord[X] = s.sph[LONGITUDE] - scene->gopt.yrot ;
	      while (v.coord[X]<-180.0) v.coord[X] += 360.0;
	      while (v.coord[X]>180.0) v.coord[X] -= 360.0;
              v.coord[X] = v.coord[X]*lonrad + scene->gopt.yrot;
	      while (v.coord[X]<-180.0) v.coord[X] += 360.0;
	      while (v.coord[X]>180.0) v.coord[X] -= 360.0;
              v.coord[X] *= earth_radius/180.0;
	   }
	   else
	   if (scene->projection>=ELLIPTIC) {
              operate(scene, &y);
	      lonrad = 1.0-y*y;
	      if (lonrad<0.0)
		 lonrad = 0.0;
	      else
	         lonrad = sqrt(lonrad);
	      if (lonrad>0.1)
	         escape_length = (int)(scene->width * lonrad * 0.5);
	      v.coord[X] = s.sph[LONGITUDE] - scene->gopt.yrot ;
	      while (v.coord[X]<-180.0) v.coord[X]+= 360.0;
	      while (v.coord[X]>180.0) v.coord[X]-= 360.0;
              v.coord[X] = v.coord[X]*lonrad + scene->gopt.yrot;
	      while (v.coord[X]<-180.0) v.coord[X]+= 360.0;
	      while (v.coord[X]>180.0) v.coord[X]-= 360.0;
              v.coord[X] *= earth_radius/180.0;
	   } else {
              operate(scene, &y);
	      v.coord[X] = earth_radius * s.sph[LONGITUDE] / 180.0;
	   }
	   v.coord[Y] = 0.5 * earth_radius * y * scene->aspect ;
           v.coord[Z] = 1.0;
	   return v;
	}
}

static int inverse_coord(ImageLayout * s, 
                     int i, int j, double *x, double *y)
{
        double fact;
        Vector a, v;
	Matrix *b;
	if (i<0 || j<0 || i>=s->width || j>=s->height) return 0;
	fact = earth_radius*s->gopt.zoom;
        a.coord[X] = ((double)(i-s->width/2))/fact;
	if (s->projection==SPHERICAL) {
           a.coord[Y] = ((double)(j-s->height/2))/fact;
	   a.coord[Z] = 1-(a.coord[X])*(a.coord[X])-(a.coord[Y])*(a.coord[Y]);
	   if (a.coord[Z]<0) return 0;
	   a.coord[Z] = sqrt(a.coord[Z]);
	   b = &s->rotation;
           v.coord[0] = a.coord[0]*b->comp[0] + a.coord[1]*b->comp[1] + a.coord[2]*b->comp[2];
	   v.coord[1] = a.coord[0]*b->comp[3] + a.coord[1]*b->comp[4] + a.coord[2]*b->comp[5];
	   v.coord[2] = a.coord[0]*b->comp[6] + a.coord[1]*b->comp[7] + a.coord[2]*b->comp[8];

	   v.coord[0] /= s->gopt.zoom;
	   v.coord[1] /= s->gopt.zoom;
	   v.coord[2] /= s->gopt.zoom;
	   *x = atan2(v.coord[1], sqrt(v.coord[0]*v.coord[0]+v.coord[2]*v.coord[2]))
                     / CONV;
	   *y = atan2(v.coord[0] , v.coord[2]) / CONV;
	   return 1;
	} else {
           a.coord[Y] = 2.0 * (((double)(j-s->height/2))/(fact*s->aspect)
                       + s->rotation.comp[1]/(earth_radius * s->aspect) );

	   if (s->projection==SINUSOIDAL) {
	      fact = cos(PI_OVER_2*a.coord[Y]);
	      if (fabs(a.coord[X])>fact) return 0;
	      *y = a.coord[X]/fact;
	   } else
	   if (s->projection>=ELLIPTIC) {
	      fact = sqrt(1-a.coord[Y]*a.coord[Y]);
	      if (fabs(a.coord[X])>fact) return 0;
	      *y = a.coord[X]/fact;
	   } else {
              *y = a.coord[X];
	      if (fabs(*y)>1.0) return 0;
	   }
 
	   inverse_operate(s, &a.coord[Y]);
	   if (fabs(a.coord[Y])>1.0) return 0;

	   *x = -a.coord[Y] * 90.0;
           *y = (*y)*180.0 + s->gopt.yrot;
           while (*y>180.0) *y -= 360.0;
	   while (*y<=-180.0) *y += 360.0;

	   return 1;
	}
}

static int str2toint(unsigned char *str)
{
   if (str[1]<=127)
      return (int) (str[1]<<8)|str[0];
   else
      return (int) (0xffff<<16)|(str[1]<<8)|str[0];
}

static int str3toint(unsigned char *str)
{
   if (str[2]<=127)
      return (int) (str[2]<<16)|(str[1]<<8)|str[0];
   else
      return (int) (0xff<<24)|(str[2]<<16)|(str[1]<<8)|str[0];
}

static int str4toint(unsigned char *str)
{
   return (int) (str[3]<<24)|(str[2]<<16)|(str[1]<<8)|str[0];
}

static int segment_in_scene (ImageLayout *s, Segment_index *ind){
	SpherePt min, max;
	Vector pt1, pt2;
	double w, h;

	min.sph[LONGITUDE] = str3toint(ind->minlong)/3600.0;
	min.sph[LATITUDE] = str3toint(ind->minlat)/3600.0;

	max.sph[LONGITUDE] = str3toint(ind->maxlong)/3600.0;
	max.sph[LATITUDE] = str3toint(ind->maxlat)/3600.0;

	pt1 = project(s, spher2cart(s, max));
	pt2 = project(s, spher2cart(s, min));

	/* check to see if it is z-clipped */
	if (s->gopt.transparent == 0 && (pt1.coord[Z] < 0 || pt2.coord[Z] < 0)) {
		return 0; 
	}
		
	/* Yeah, this is ugly, but my head hurts right now. I've got the flu
	   I'll try to figure it out later */

	w = s->width/2;
	h = s->height/2;

	if ( (pt1.coord[X] > w && pt2.coord[X] > w) ||
	     (pt1.coord[Y] > h && pt2.coord[Y] > h) || 
	     (pt1.coord[X] < -w && pt2.coord[X] < -w) ||
	     (pt1.coord[Y] < -h && pt2.coord[Y] < -h) ) {
		return 0;
	}
	
	/* See if the whole segment can fit into one point, 
           if so, then just draw the point! */
	if ( (int)pt1.coord[X] == (int)pt2.coord[X] && 
             (int)pt1.coord[Y] == (int)pt2.coord[Y] ) {
		return POINT_ONLY;
	}

	return 1;
}


void drawCircles(ImageLayout *s) {
        int i, j, b;
	double x, y; 
	SpherePt spt1, spt2;
	Vector pt1, pt2;	
	double step; /* small step */
	double main_circles_lat[5] = { 0.0, 23.435, 66.565 };

	step = sqrt(4.0/(1+s->gopt.zoom));
	b = (s->projection==SPHERICAL);

	if (s->color[C_LON_GRID].bool == ON) 
	for (i=0; i<=1; i++)
	for (x=i*s->ropt.spacing_lon; x<180.001-i*b*(s->ropt.spacing_lon/5); 
             x+=s->ropt.spacing_lon) {
	        j = 0;
		spt1.sph[LONGITUDE] = (i==0)? x : -x;
		spt2.sph[LONGITUDE] = spt1.sph[LONGITUDE];
		y = -89.999999;
		while (y<90.0) {
		    if (j==0) {
		       spt1.sph[LATITUDE]  = y;
		       pt1 = project(s, spher2cart(s, spt1));
		       ++j;
		    } else
		       pt1 = pt2;
		    y += step;
		    if (pt1.coord[Z] < 0) {
		       j = 0;
		       continue;
		    }
		    if (y>=90.0)
		       spt2.sph[LATITUDE] = 89.999999;
		    else
		       spt2.sph[LATITUDE] = y;
		    pt2 = project(s, spher2cart(s, spt2));
	            escape_length = LARGE_INTEGER;
		    drawSegment(s, pt1, pt2, s->color[C_LAT_GRID].pix);
		}
	}

	for (i=0; i<=6; i++)
	for (x=i*s->ropt.spacing_lat; x<90.0-s->ropt.spacing_lat/5 || i>=2; 
             x+=s->ropt.spacing_lat) {
                Pixel pix;
	        if (i<=1 && s->color[C_LAT_GRID].bool != ON) break;
	        if (i>1 && s->color[C_MAIN_LINES].bool != ON) break;
	        if (i>=2) {
                   x = main_circles_lat[(i-1)>>1];
		   pix = s->color[C_MAIN_LINES].pix;
		} else
		   pix = s->color[C_LON_GRID].pix;
	        j = 0;
		spt1.sph[LATITUDE]  = (i&1)? x : -x;
		spt2.sph[LATITUDE]  = spt1.sph[LATITUDE];
		y = -180;
		while (y<180.0) {
		    if (j==0) {
		       spt1.sph[LONGITUDE] = y;
		       pt1 = project(s, spher2cart(s, spt1));
		       ++j;
		    } else
		       pt1 = pt2;
		    y += step;
		    if (pt1.coord[Z] < 0) {
		       j = 0;
		       continue;
		    }
		    if (y>=180.0)
		       spt2.sph[LONGITUDE] = 180.0;
		    else
                       spt2.sph[LONGITUDE] = y;
		    pt2 = project(s, spher2cart(s, spt2));
          	    escape_length = LARGE_INTEGER;
		    drawSegment(s, pt1, pt2, pix);
		}
		if (i>=2) break;
	}

	if (s->color[C_EARTH_CONTOUR].bool == ON &&
	    s->projection==SPHERICAL) {
	   y = earth_radius * s->gopt.zoom;
	   pt2.coord[X] = y;
	   pt2.coord[Y] = 0;
	   for (x=step; x<360.1-step; x+=step) {
	       pt1 = pt2;
	       pt2.coord[X] = y * cos(x*CONV);
	       pt2.coord[Y] = y * sin(x*CONV);
	       drawSegment(s, pt1, pt2, s->color[C_EARTH_CONTOUR].pix);
	   }
	}
}

int size_of_city(ImageLayout *s, int n)
{
    int i;    
    for (i=0; i<5; i++)
        if (n>s->ropt.city_size[i]) return i;
    return 0;
}

int size_of_peak(ImageLayout *s, int n)
{
    int i;    
    for (i=0; i<5; i++)
        if (n>s->ropt.peak_size[i]) return i;
    return 0;
}

void drawLocation (ImageLayout *s, Location * loc) {
	Vector pt;
        SpherePt spt;
	int x, y, i, j, d, u, ind=0, size=0, bool, booltext=0;
	int text_x = 0, text_y = 0;
	Pixel pixel = black;
        unsigned short * bits;
	char name[128];
	char *ptr = NULL, *seatype = NULL;

	i = loc->tz;
	if (i<0) return;
	if (loc->obj=='c' || loc->obj=='p')
	   i = 0x00000001<<timezones[i]->continent;
	if (loc->obj=='l')
	   i = 0x00000001<<(i-1);
	if (!(s->gopt.desired_continents & i)) return;

	if (loc->obj=='c') {
	  if (s->color[C_CITY_CIRCLE].bool!=ON && 
              s->color[C_CITY_TEXT].bool!=ON)
              return;
	}
	if (loc->obj=='p') {
	  if (s->color[C_PEAK_TRIANGLE].bool!=ON && 
              s->color[C_PEAK_TEXT].bool!=ON)
              return;
	}
	if (loc->obj=='l') {
	  seatype = index("BKOSTV", loc->status);
	  if (s->color[C_LOC_LAND].bool!=ON && !seatype) return;
	  if (s->color[C_LOC_WATER].bool!=ON && seatype) return;
	}

	spt.sph[LONGITUDE] = loc->lon;
	spt.sph[LATITUDE] = loc->lat;

	/* determine the point, post transformation */
	pt = project(s, spher2cart(s, spt));

	if (pt.coord[Z] < 0) 
		return;

	x = pt.coord[X] + s->width/2;
	y = pt.coord[Y] + s->height/2;

	/* verify that the point is on the map, return if not */
	if (x > s->width || x < 0 || y > s->height || y < 0)
		return;
	
	/* draw a mark around the point in question if it's a city */
	bool = 0;
	if (loc->obj=='c') {
	   if (!index(s->ropt.city_filter, loc->status)) return;
	   size = size_of_city(s, loc->population);
	   ind = s->ropt.city_mark[size]-1;
	   if (ind<0) return;
	   if (s->ropt.mark_step[size] > 
              (int)(2*s->gopt.zoom*earth_radius)) return;
	   bool = s->color[C_CITY_CIRCLE].bool;
	   booltext = s->color[C_CITY_TEXT].bool;
	   pixel = s->color[C_CITY_CIRCLE].pix;
        } else
        if (loc->obj=='p') {
	   bool = s->color[C_PEAK_TRIANGLE].bool;
	   booltext = s->color[C_PEAK_TEXT].bool;
	   pixel = s->color[C_PEAK_TRIANGLE].pix;
	   size = size_of_peak(s, loc->alt);
	   ind = s->ropt.peak_mark[size]+4;
	   if (ind<=4) return;
	   if (s->ropt.mark_step[size] > 
              (int)(2*s->gopt.zoom*earth_radius)) return;
        }
        if (bool) {
	   loc->drawn = '+';
	   bits = spot_bits[ind];
	   d = bits[0]/2;
	   for (j=0; j<bits[0]; j++) {
              if (y-d+j >= s->height) break;
              u = bits[j+1];
              for (i=0; i<bits[0]; i++) {
                 if (u&1) putPixel(s->image, x-d+i, y-d+j, pixel);
                 u = u>>1;
	      }
	   }
	   /* determine the text location relative to the circle */
	   d = (d*7)/10+1;
	   ind = size;
	   if (!booltext) return;
	   if (s->ropt.name_step[size] > 
              (int)(2*s->gopt.zoom*earth_radius)) return;
	   if ((ptr=index(loc->name,'('))) {
	      strcpy(name, ptr+1);
	      ptr = index(name,')');
	      if (ptr) *ptr = '\0';
	   } else
	      strcpy(name, loc->name);
	   ptr = name;
	   text_x = (s->ropt.placetext && x > s->width/2) ? 
	     (x - XTextWidth(xfont[ind], name, strlen(name)) - d) :
	     (x + d);
	   text_y = (s->ropt.placetext && y < s->height/2) ? 
             (y+fonth[ind]) : (y - d);
	   if (loc->obj=='c')
	      pixel = s->color[C_CITY_TEXT].pix;
	   else
	   if (loc->obj=='p')
	      pixel = s->color[C_PEAK_TEXT].pix;
	} 

	if (loc->obj=='l') {
	   ind = loc->population;
	   if (!ind) return;
	   if (s->ropt.name_step[ind-1] > 
              (int)(2*s->gopt.zoom*earth_radius)) return;
           if (!index(s->ropt.loc_filter, loc->status)) return;
	   ptr = loc->name;
	   text_x = x;
	   text_y = y;
	   if (loc->rank) {
              char *ptr1, *ptr2;
	      int i1;
	      i = 0;
	      if (loc->rank<=3)
		 j = fontasc[ind]-fonth[ind]-1;
	      else
	      if (loc->rank>=5)
		 j = fontasc[ind]+1;
	      else
		 j = 0;
	      ptr1 = ptr;
	      while(ptr1) {
		 ptr2 = strstr(ptr1, "\\n");
		 if (!ptr2) ptr2 = ptr1+strlen(ptr1);
                 i1 = XTextWidth(xfont[ind], ptr1, (int)(ptr2-ptr1));
		 if (i1>i) i = i1;
		 if (*ptr2=='\\') {
                    ptr1 = ptr2+2;
		    if (loc->rank<=3) j -= fonth[ind];
		 } else
		    ptr1 = NULL;
	      }
	      if (loc->rank>=3 && loc->rank<=5) ++i;
	      else
	      if (loc->rank==2 || loc->rank==6) i = i/2;
	      else
		 i = -1;
	      text_x -= i; 
              text_y += j;
	   }
	   if (seatype)
	      pixel = s->color[C_LOC_WATER].pix;
           else
	      pixel = s->color[C_LOC_LAND].pix;
	}

	/* add text */
	if (!textpix) create_textpix(s->width);
	drawString(s->image, text_x, text_y, ptr, ind, pixel);
}

/* This function will generate the proper transformation 
 * matrix (rotation and scaling) and copy the values 
 * to the ImageLayout struct
 */

void computeMatrix(ImageLayout *s) {
     double y, cosx, cosy, cosz, sinx, siny, sinz;
     int i;

     if (s->projection==SPHERICAL) {
        /* generate the proper rotation & scale matrix */
        cosx = cos(s->gopt.xrot * CONV);
        cosy = cos(s->gopt.yrot * CONV);
        cosz = cos(s->gopt.zrot * CONV);

        sinx = sin(s->gopt.xrot * CONV);
        siny = sin(s->gopt.yrot * CONV);
        sinz = sin(s->gopt.zrot * CONV);

	s->rotation.comp[0] = s->gopt.zoom*(cosy*cosz + sinx*siny*sinz);
	s->rotation.comp[1] = s->gopt.zoom*(-cosy*sinz + sinx*siny*cosz);
	s->rotation.comp[2] = s->gopt.zoom*(cosx*siny);

	s->rotation.comp[3] = s->gopt.zoom*(-cosx*sinz);
	s->rotation.comp[4] = s->gopt.zoom*(-cosx*cosz);
	s->rotation.comp[5] = s->gopt.zoom*(sinx);

	s->rotation.comp[6] = s->gopt.zoom*(-siny*cosz + sinx*cosy*sinz);
	s->rotation.comp[7] = s->gopt.zoom*(siny*sinz + sinx*cosy*cosz);
	s->rotation.comp[8] = s->gopt.zoom*(cosx*cosy);

     } else {
        s->rotation.comp[0] = s->gopt.yrot * earth_radius / 180.0;
        y = s->gopt.xrot / 90.0;
	operate(s, &y);
        s->rotation.comp[1] = - 0.5 * earth_radius * y * s->aspect;
        s->rotation.comp[2] = s->gopt.zoom;
     }
     for (i=0; i<s->num_positions; i++) {
         if (s->position[i].xrot == s->gopt.xrot &&
	     s->position[i].yrot == s->gopt.yrot &&
	     s->position[i].zrot == s->gopt.zrot) return;
     }
     ++s->num_positions;
     i = s->ind_position = s->num_positions-1;
     s->position = (XYZPosition *)realloc(s->position, 
                       s->num_positions*sizeof(XYZPosition));
     s->position[i].xrot = s->gopt.xrot;
     s->position[i].yrot = s->gopt.yrot;
     s->position[i].zrot = s->gopt.zrot;
     s->position[i].loc = -1;
}
	
void openOutfile(ImageLayout * scene, void **fo)
{
        int i;
	struct stat buf;
	char str[256], strgz[256];

	if (!*scene->outfile) {
	   *fo = NULL;
	   overwrite = -1;
	   return;
	}

        i = strlen(scene->outfile)-3;
        if (strcmp(scene->outfile+i, ".gz")) {
	   sprintf(str,"%s", scene->outfile);
	   sprintf(strgz,"%s.gz", scene->outfile);
        } else {
	   strcpy(strgz, scene->outfile);
           strcpy(str, strgz);
	   str[i-3] = '\0';
        }
        if (overwrite==0) {
           stat(str, &buf);
	   if (S_ISREG(buf.st_mode)) overwrite = 1;
           stat(strgz, &buf);
           if (S_ISREG(buf.st_mode)) overwrite = 1;
	   if (overwrite==1) return;
	}
 
        #ifdef ZLIB
	if (im_compress) {
	   *fo = (void *)gzopen(strgz, "wb6");
	   strcpy(str, strgz);
	} else
	#endif
	   *fo = (void *)fopen(str, "w");

	if (!*fo) {
	   fprintf(stderr, msg[OPENING_FAILED], str);
	   fprintf(stderr, "\n");
	}
	overwrite = -1;
}

void closeOutfile(void *fo)
{
    if (fo) {
#ifdef ZLIB
       if (im_compress) 
          gzclose((gzFile *)fo);
       else
#endif
          fclose((FILE *)fo);
    }
}

void set_bg_sky(ImageLayout *scene)
{
    int c, i, j, i1_0=0, i2_0=0, i1, i2, j1=0, j2=0, k, l;
    Pixel bg, sky, star;
    /* pixels for bigger stars */
    int dx[5] = { 0, 0, 0, 1, -1};
    int dy[5] = { 0, 1, -1, 0, 0};
    double R, val, y;

    bg = scene->color[C_BG_MAP].pix;
    if (scene->color[C_BG_SKY].bool) {
       sky = scene->color[C_BG_SKY].pix;
       if (scene->color[C_FG_STARS].bool)
          star = scene->color[C_FG_STARS].pix;
       else
	  star = sky;
    } else
       star = sky = bg;

    c = random() % 191; 
    R = earth_radius*scene->gopt.zoom;
    i = (int)R;
    if (scene->projection==SPHERICAL) {
       j = i;
       R = R*R;
       j1 = scene->height/2 - j;
       j2 = scene->height/2 + j;
    } else {  
       i1_0 = scene->width/2 - i;
       i2_0 = scene->width/2 + i;
       if (scene->projection==RECTANGULAR ||
           scene->projection==SINUSOIDAL ||
           scene->projection>=ELLIPTIC) {
	  y = scene->gopt.xrot/90.0;
	  operate(scene, &y);
	  y *= 0.5;
          j1 = scene->height/2-(int)(R*scene->aspect*(0.5-y));
          j2 = scene->height/2+(int)(R*scene->aspect*(0.5+y));
       } else
       if (scene->projection==CYLINDRICAL) {
	  y = 0.5*sin(scene->gopt.xrot * CONV);
          j1 = scene->height/2-(int)(R*scene->aspect*(0.5-y));
          j2 = scene->height/2+(int)(R*scene->aspect*(0.5+y));
       } else
       if (scene->projection==MERCATOR) {
	  j1 = 0;
	  j2 = scene->height-1;
       } else
       if (scene->projection==MILLER) {
          y = scene->gopt.xrot/90.0;
	  operate(scene, &y);
	  y *= 0.5;
	  j1 = scene->height/2 - (int)(R*scene->aspect*(MAX_MILLER-y));
	  j2 = scene->height/2 + (int)(R*scene->aspect*(MAX_MILLER+y));
       }
    }

    for (j=0; j<scene->height; j++) {
       if (j<j1 || j>j2) 
	  i1 = i2 = scene->width;
       else {
          if (scene->projection==SPHERICAL) {
	     i = j - scene->height/2;
	     val = R - ((double)i*(double)i);
	     if (val<=0) 
                val = 0.0;
	     else
	        val = sqrt(val);
	     i = (int)val;
             i1 = scene->width/2 - i;
             i2 = scene->width/2 + i;
	  }
          else 
          if (scene->projection==SINUSOIDAL) {
	     k = j-scene->height/2-(int)(R*scene->aspect*y);
	     k = (int)(R*cos((M_PI*k)/(R*scene->aspect)));
	     i1 = scene->width/2 - k;
	     i2 = scene->width/2 + k;
	  } 
          else 
          if (scene->projection>=ELLIPTIC) {
	     k = j-scene->height/2-(int)(R*scene->aspect*y);
	     val = (2.0*(double)k)/(R*scene->aspect);
             val = 1-val*val;
	     if (val<=0) 
                val = 0;
	     else
	        val = sqrt(val);
	     k = (int)(R*val);
	     i1 = scene->width/2 - k;
	     i2 = scene->width/2 + k;
	  } 
          else {
             i1 = i1_0;
             i2 = i2_0;
	  }
       }
       for (i=0; i<scene->width; i++) {
	  if (i>=i1 && i<=i2)
	     XPutPixel(scene->image, i, j, bg);
	  else {
	     if ((i<i1-2 || i>i2+2) && (i*i+j)%971 == (j*j*j+i)%593 + c) {
		if (((9*i+j*j)%7) == 0) l=4; else l=0;
		for (k=0; k<=l; k++)
	           putPixel(scene->image, i+dx[k], j+dy[k], star);
	        } else
	           XPutPixel(scene->image, i, j, sky);
	  }
       }
    }
}

Bool evpred(d, e, a)
Display *              d;
XEvent *               e;
XPointer               a;
{
   if (e->type==Expose) {
      if (e->xexpose.window==mainwin) exposed |= MAIN_WINDOW;
      if (e->xexpose.window==datawin) exposed |= DATA_WINDOW;
   }
   return (True);
}

void discard_events(Window win) 
{
   XEvent ev;
   while (XCheckIfEvent(dpy, &ev, evpred, (XPointer)&win))
      if (ev.type==ButtonRelease || ev.type==LeaveNotify) but_pressed = 0;
   if (win==mainwin) exposed &= ~MAIN_WINDOW;
   if (win==datawin) exposed &= ~DATA_WINDOW;
}

int essential_change(ImageLayout *scene1, ImageLayout *scene2)
{
    return 
      memcmp(scene1, scene2, 
        3*sizeof(int)+sizeof(double)+sizeof(Matrix)
	+sizeof(GeomOptions)) ||
      memcmp(scene1->color, scene2->color, C_BG_MAP*sizeof(Colordata));
}

void generate (ImageLayout *scene) {
        /* misc variables */
	SpherePt sp;
	Vector pt1, pt2;
	int numseg, u, i, j, k, l, m, ret, lon, lat;
	int maxlength, nstrokes, address0 = 0;
	int address1, address2, index1, index2;
	Pixel pix;
	/* filehandles */
        void *fh = NULL, *fo = NULL;

	cmd_string = realloc(cmd_string, 2);
	*cmd_string = '\0';
	cmd_killed_char = '\0';
	caret_pos = 0;
	setup_change = 0;

	if (scene->preimage && scene->gopt.hierarchy[0]=='0' &&
            !essential_change(scene, scene->backup)) {
	   l = 0;
           memcpy(scene->image->data, scene->preimage,
	      scene->image->bytes_per_line * scene->image->height);
	   goto direct;
	}

	/* initialize the image stuff */
	image_init(scene);

	/* compute the rotation matrix */
	computeMatrix(scene);

	/* set the background color or starry sky */
	set_bg_sky(scene);

#ifdef DEBUG
	fprintf(stderr,"-- GENERATE --\n");
	fprintf(stderr,"o Options\n");
	fprintf(stderr,"  - slatitude   [%g]\n", scene->ropt.spacing_lat);
	fprintf(stderr,"  - slongitude  [%g]\n", scene->ropt.spacing_lon);
	fprintf(stderr,"  - zoom        [%g]\n", scene->gopt.zoom);
	fprintf(stderr,"  - aspect      [%g]\n", scene->aspect);
	fprintf(stderr,"  - latitude    [%g]\n", scene->gopt.xrot);
	fprintf(stderr,"  - longitude   [%g]\n", scene->gopt.yrot);
	fprintf(stderr,"  - inclination [%g]\n", scene->gopt.zrot);
	fprintf(stderr,"  - projection  [%d]\n", scene->projection);
	fprintf(stderr,"  - mapfile     [%s]\n", scene->mapfile);
        fprintf(stderr,"  - rcfile      [%s]\n", scene->rcfile);
	fprintf(stderr,"  - locfile     [%s]\n", scene->locfile);
	fprintf(stderr,"  - des_cat     [%X]\n", scene->gopt.desired_categories);
	fprintf(stderr,"  - des_cont    [%X]\n", scene->gopt.desired_continents);
	fprintf(stderr,"o ImageLayout (Scene)\n");
	fprintf(stderr,"  - height      [%d]\n", scene->height);
	fprintf(stderr,"  - width       [%d]\n", scene->width);
	fprintf(stderr,"  - transp.     [%d]\n", scene->gopt.transparent);
	fprintf(stderr,"o Colormap\n");
	for (i=0; i<C_ENDCOLOR; i++) 
			if (scene->color[i].bool != INACTIVE)
				fprintf(stderr,"    - (%2d) [%s]  %s\n",i , (scene->color[i].bool)? "on":"off", scene->color[i].name);
	fprintf(stderr, "o Rotation Matrix\n");
	for (i=0; i<3; i++) 
		fprintf(stderr,"    [%10.5f]  [%10.5f]  [%10.5f]\n", 
                    scene->rotation.comp[3*i],  
		    scene->rotation.comp[3*i+1], 
                    scene->rotation.comp[3*i+2]);
	fprintf(stderr,"-- -------- --\n");
#endif /* DEBUG */

	/* Open the mapfile and find graph of the appropriate sectors */
	if (!fullmap_buffer) {
#ifdef ZLIB
 	    fh = (void *)gzopen(scene->mapfile, "r");
	    if (fh == NULL) {
	        char *name;
	        name = scene->mapfile+strlen(scene->mapfile)-3;
		if (strcmp(name, ".gz")) {
	           name = (char *)malloc(512);
		   strcpy(name, scene->mapfile);
		   strcat(name, ".gz");
		   fh = gzopen(name, "r");
		   free(name);
		}
	    }
	    if (fh == NULL) {
	        char *err_msg = malloc(512);
		sprintf(err_msg, msg[COULDNT_OPEN_VECTOR_MAP],
                                 scene->mapfile);
                fprintf(stderr, "%s\n", err_msg);
		if (runlevel) {
		   
		   drawString(scene->image, 8, fontasc[0]+fonth[0]+18,
                      err_msg, 0, scene->color[C_FG_CARET].pix);
		   putPicture(scene);
		}
		free(err_msg);
		return;
	    }
#else
            fh = (FILE *)fopen(scene->mapfile, "r");
            if (fh == NULL) {
		fprintf(stderr, msg[COULDNT_OPEN_VECTOR_MAP],
                                scene->mapfile);
                fprintf(stderr, "\n");
		exit (-1);
	    }
#endif
	}

	/* read the header */
	if (! header) {
            header = (Header *) malloc(sizeof(Header));
#ifdef ZLIB
	    gzread((gzFile *)fh, header, sizeof(Header));
#else
	    fread(header, 1, sizeof(Header), (FILE *)fh);
#endif
	    if (strcmp(header->magic, HEADER_MAGIC)) {
		fprintf(stderr, msg[MAGIC_INFORMATION1],
                        scene->mapfile, header->magic);
		fprintf(stderr, "\n\n");
		fprintf(stderr, msg[MAGIC_INFORMATION2], HEADER_MAGIC);
		fprintf(stderr, "\n%s\n", msg[MAGIC_INFORMATION3]);
		exit(-2);
	    }
	}

        numseg = str4toint(&header->index[556]);

	if (!segment_index) {
#ifdef TRACE
           fprintf(stderr, 
               "Allocating index buffer of size %d (%d headers)\n",
               numseg * sizeof(Segment_index), numseg);
#endif	   
	   segment_index = (Segment_index *)
                    malloc(numseg * sizeof(Segment_index));
#ifdef ZLIB
	   gzseek((gzFile *)fh, sizeof(Header), SEEK_SET);
	   gzread((gzFile *)fh, segment_index, 
                      numseg * sizeof(Segment_index));
#else
	   fseek((FILE *)fh, sizeof(Header), SEEK_SET);
	   fread(segment_index, 1,
                    numseg * sizeof(Segment_index), (FILE *)fh);
#endif
	}

	address0 = sizeof(Header) + numseg*sizeof(Segment_index);

	if (use_memory && !fullmap_buffer) {
	   i = str4toint(segment_index[numseg-1].address) - address0;
#ifdef TRACE
	   fprintf (stderr, 
	        "Addresses %d + %d = %d\n", address0, i, address0+i);
           fprintf(stderr, 
               "Allocating index fullmap buffer of size %d (%d segments)\n",
               i * sizeof(Segment_index), numseg);
#endif	   
	   fullmap_buffer = (signed char *) malloc(i);
#ifdef ZLIB
           gzseek((gzFile *)fh, address0, SEEK_SET);
           gzread((gzFile *)fh, fullmap_buffer, i);
	   gzclose((gzFile *)fh);
#else
           fseek((FILE *)fh, address0, SEEK_SET);
           fread(fullmap_buffer, 1, i, (FILE *)fh);
	   fclose((FILE *)fh);
#endif
	   fh = NULL;
	   segment_buffer = fullmap_buffer;
        }

	if (!segment_buffer) {
	   maxlength = str4toint(header->maxlength);
#ifdef TRACE
           fprintf(stderr, 
               "Allocating segment buffer of size %d (max %d strokes)\n",
               6+2*maxlength, maxlength);
#endif	   
	   segment_buffer = (char *) malloc(6+2*maxlength);
	}

	/* draw gridlines and segments with appropriate hierarchy ordering */
	for (l=4; l>=0; l--) if (scene->gopt.hierarchy[l]=='0') {
	   if (l==0) {
	      i = scene->image->bytes_per_line * scene->image->height;
	      scene->preimage = (char *) malloc(i);
              memcpy(scene->preimage, scene->image->data, i);
	   }
	 direct:
	   /* draw grid and contour */
           drawCircles(scene);
 	   /* iterate through locations and show spot/texts */
	   for (i=0; i<numloc; i++) {
	      locations[i]->drawn = '\0';
	      drawLocation(scene, locations[i]);
	   }
	}
	else
	for (u=0; u<140; u++) {
	    i = u/28; /* number of continent */
	    if ( !((1<<i) & scene->gopt.desired_continents) ) continue;

            /* number of item */
	    j = u%28; 
            /* type of item, INT=0, NAT=1, RIV=2 or CIL=3 */
	    if (j <= 2) 
               k = 0;
            else
	    if (j <= 4) 
               k = 1;
	    else
	    if (j <= 15) 
               k = 2;
	    else
	       k = 3;
	    if (k!=scene->gopt.hierarchy[l]-'1') continue;

	    if ( !((1<<k) & scene->gopt.desired_categories) ) continue;

	    if (scene->color[j].bool!=ON) {
#ifdef TRACE
	       fprintf(stderr, "Skipping u = %d (type = %d)\n", u, j);
#endif TRACE
               continue;
	    }

	    /* initialize escape length */
	    escape_length = scene->width/2;

            pix = scene->color[j].pix;
#ifdef TRACE
	    fprintf(stderr, "Section u = %d (type = %d)\n", u, j);
#endif
            if (u==0)
	       index1 = 0;
	    else
               index1 = str4toint(&header->index[4*(u-1)]);
	    index2 = str4toint(&header->index[4*u]);
#ifdef TRACE
	    fprintf(stderr, "ind1 = %d, ind2 = %d\n", index1, index2);
#endif
	    for (i=index1; i<index2; i++) {
	       ret = segment_in_scene(scene, &segment_index[i]);
	       if (ret == YES) {
		  if (i)
 	              address1 = str4toint(segment_index[i-1].address);
                  else
		      address1 = address0;
		  address2 = str4toint(segment_index[i].address);
		  nstrokes = (address2-address1-6)/2;

		  if (fullmap_buffer) {
                      segment_buffer = &fullmap_buffer[address1 - address0];
		  } else {
#ifdef ZLIB
	              gzseek((gzFile *)fh, address1, SEEK_SET);
	              gzread((gzFile *)fh, segment_buffer, address2-address1);
#else
	              fseek((FILE *)fh, address1, SEEK_SET);
	              fread(segment_buffer, 1, address2-address1, (FILE *)fh);
#endif
		  }
                  lon = str3toint(segment_buffer);
		  lat = str3toint(&segment_buffer[3]);
	          sp.sph[LONGITUDE] = ((double)lon)/3600.0;
	          sp.sph[LATITUDE]  = ((double)lat)/3600.0;
                  pt2 = project(scene, spher2cart(scene, sp));

#ifdef SUPERTRACE
	          fprintf(stderr, "i = %d  addr1 = %d, addr2 = %d  n = %d\n"
			          "  orig_lat = %g  orig_lon = %g\n",
                       i, address1, address2, nstrokes, 
                       s2.sph[LATITUDE], s2.sph[LONGITUDE]);
#endif

	          for (j=0; j<nstrokes; j++) {
		     /* iterate and draw all the appropriate points */
		     pt1 = pt2;
		     m = 6+2*j;
		     lon += (int)segment_buffer[m];
		     lat += (int)segment_buffer[m+1];
#ifdef SUPERTRACE
                     fprintf(stderr, "dx = %d, dy = %d\n",
			     (int)segment_buffer[m],
                             (int)segment_buffer[m+1]);
#endif
	             sp.sph[LONGITUDE] = ((double)lon)/3600.0;
	             sp.sph[LATITUDE]  = ((double)lat)/3600.0;
	             pt2 = project(scene, spher2cart(scene, sp));
	             drawSegment(scene, pt1, pt2, pix);
		  }
	       } else 
               if (ret == POINT_ONLY) {
		  /* the whole segment amounts to a point 
                     based on the perspective of the scene, 
                     so just draw a point	*/
		  sp.sph[LONGITUDE] = 
                     str3toint(segment_index[i].maxlong)/3600.0;
		  sp.sph[LATITUDE] = 
                     str3toint(segment_index[i].maxlat)/3600.0;
		  pt1 = project(scene, spher2cart(scene, sp));
		  drawPoint(scene, pt1, pix);
	       }
	    }
	}

        if (fh)
#ifdef ZLIB
	   gzclose((gzFile *)fh);
#else
	   fclose((FILE *)fh);
#endif

        putPicture(scene);
	memcpy(scene->backup, scene, sizeof(ImageLayout));

	if (img_output) {
	   openOutfile(scene, &fo);
	   printImage(fo, scene);
           closeOutfile(fo);
	}
	discard_events(mainwin);
}

void check_datachanges(ImageLayout * scene)
{
      ImageLayout *backup;
      backup = (ImageLayout *)scene->backup;
      if (strcmp(scene->rcfile, backup->rcfile))
         parse_rcfile(scene);
      if (strcmp(scene->locfile, backup->locfile))
         parse_locfile(scene);	   
      if (strcmp(scene->mapfile, backup->mapfile))
	 free_data_buffers(scene);
}


void explain_tz(char *out, char *data)
{
   char *ptr, *ptrp;
   char explan[256], compl[80];
   int month[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
   int i, j, k, l;

   ptr = data;
   while (*ptr>='A' && *ptr<='Z') ++ptr;
   strncpy(explan, data, ptr-data);
   explan[ptr-data] = '\0';
   strcat(out, explan);

   ptrp = ptr;
   while (*ptrp && *ptrp<'A') ++ptrp;
      if (*ptr=='-') {
      strncpy(explan, ptr, ptrp-ptr);
      explan[0] = '+';
      explan[ptrp-ptr] = '\0';
   } else
   if (*ptr=='+') {
      strncpy(explan, ptr, ptrp-ptr);
      explan[0] = '-';
      explan[ptrp-ptr] = '\0';
   } else {
      if (atoi(ptr)) {
         strncpy(explan+1, ptr, ptrp-ptr);
         explan[0] = '-';
         explan[ptrp-ptr+1] = '\0';
      } else {
	 strcpy(explan, "0");
      }
   }
   i = atoi(explan);
   j = -1;
   if ((ptrp=index(explan, ':'))) j = atoi(ptrp+1);
   strcat(out, explan);

   if (!index(ptr,',')) return;
   strcat(out, "/");

   while (*ptr && *ptr<'A') ++ptr;
   ptrp = ptr;
   while (*ptrp && *ptrp!=',') ++ptrp;
   strncpy(explan, ptr, ptrp-ptr);
   explan[ptrp-ptr] = '\0';
   strcat(out, explan);
   if (j==-1) {
      if (i+1>0)
         sprintf(explan, "+%d", i+1);
      else
         sprintf(explan, "%d", i+1);
   }
   else {
      if (i+1>=0)
         sprintf(explan, "+%d:%d ", i+1, j);
      else
      if (i+1<0)
         sprintf(explan, "%d:%d", i+1, j);
   }
   strcat(out, explan);
   if (!*ptrp) return;
   strcat(out, ", ");

   if (*ptrp==',') ++ptrp;
 repscan:
   if (*ptrp=='M') {
      sscanf(ptrp+1, "%d.%d.%d/%s", &i, &j, &k, compl);
      if (j<=0 || j>5) j=1;
      strcpy(explan, msg[ORDINALS+j-1]);
      strcat(explan, " ");
      if (k<0 || k>=7) k=0;
      l = msg[LENGTH_DAY_ABBREVIATION][k]-'0';
      strncat(explan, msg[NAME_OF_DAYS+k], l);
      strcat(explan, " ");
      if (i<0 || i>11) i=0;
      l = msg[LENGTH_MONTH_ABBREVIATION][i]-'0';
      strncat(explan, msg[NAME_OF_MONTHS+i], l);
      strcat(out, explan);
      if (!*compl) return;
      if ((ptrp=index(compl,','))) *ptrp = '\0';
      strcat(out, " @");
      strcat(out, compl);
      if (!ptrp) return;
      ++ptrp;
   } else
   if (*ptrp=='J') {
      sscanf(ptrp+1, "%d/%s", &j, compl);
      if (j>365) j=365;
      i = 0;
      while (j>month[i]) {
	 j -= month[i];
	 ++i;
      }
      sprintf(explan, "%d", j);
      strcat(explan, " ");
      l =  msg[LENGTH_MONTH_ABBREVIATION][i]-'0';
      strncat(explan, msg[NAME_OF_MONTHS+i], l);
      strcat(out, explan);
      if (!*compl) return;
      if ((ptrp=index(compl,','))) *ptrp = '\0';
      strcat(out, " @");
      strcat(out, compl);
      if (!ptrp) return;
      ++ptrp;     
   } else
      return;
   if (*ptrp) {
         strcat(out, "/");
         goto repscan;
   }
}

void show_coord(ImageLayout * scene, double x, double y)
{
char str[256], compl1[40], compl2[40];
int i;

   if (cmd_mode==MODE_RUN && but_pressed==3) {
      strcpy(compl1, msg[RECENTER_MAP]);
      strcpy(compl2, msg[CANCEL_MOVE]);
   } else
      *compl1 = *compl2 = '\0';
   sprintf(str, "%slat = %.3f  lon = %.3f%s", compl1, x, y, compl2);
   i = cmd_mode;
   cmd_mode = MODE_RUN;
   drawWinString(scene, str, line_pos);
   cmd_mode = i;
   point_shown = 1;
}

void set_time(int num, char *str)
{
   struct tm ltp;
   time_t gtime;

   time(&gtime);
   setenv("TZ", timezones[num]->posix, 1);
   ltp = *localtime(&gtime);
   sprintf(str, "%s %02d:%02d:%02d %s, %s %d %s %d", 
                   msg[DATE_TIME],
                   ltp.tm_hour, ltp.tm_min, ltp.tm_sec, 
#ifdef NEW_CTIME
                   ltp.tm_zone,
#else
                   tzname[ltp.tm_isdst],
#endif
	           msg[NAME_OF_DAYS+ltp.tm_wday], ltp.tm_mday,
                   msg[NAME_OF_MONTHS+ltp.tm_mon], ltp.tm_year+1900);
}

void show_mark(ImageLayout * scene)
{
   Vector pt;
   SpherePt spt;
   int u, v;

   if (city_found<0) return;
   if (cmd_mode!=MODE_RUN) return;
   if (locations[city_found]->obj=='l') return;

   putPicture(scene);
   city_shown = 1;
   exposed |= DONE_MARK;

   spt.sph[LONGITUDE] = locations[city_found]->lon;
   spt.sph[LATITUDE] = locations[city_found]->lat;
   pt = project(scene, spher2cart(scene, spt));
   u = pt.coord[X] + scene->width/2;
   v = pt.coord[Y] + scene->height/2;
   XSetForeground(dpy, gc, scene->color[C_MARKED_POINT].pix);
   XDrawLine(dpy, mainwin, gc, u-6, v, u+6, v);
   XDrawLine(dpy, mainwin, gc, u, v-6, u, v+6);
   XSetForeground(dpy, gc, scene->color[C_FG_TEXT].pix);
}

void show_datawin(ImageLayout * scene, int raise)
{
   XpmAttributes attrib;
   XImage *image1 = NULL, *image2=NULL, *pixmap=NULL;
   XpmColorSymbol flagbg;
   struct stat buf;
   int i, j=0, k;
   char str[11][256],
	path1[256], path2[256], name[256], 
        alt[20], pop[20], rank1[20], rank2[20], rank3[20], adm_status[40];
   char *flagpath, *ptr, *ptrp;

   if (city_found<0) return;
   XSetWindowBackground(dpy, datawin, scene->color[C_BG_TEXT].pix);
   datawin_on = 1;
   exposed &= ~DATA_WINDOW;
   data_text = realloc(data_text, 4*sizeof(char));
   *data_text = '\0';

   if (locations[city_found]->obj == 'l') {
      scene->position[scene->num_positions-1].loc = city_found;
      if (raise)
         XMapRaised(dpy, datawin);
      else
         XMapWindow(dpy, datawin);
      datawin_height = 2*fonth[0]+20;
      XResizeWindow(dpy, datawin, datawin_width, datawin_height);
      XFlush(dpy);
      usleep(20000);
      XClearArea(dpy, datawin, 0, 0, datawin_width, datawin_height, False);
      strcpy(name, locations[city_found]->name);
      simplify_string(name);
      sprintf(str[0], "%s, %s  (%c, %d)",
    	 name, msg[BEGIN_CONTINENT_DESCRIPTION+locations[city_found]->tz],
         locations[city_found]->status, locations[city_found]->population);
      sprintf(str[1], "%s = %.3f  %s = %.3f",
           msg[LATITUDE_ABBREVIATED],
	   locations[city_found]->lat, 
           msg[LONGITUDE_ABBREVIATED],
           locations[city_found]->lon);
      k = 90;
      for (i=0; i<=1; i++) {
	 j = strlen(str[i]);
         XDrawImageString(dpy, datawin, gc, 4, 
            (fonth[0]+4)*(i+1)+2, str[i], j);
	 k += j+2;
         data_text = realloc(data_text, (k+2)*sizeof(char));
	 strcat(data_text, str[i]);
         strcat(data_text, "\n");
      }
      for (j=1; j<=78; j++) strcat(data_text, "-");
      strcat(data_text, "\n");
      discard_events(datawin);
      return;
   }

   if (locations[city_found]->obj != 'c' && 
       locations[city_found]->obj != 'p') return;
   scene->position[scene->num_positions-1].loc = city_found;
   if (locations[city_found]->alt==-30000) 
      strcpy(alt, "?");
   else
      sprintf(alt, "%dm", locations[city_found]->alt);
   if (locations[city_found]->status=='C') 
      sprintf(adm_status, "   (%s)", msg[CAPITAL_OF_STATE]);
   else
   if (locations[city_found]->status=='R') 
      sprintf(adm_status, "   (%s)", msg[CAPITAL_OF_REGION]);
   else
      *adm_status = '\0';
   k = 0;
   for (j=0; j<=strlen(locations[city_found]->name); j++) {
      if (locations[city_found]->name[j]=='(') {
	 name[j] = ' ';
	 k = 1;
      }
      name[j+k] = locations[city_found]->name[j];
   }

   if (locations[city_found]->rank) 
      sprintf(rank1, "%d", locations[city_found]->rank);
   else
      strcpy(rank1, "-");
   if (locations[city_found]->contrank) 
      sprintf(rank2, "%d", locations[city_found]->contrank);
   else
      strcpy(rank2, "-");
   if (locations[city_found]->worldrank) 
      sprintf(rank3, "%d", locations[city_found]->worldrank);
   else
      strcpy(rank3, "-");

   sprintf(str[1], "%s = %.3f  %s = %.3f   %s = %s",
           msg[LATITUDE_ABBREVIATED],
	   locations[city_found]->lat, 
           msg[LONGITUDE_ABBREVIATED],
           locations[city_found]->lon, 
           msg[ALTITUDE_ABBREVIATED], alt);

   if (locations[city_found]->obj=='c') {
      sprintf(str[0], "%s, %s%s",
  	 name, countries[timezones[locations[city_found]->tz]->country]->name,
         adm_status);
      sprintf(str[2], msg[ADMINISTRATIVE_REGION],
                 (locations[city_found]->region[0])?
                  locations[city_found]->region : "-");
      if (locations[city_found]->population) {
         sprintf(pop, "%d", locations[city_found]->population);
         sprintf(str[4], msg[POPULATION_NUMBER], pop);
         sprintf(str[5], msg[CONTINENT_AND_WORLD_RANKS], rank1, rank2, rank3);
         sprintf(str[3], "%s    %s", str[4], str[5]);
      }
      else
         sprintf(str[3], msg[POPULATION_NUMBER], "?");
   }
   if (locations[city_found]->obj=='p') {
      sprintf(str[0], name);
      sprintf(str[2], locations[city_found]->region);
      sprintf(str[3], msg[CONTINENT_AND_WORLD_RANKS], rank1, rank2, rank3);
   }

   j = locations[city_found]->tz;
   k = timezones[j]->country;
   sprintf(str[4], "%s", countries[k]->name);
   sprintf(str[5], msg[DATA_CURRENCY],
            countries[k]->currency, countries[k]->curr_code);
   sprintf(str[6], msg[DATA_TELEPHONE], countries[k]->telephone);
   if (countries[k]->capital[0]=='@') {
      k = (countries[k]->capital[1]-'A')*26+(countries[k]->capital[2]-'A');
      k = country_table[k];
      sprintf(str[7], msg[DATA_TERRITORY], countries[k]->name);
   } else
      sprintf(str[7], msg[DATA_AREA], countries[k]->area);
   sprintf(str[8], msg[DATA_POPULATION],
	    countries[k]->capital, 
            countries[k]->population);
   sprintf(str[9], "%s %s; ", msg[AREA_AND_TIMEZONE], timezones[j]->name);
   explain_tz(str[9], timezones[j]->posix);
   set_time(j, str[10]);
   time_count = 1;

   flagpath = (big_flags)? BIGFLAGPATH : SMALLFLAGPATH;

 fallback:
   if (stat(flagpath, &buf) || !S_ISDIR(buf.st_mode)) {
      if (flagpath==BIGFLAGPATH) {
         flagpath = SMALLFLAGPATH;
         big_flags = -1;
         goto fallback;
      }
   }

   if (stat(flagpath, &buf) || !S_ISDIR(buf.st_mode)) {
      pixmap1_width = pixmap1_height = 0;
      pixmap2_width = pixmap2_height = 0;
      goto skipflags;
   }
   else {
      ptr = timezones[j]->flags;
      if ((ptrp = index(ptr, ','))) {	
         *ptrp = '\0';
	 if (big_flags==1) {
            sprintf(path1, "%s/%s.xpm.gz", flagpath, ptr);
	    *path2 = '\0';
	 } else
	 if (big_flags==2) {
            sprintf(path1, "%s/%s.xpm.gz", flagpath, ptrp+1);
	    *path2 = '\0';
	 } else {
            sprintf(path1, "%s/%s.xpm.gz", flagpath, ptr);
            sprintf(path2, "%s/%s.xpm.gz", flagpath, ptrp+1);
	 }
         *ptrp = ',';
      } else {
         sprintf(path1, "%s/%s.xpm.gz", flagpath, ptr);
         *path2 = '\0';
      }
   }

   j = 0;
   flagbg.name = NULL;
   flagbg.value = "None";
   flagbg.pixel = scene->color[C_BG_TEXT].pix;
   attrib.valuemask = XpmColormap | XpmReturnPixels | XpmColorSymbols;
   attrib.colormap = cmap;
   attrib.numsymbols = 1;
   attrib.colorsymbols = &flagbg;
   if (XpmReadFileToImage(dpy, path1, &image1, NULL, &attrib)) {
      image1 = NULL;
      pixmap1_width = XTextWidth(xfont[0], path1, strlen(path1))+20;
      pixmap1_height = 3*fonth[0]+16;
      fprintf(stderr, msg[CANNOT_READ_XPM_FILE], path1);
      fprintf(stderr, "\n");
      fflush(stderr);
   } else {
      pixmap1_width = image1->width;
      pixmap1_height = image1->height;
   }
   if (*path2) {
      attrib.valuemask = XpmColormap | XpmReturnPixels | XpmColorSymbols;
      attrib.colormap = cmap;
      attrib.numsymbols = 1;
      attrib.colorsymbols = &flagbg;
      if (XpmReadFileToImage(dpy, path2, &image2, NULL, &attrib)) {
	 image2 = NULL;
         pixmap2_width = XTextWidth(xfont[0], path2, strlen(path2))+20;
	 pixmap2_height = 3*fonth[0]+16;
         fprintf(stderr, msg[CANNOT_READ_XPM_FILE], path2);
         fprintf(stderr, "\n");
	 fflush(stderr);
      } else {
         pixmap2_width = image2->width;
         pixmap2_height = image2->height;
      }
   }

 skipflags:
   pixmap_height = pixmap1_height;
   if (pixmap2_height > pixmap_height) pixmap_height = pixmap2_height;
   if (pixmap_height>0)
      datawin_height = pixmap_height+80+11*fonth[0]; 
   else {
      pixmap_height = -8;
      datawin_height = 74+11*fonth[0];
   }
   if (datawin_x>-100000) {
      XSizeHints xsh;
      xsh.flags = USPosition | PSize;
      xsh.x = datawin_x;
      xsh.y = datawin_y;
      xsh.width = datawin_width;
      xsh.height = datawin_height;
      XSetNormalHints(dpy, datawin, &xsh);
   }

   if (raise)
      XMapRaised(dpy, datawin);
   else
      XMapWindow(dpy, datawin);
   XResizeWindow(dpy, datawin, datawin_width, datawin_height);
   if (datawin_x>-100000) {
      XMoveWindow(dpy, datawin, datawin_x, datawin_y);
      datawin_x = -1000000;
   }
   XFlush(dpy);
   usleep(30000);
   discard_events(datawin);
   
   show_mark(scene);
   XSetWindowBackground(dpy, datawin, scene->color[C_BG_TEXT].pix);
   XClearArea(dpy, datawin, 0, 0, datawin_width, pixmap_height+8, False);
   XFlush(dpy);

   if (image1) {
      XPutImage(dpy, datawin, gc, image1, 0, 0, 4, 4, 
               image1->width, image1->height);
      XDestroyImage(image1);
   } else if (*path1) {
      XDrawRectangle(dpy, datawin, gc, 4, 4,
	                  pixmap1_width, pixmap1_height);
      XDrawImageString(dpy, datawin, gc, 12, 8+fonth[0], path1, strlen(path1));
      XDrawImageString(dpy, datawin, gc, 12, 8+2*fonth[0], "???", 3);
   }
      
   if (image2) {
      XPutImage(dpy, datawin, gc, image2, 0, 0, pixmap1_width+12, 4, 
               image2->width, image2->height);
      XDestroyImage(image2);
   } else if (*path2) {
      XDrawRectangle(dpy, datawin, gc, pixmap1_width+12, 4,
	                  pixmap2_width, pixmap2_height);
      XDrawImageString(dpy, datawin, gc, pixmap1_width+20, 8+2*fonth[0], 
                            path2, strlen(path2));
      XDrawImageString(dpy, datawin, gc, pixmap1_width+20, 8+3*fonth[0], 
                            "???", 3);
   }

   XClearArea(dpy, datawin, 0, pixmap_height+8, 
                   datawin_width, 11*(fonth[0]+4)+20, False);
   k = 90;
   for (i=0; i<=10; i++) {
      j = strlen(str[i]);
      XDrawImageString(dpy, datawin, gc, 4, 
         pixmap_height+(fonth[0]+4)*(i+1)+8+((i>=4)?10:0)+((i>=9)?10:0), 
         str[i], j);
	 k += j+2;
         data_text = realloc(data_text, (k+2)*sizeof(char));
	 if (i!=4 && i!=9) strcat(data_text, str[i]);
	 if (i==9) {
            for (j=1; j<=78; j++) strcat(data_text, "-");
            strcat(data_text, "\n");
	 }
         strcat(data_text, "\n");
   }

   j = pixmap_height+5*(fonth[0]+4)+12-fontasc[0];
   XDrawLine(dpy, datawin, gc, 0, j, datawin_width, j); 

   for (k=0; k<=1; k++) {
      XpmCreateImageFromData(dpy, (k==0)?icon_editor:icon_talk, 
         &pixmap, NULL, &attrib);
      i = datawin_width-28;
      j = pixmap_height+5*(fonth[0]+4)+12-fontasc[0]+15+25*k;
      XDrawRectangle(dpy, datawin, gc, i, j, 23, 21); 
      XPutImage(dpy, datawin, gc, pixmap, 0, 0, 
                  i+4, j+4, 16, 14);
   }
   j = pixmap_height+11*(fonth[0]+4)+4-fontasc[0];
   XDrawLine(dpy, datawin, gc, 0, j, datawin_width, j); 
   XDestroyImage(pixmap);
}

void redisplay(ImageLayout * scene)
{
   exposed &= ~MAIN_WINDOW;
   if (cmd_mode==MODE_RUN) {
      exposed &= ~DONE_MARK;
      if (datawin_on)
         show_mark(scene);
      if (!datawin_on || !(exposed & DONE_MARK))
         putPicture(scene);
   }
   if (cmd_mode!=MODE_RUN) 
      drawWinCmd(scene, 0);
}

void close_datawin(ImageLayout * scene)
{
   int i;
   city_found = -1;
   getWindowPlacement(datawin, &datawin_x, &datawin_y, &i, &i);
   big_flags = 0;
   datawin_on = 0;
   XUnmapWindow(dpy, datawin);
   redisplay(scene);
}

void show_point(ImageLayout * scene, int u, int v)
{
   double x, y, eps;
   int i=0, j, k, found, popul, max;
   int up, vp, uf=0, vf=0;

   found = -1;
   if (inverse_coord(scene, u, v, &x, &y)) {
      eps = 1.5/scene->gopt.zoom;
      if (cmd_mode==MODE_RUN && but_pressed==1) {
	 popul = 0;
	 max = 10;
         for (i=0; i<numloc; i++) {
	    if ((locations[i]->obj=='c' || locations[i]->obj=='p') &&
                locations[i]->drawn=='+' &&
	        fabs(x-locations[i]->lat)+fabs(y-locations[i]->lon)<eps) {
	       Vector pt;
               SpherePt spt;
	       spt.sph[LONGITUDE] = locations[i]->lon;
	       spt.sph[LATITUDE] = locations[i]->lat;
	       pt = project(scene, spher2cart(scene, spt));
	       up = pt.coord[X] + scene->width/2;
               vp = pt.coord[Y] + scene->height/2;
	       j = abs(u-up);
	       k = abs(v-vp);
               if (j<3 && k<3 && j+k<=max) {
		  if (j+k<max) popul = 0;
		  max = j+k;
		  if (locations[i]->population>popul) {
		     found = i;
		     uf = up;
		     vf = vp;
                     popul = locations[i]->population;
		  }
	       }
	    }
	 }
      }
      if (city_shown && but_pressed==1) {
	 if (found==-1) return;
	 if (found==city_found) return;
      }

      if (v>scene->height/2)
	 j = TOP;
      else
	 j = BOTTOM;
      if (found==-1) {
	 if (line_pos!=j) putPicturePiece(scene, line_pos);
	 line_pos = j;
         show_coord(scene, x, y);
      } else {
	 city_found = found;
         show_datawin(scene, 1);
      }
   } else {
      if (point_shown) putPicturePiece(scene, line_pos);
      point_shown = 0;
   }
}

void do_escape(ImageLayout * scene)
{
int i;
	line_clicked = -1;
	*cmd_string = '\0';
	caret_pos = 0;
        cmd_mode = MODE_RUN;
	cmd_type = M_RUN_START;
        show_title(scene, mainwin);
	memcpy(scene, scene->backup, sizeof(ImageLayout));
	if (setup_change & FONT_CHANGE) {
           for (i=0; i<=5; i++)
	      set_font(scene->ropt.font[i], i);
           setup_change &= ~FONT_CHANGE;
	}
	if (setup_change & COLOR_CHANGE) {
	   get_all_colors(scene);
	   setup_change &= ~COLOR_CHANGE;
	   if (color_depth<=8) {
	      generate(scene);
              return;
	   }
	}
	putPicture(scene);
}

void do_activate(ImageLayout * scene, KeySym keysym)
{
char str[512], outfilename[256], compl[6];
struct stat buf;

	if (cmd_mode==MODE_MENU && cmd_type==M_MENU_SEARCH) {
	   if (!*cmd_string) {
              drawWinCmd(scene, 2);
	      if (keysym==XK_Return) return;
	   }
	   line_clicked = -1;
           search_index = -1;
	   free(search_string);
           search_string = strdup(cmd_string);
	   search_city(scene);
	   search_output(scene, (keysym==XK_exclam));
	   return;
	} 

	if (cmd_mode==MODE_FILE && 
            (cmd_type==M_FILE_DATA || cmd_type==M_FILE_PRINTCFG)) {
           str2value(scene, line_clicked, cmd_string);
	   if (keysym==XK_Return) {
              drawWinCmd(scene, 2);
	      return;
	   }
	}
	if (cmd_mode==MODE_FILE && cmd_type==M_FILE_SAVE) {
           void * fo = NULL;
	   if (line_clicked<=1) {
      	      strncpy(scene->outfile, cmd_string, 255);
	      scene->outfile[255] = '\0';
	      drawWinCmd(scene, 2);
	      return;
	   }
	   #ifdef ZLIB
 	   if (im_compress && *scene->outfile &&
              strcmp(scene->outfile+strlen(scene->outfile)-3,".gz"))
   	      strcpy(compl, ".gz");
   	   else
	   #endif
   	      *compl = '\0';
           sprintf(outfilename, "%s%s", scene->outfile, compl);
	   if (runlevel && line_clicked==2 && !*scene->outfile) {
	     noname:
	      drawWinString(scene, msg[NO_FILENAME_SPECIFIED], BOTTOM);
	      return;
	   }
	   if (line_clicked==3) {
	      if (!*scene->outfile) goto noname;
              stat(outfilename, &buf);
              if (!S_ISREG(buf.st_mode)) {
	       notsaved:
 	         drawWinString(scene, msg[FILE_NOT_SAVED], BOTTOM);
		 return;
	      }
	      sprintf(str, msg[VIEWING_IMAGE], outfilename);
	      drawWinString(scene, str, BOTTOM);
	      viewOutfile(scene);
	      return;
	   }
	   if (line_clicked==4) {
	      if (!*scene->outfile) goto noname;
              stat(outfilename, &buf);
              if (!S_ISREG(buf.st_mode)) goto notsaved;
	      if (strcmp(outfilename+strlen(outfilename)-3,".ps") &&
	          strcmp(outfilename+strlen(outfilename)-6,".ps.gz"))
		 goto notsaved;
	      sprintf(str, msg[PRINTING_IMAGE], outfilename);
	      drawWinString(scene, str, BOTTOM);
	      if (strcmp(outfilename+strlen(outfilename)-3,".gz"))
                 sprintf(str, "%s %s &", print_cmd, outfilename);
	      else
                 sprintf(str, "gzip -cd %s | %s - &", outfilename, print_cmd);
	      system(str);
              return;
	   }
	   if (overwrite==1) overwrite=-1;
	   openOutfile(scene, &fo);
	   if (overwrite==-1) {
	      get_output_type(scene);
   	      sprintf(str, msg[WRITING_IMAGE_TO_FILE], outfilename);
	      drawWinString(scene, str, BOTTOM);  
	      printImage(fo, scene);
              closeOutfile(fo);
	      strcpy(((ImageLayout *)scene->backup)->outfile, scene->outfile);
	      img_output = OUT_NONE;
	      overwrite = 0;
	   }
	}
	if (cmd_mode==MODE_CMD) {
	   if (*cmd_string && parse_cmd_line(scene, cmd_string)) return;
	}
	if (keysym==XK_exclam) {
	   if (cmd_mode == MODE_MENU && cmd_type==M_MENU_LIST &&
	       search_index>=0) {
	      if (num_found==1) {
		 search_output(scene, 1);
		 return;
	      }
	      if (search_index!=-1) city_found = search_index;
	   }
           if (!memcmp((char *)scene, (char *)scene->backup, 
               sizeof(ImageLayout))) { 
	      drawWinString(scene, msg[NO_CHANGE_MADE], BOTTOM);
	      return;
	   }
	   cmd_mode = MODE_RUN;
	   cmd_type = M_RUN_START;
	   if (loc_pointer) {
	      free(loc_pointer);
	      loc_pointer = NULL;
	   }
           show_title(scene, mainwin);
	   putPicture(scene);
           /* look whether some options have changed that force to 
              regenerate image */
           if (memcmp((char *)scene, (char *)scene->backup, 
	        sizeof(ImageLayout)-256-
                (C_ENDCOLOR-C_FG_TEXT)*sizeof(Colordata))) { 
              check_datachanges(scene);
	      generate(scene);
	   }
	   if (search_index==city_found && search_index>=0) 
              show_datawin(scene, 1);
	   return;
	}
        /* Remaining text_input cases with XK_Return */
	if (text_input_mode()) {
	   if (line_clicked>=1 && line_clicked<=box_numitems) {
	      str2value(scene, line_clicked, cmd_string);
	      if (setup_change) {
		 show_title(scene, mainwin);
		 drawWinCmd(scene, 0);
	      } else
		 drawWinCmd(scene, 2);
	   }
        } else
	   drawWinCmd(scene, 0);
}

void do_caret(ImageLayout * scene, int u)
{
        int l;
        l = strlen(cmd_string);
        caret_pos = 0;
        while (caret_pos<l &&
           XTextWidth(xfont[0], cmd_string, caret_pos+1)<u-4) ++caret_pos;
	drawWinString(scene, cmd_string, BOTTOM);
        XFlush(dpy);
}

void filemenu(ImageLayout * scene)
{
	overwrite = 0;
	cmd_mode = MODE_FILE;
	cmd_type = M_FILE_START;
	show_title(scene, mainwin);
	cmd_string = (char *)realloc(cmd_string, 1);
	*cmd_string = '\0';
	cmd_killed_char = '\0';
	caret_pos = 0;
}

void commands(ImageLayout * scene)
{
    cmd_mode = MODE_CMD;
    cmd_type = M_CMD_START;
    show_title(scene, mainwin);
    drawWinCmd(scene, 0);
}

void list_key_shortcuts(char *str)
{
char *ptr, *ptrp;
int i, j;

   sprintf(str, "%s ", msg[KEY_SHORTCUTS]);
   for (i=BEGIN_LIST_SHORTCUTS; i<END_LIST_SHORTCUTS; i++) {
      ptr = msg[i];
      j = strlen(str);
      str[j] = ' ';
      ++j;
      while (*ptr && isspace(*ptr)) ++ptr;
      ptrp = ptr;
      while (*ptr && !isspace(*ptr)) {
	 str[j] = *ptr;
	 ++ptr;
	 ++j;
      }
      str[j] = '\0';
   }
}

void search_output(ImageLayout * scene, int force)
{
int i;
    if (num_found==0) {
       i = caret_pos;
       caret_pos = -10;
       drawWinString(scene, msg[NO_MATCH_IN_SEARCH], BOTTOM);
       usleep(WARNING_TIME);
       caret_pos = i;
       drawWinString(scene, search_string, BOTTOM);
       if (force) {
          cmd_mode = MODE_RUN;
          cmd_type = M_RUN_START;
          putPicture(scene);
       }
    } else
    if (num_found==1 && force==1) {
	  city_found = loc_pointer[0];
	  scene->gopt.xrot = locations[city_found]->lat;
	  scene->gopt.yrot = locations[city_found]->lon;
	  search_index = city_found;
	  cmd_mode = MODE_RUN;
          cmd_type = M_RUN_START;
          show_title(scene, mainwin);
          check_datachanges(scene);
	  generate(scene);
          show_datawin(scene, 1);
    } else {
       cmd_mode = MODE_MENU;
       cmd_type = M_MENU_SEARCH;
       drawWinCmd(scene, 0);
       cmd_type = M_MENU_LIST;
       drawWinCmd(scene, 1);
    }
}

int main (int argc, char **argv) {
        /* Global geometric structure */
        ImageLayout *scene; 
	/* misc variables */
        XEvent ev;
        KeySym keysym = 0;
	char *string;
	struct stat buf;
	int i, j, l, m;
        char str[256];

	/* Table for Mollweide function
	for (i=0; i<=1000; i++) {
	    double d, e;
	    d = e = (double)i*0.001;
            mollweide(&e);
	    printf("%12.10f --> %12.10f %12.10f\n", d, e, 
              log(1-e)/log(1-d));
	}
	*/

	/* the default behaviour */
	ps_viewer = strdup(PS_VIEWER); /* Should have gv installed !! */
	im_viewer = strdup(IM_VIEWER); /* Should install ImageMagick !! */
	print_cmd = strdup(PRINT_CMD); /* BSD type printing system... */
	editor_cmd = strdup(EDITOR);   /* Default editor */
        *language = '\0';
	bzero(feature_descr, 6*sizeof(char **));
	cmd_string = (char*) malloc(1);
        *cmd_string = '\0';

	search_string = strdup("");
	search_index = -1;

	/* the main struct for the data file */
	scene = (ImageLayout *) malloc(sizeof(ImageLayout));
	scene->backup = (ImageLayout *)NULL;
      
        scene->projection = SPHERICAL;
	scene->position = NULL;
	scene->num_positions = 0;
	scene->aspect = 1.0;
	scene->width = 0;
	scene->height = 0;
        scene->image = NULL;
        scene->preimage = NULL;

	bzero(&scene->rotation, sizeof(Matrix));
	bzero(&scene->gopt, sizeof(GeomOptions));
	bzero(&scene->ropt, sizeof(RenderOptions));

	scene->gopt.zoom = 1.0;
	scene->gopt.desired_categories = 0xFFFFFFFF; /* default is all */
	scene->gopt.desired_continents = 0xFFFFFFFF; /* default is all */
	strcpy(scene->gopt.hierarchy, hierarchy0);

	scene->ropt.spacing_lat = 15.0;
	scene->ropt.spacing_lon = 15.0;
        memcpy(scene->ropt.mark_step, mark_step0, sizeof(mark_step0));
        memcpy(scene->ropt.name_step, name_step0, sizeof(name_step0));
	memcpy(scene->ropt.peak_mark, peak_mark0, sizeof(peak_mark0));
	memcpy(scene->ropt.peak_size, peak_size0, sizeof(peak_size0));
	memcpy(scene->ropt.city_mark, city_mark0, sizeof(city_mark0));
	memcpy(scene->ropt.city_size, city_size0, sizeof(city_size0));
	strcpy(scene->ropt.city_filter, city_filter0);
	strcpy(scene->ropt.loc_filter, loc_filter0);
	
	*scene->mapfile = '\0';
	*scene->outfile = '\0';
	*scene->locfile = '\0';


	for (i=0; i<=5; i++) {
            strcpy(scene->ropt.font[i], "6x13");
	    xfont[i] = NULL;
	}

	/* initial setting (which is to print nothing !!) */
	for (i=0; i< C_ENDCOLOR; i++)
            scene->color[i].bool = INACTIVE;

	list_lang = list_languages();

        /* read i18n file */
	for (i=1; i<argc-1; i++) 
 	   if (!strpcmp(argv[i], "language", 4)) {
	      strncpy(language, argv[i+1], 2);
	      language[2] = '\0';
	      read_i18n_file(scene);
	      argv[i+1] = language;
	      break;
	   }
        if (!*language) read_i18n_file(scene);

	/* parse system wide rcfile */
	strncpy(scene->rcfile, RCFILE, 255);
	scene->rcfile[255] = '\0';
        parse_rcfile(scene);

	/* Check whether user as specified a non standard rcfile */
	*scene->rcfile = '\0';
	for (i=1; i<argc-1; i++) 
	  if (!strpcmp(argv[i], "rcfile", 3)) {
	     strncpy(scene->rcfile, argv[i+1], 255);
	     scene->rcfile[255] = '\0';
	     break;
	  }
	if (*scene->rcfile) {
	   if (!stat(scene->rcfile, &buf)) 
              parse_rcfile(scene);
	   else {
	      fprintf(stderr, msg[RCFILE_DOES_NOT_EXIST], scene->rcfile);
              fprintf(stderr, "\n");
	   }
	} else {   
	   /* if ~/.xrmaprc exists, use that file instead (for a rcfile) */
	   *scene->rcfile = '\0';
	   string = getenv("HOME");
	   if (string != NULL) {
	      sprintf(scene->rcfile, "%s/.xrmaprc", string);
	      if (stat(scene->rcfile, &buf)) *scene->rcfile = '\0';
	   }
	   if (*scene->rcfile) parse_rcfile(scene);
	}

	fix_features_description(scene);

	/* parse the command line options */
	*cmd_string= '\0';
	for (i=1; i<argc; i++) {
	    j = strlen(cmd_string);
	    cmd_string = (char *) realloc(cmd_string, j+strlen(argv[i])+2);
	    cmd_string[j] = ' ';
	    cmd_string[j+1] = '\0';
	    strcat(cmd_string, argv[i]);
	}
	if (*cmd_string && parse_cmd_line(scene, cmd_string)) exit(-1);

	cmd_string = (char *) realloc(cmd_string, 1);
	*cmd_string = '\0';
	cmd_killed_char = '\0';
	if (scene->projection==SPHERICAL) scene->aspect = 1.0;

        /* set the window dimensions */
	if (scene->width==0 && scene->height==0) {
	   if (scene->projection==RECTANGULAR || 
               scene->projection==CYLINDRICAL ||
               scene->projection>=SINUSOIDAL) {
	      scene->width = 1000;
	      scene->height = 500 * scene->aspect;
	   } else {
	      scene->width = 640;
	      scene->height = 640 * scene->aspect;
	   }
	} else
	if (scene->height==0) {
	   if (scene->projection==RECTANGULAR || 
               scene->projection==CYLINDRICAL ||
               scene->projection>=SINUSOIDAL) {
	      scene->height = scene->width * (0.5*scene->aspect);
	   } else {
	      scene->height = scene->width * scene->aspect;
	   }
	} else
	if (scene->width==0) {
	   if (scene->projection==RECTANGULAR ||
               scene->projection==CYLINDRICAL ||
               scene->projection>=SINUSOIDAL) {
	      scene->width = scene->height / (0.5*scene->aspect);
	   } else {
	      scene->width = scene->height / scene->aspect;
	   }
	}

	/* set some default values */
        set_radius(scene);
        check_spacing(scene);

	if (!*scene->mapfile) {
	   strncpy(scene->mapfile, MAPFILE, 255); /* default */
	   scene->mapfile[255] = '\0';
	}
	if (!*scene->locfile) {
	   strncpy(scene->locfile, LOCFILE, 255); /* default */
	   scene->locfile[255] = '\0';
	   parse_locfile(scene);
	}

        /* We are now entering the GUI program */
        runlevel = 1;

	/* Open X display */
        dpy = XOpenDisplay(NULL);
        scr = DefaultScreen(dpy);
        color_depth = DefaultDepth(dpy, scr);
        bigendian = (ImageByteOrder(dpy) == MSBFirst);

        if (color_depth > 16)
            color_pad = 32;
        else if (color_depth > 8)
            color_pad = 16;
        else
            color_pad = 8;

        Root = RootWindow(dpy, scr);
        visual = DefaultVisual(dpy, scr);
	if (color_depth > 8)
           cmap = DefaultColormap(dpy, scr);

        /* Load X fonts */
	for (i=0; i<=5; i++) {
    	   xfont[i] = XLoadQueryFont(dpy, scene->ropt.font[i]);
	   if (xfont[i] == (XFontStruct *)NULL) {
	      fprintf(stderr, msg[REPLACING_BY_DEFAULT_FONT],scene->ropt.font[i]);
              fprintf(stderr, "\n");
              strcpy(scene->ropt.font[i], "fixed");
              xfont[i] = XLoadQueryFont(dpy, scene->ropt.font[i]);
	   }
	   if (xfont[i] == (XFontStruct *)NULL) {
	      fprintf(stderr, msg[CANT_OPEN_FONT], scene->ropt.font[i]);
              fprintf(stderr, "\n");
	      exit(1);
	   }
           set_font_params(i);
	}

	/* get all colors from X colormap */
        get_all_colors(scene);

        /* open the X window */
        mainwin = makeWindow(scene->width, scene->height);
        show_title(scene, mainwin);
	datawin_width = 640;
	datawin = makeWindow(datawin_width, 78+4*fonth[0]);
	sprintf(str, "%s / %s", PACKAGE, msg[CITY_AND_COUNTRY]);
	XStoreName(dpy, datawin, str);

	if (!img_output || preview>0.0) 
           XMapWindow(dpy, mainwin);
       	createGCs(scene);

	if (!scene->backup)
           scene->backup =
           (ImageLayout *) malloc(sizeof(ImageLayout));
	memcpy(scene->backup, scene, sizeof(ImageLayout));

	/* generate the image */
	if (start_menu && !img_output) {
	   image_init(scene);
	   ((ImageLayout *)scene->backup)->image = scene->image;
	   keysym = (KeySym) start_menu;
	   goto startkey;
	} else {
	   if (*search_string) {
	      search_city(scene);
	      if (num_found!=1)
		 generate(scene);
	      search_output(scene, 1);
	   } else
	      generate(scene);
	}

        /* quit after preview time if user only wants file output */
	if (img_output) {
	   if (preview>0.0) {
	      putPicture(scene);
	      usleep((int)(preview*1000000));
	   }
	   exit(0);
        }

	/* check X events */
	for (;;) {
	  if (!XCheckIfEvent(dpy, &ev, evpred, (XPointer)0)) {
	     usleep(20000);
	     i = 0;
	     if (datawin_on && (exposed & DATA_WINDOW)) {
	        show_datawin(scene, 0);
	        i = 1;
	     }
	     if (exposed & MAIN_WINDOW) {
	        redisplay(scene);
		usleep(100000);
		i = 1;
	     }
	     if (i)
                discard_events(mainwin);

	     if (time_count==0 && city_found>=0 && city_found<numloc &&
                 locations[city_found]->obj!='l') {
	        i = locations[city_found]->tz;
	        set_time(i, str);
                XDrawImageString(dpy, datawin, gc, 4, 
                   pixmap_height+11*(fonth[0]+4)+28, str, strlen(str));
	     }
	     time_count = (time_count+1)%33;
	     continue;
	  }

	  if (ev.type==EnterNotify || ev.type==LeaveNotify) {
             move_notify = 1;
	     but_pressed = 0;
	  }
          if (ev.type==ButtonRelease) but_pressed = 0;
          if (ev.type==ButtonPress)
	     but_pressed = ev.xbutton.button;

          if (ev.xclient.message_type == wm_protocols &&
              ev.xclient.format == 32 &&
              ev.xclient.data.l[0] == wm_delete_window) {
	     if (ev.xexpose.window == mainwin)
	        exit(0);
	     if (ev.xexpose.window == datawin) {
                close_datawin(scene);
	     }
	  }

	  if (ev.xexpose.window == datawin)
	  switch(ev.type) {
	     case Expose:
	        exposed |= DATA_WINDOW;
		break;

	     case ButtonRelease:
	        for (i=0; i<=1; i++)
                if (ev.xbutton.x>datawin_width-28 &&
                    ev.xbutton.x<datawin_width-5 &&
                    ev.xbutton.y>
                       pixmap_height+5*(fonth[0]+4)+27+25*i-fontasc[0] &&
                    ev.xbutton.y<
                       pixmap_height+5*(fonth[0]+4)+48+25*i-fontasc[0]) {
		   char cmd[512], path[256], code[4];
		   struct stat buf;
                   char tmpname[80];
                   FILE *ft;
		   if (locations[city_found]->obj!='l') {
		      strcpy(code, countries[timezones[
                              locations[city_found]->tz]->country]->code2);
		      code[0] = tolower(code[0]);
		      code[1] = tolower(code[1]);
		      if (i==0)
		         sprintf(path, "%s/%s.txt", FACTBOOKPATH, code);
		      else
		         sprintf(path, "%s/%s.ant", ANTHEMPATH, code);
		      stat(path, &buf);
                      if (S_ISREG(buf.st_mode)) {
                         sprintf(tmpname, "/tmp/xrmap-%s-XXXXXX", code);
                         mktemp(tmpname);
			 if (i==0)
                            strcat(tmpname, ".txt");
			 if (i==1)
                            strcat(tmpname, ".ant");
                         ft = fopen(tmpname, "w");
                         if (ft) {
                            if (data_text && *data_text)
                               fprintf(ft, data_text);
			    if (i==1)
			       fprintf(ft, "%s\n\n", msg[NATIONAL_ANTHEM]);
                            fclose(ft);
                            sprintf(cmd,
                              "( cat %s >> %s ; %s %s ; rm -f %s ) &",
                              path, tmpname, editor_cmd, tmpname, tmpname);
                            system(cmd);
			 }
		      }
		   }
		   break;
		}
	        if (ev.xbutton.button == 3 && cmd_mode==MODE_RUN) {
		   if (city_found>=0 && city_found<numloc) {
                      XMapRaised(dpy, mainwin);
		      scene->gopt.xrot = locations[city_found]->lat;
		      scene->gopt.yrot = locations[city_found]->lon;
		      drawWinString(scene, msg[RECENTERING_TO_POINT], BOTTOM);
		      generate(scene);
		      show_mark(scene);
 		      city_shown = 1;
		      break;
		   }
		}
                if (ev.xbutton.x>=4 && ev.xbutton.x<4+pixmap1_width &&
                    ev.xbutton.y>=4 && ev.xbutton.y<4+pixmap1_height) {
		   if (big_flags) big_flags = 0;
		      else
		   if (!big_flags) big_flags = 1;
 		   show_datawin(scene, 1);
		}
		else
                if (ev.xbutton.x>=12+pixmap1_width && 
                    ev.xbutton.x<12+pixmap1_width+pixmap2_width &&
                    ev.xbutton.y>=4 && ev.xbutton.y<4+pixmap2_height) {
		   if (big_flags) big_flags = 0;
		      else
		   if (!big_flags) big_flags = 2;
		   show_datawin(scene, 1);
		}
		else
		if (cmd_mode==MODE_RUN)
		   show_mark(scene);
                break;

	     case ConfigureNotify:
	        {
	           int x, y;
                   (void) getWindowPlacement(datawin, 
                       &x, &y, &datawin_width, &datawin_height);
		}
	        time_count = 0;
		break;

	     case KeyPress:
	     case KeyRelease: {
                  char buffer[1];
                  XLookupString((XKeyEvent *)&ev, buffer, 1, &keysym, NULL);
                  if (keysym==XK_k || keysym==XK_K) {
		     close_datawin(scene);
		     break;
		  }
	          ev.xexpose.window = mainwin;
		  XMapRaised(dpy, mainwin);
                  XSetInputFocus(dpy, mainwin, 
                                 RevertToPointerRoot, CurrentTime);
		  break;
	     }

  	     default:
		break;
	  }

	  if (ev.xexpose.window == mainwin)
	  switch(ev.type) {

	     case KeyPress : {
                   char buffer[1];
                   XLookupString((XKeyEvent *) &ev, buffer, 1, &keysym, NULL);
		   if (keysym == XK_Control_L ||
                       keysym == XK_Control_R ) {
		       control_pressed = 1;
		   }
		}
		break;
                	     
	     case KeyRelease : {
                char buffer[1];
                XLookupString((XKeyEvent *) &ev, buffer, 1, &keysym, NULL);
		/* printf("%d\n", keysym); */
	     startkey:
		if (keysym == XK_Control_L ||
                    keysym == XK_Control_R ) {
		    control_pressed = 0;
		    break;
		}
		if (keysym>=XK_KP_0 && keysym<=XK_KP_9)
		   keysym += XK_0-XK_KP_0;
		if (keysym==XK_KP_Add) keysym = XK_plus;
		if (keysym==XK_KP_Subtract) keysym = XK_minus;
		if (keysym==XK_KP_Multiply) keysym = XK_asterisk;
		if (keysym==XK_KP_Divide) keysym = XK_slash;
		if (keysym==XK_KP_Delete) keysym = XK_BackSpace;
		if (keysym==XK_KP_Decimal) keysym = XK_period;
		if (keysym==XK_KP_Enter) keysym = XK_Return;
		if (keysym==XK_BackSpace) keysym = XK_Delete;

                if ((keysym==XK_k || keysym==XK_K) && !text_input_mode()) {
		   close_datawin(scene);
		   break;
		}

		if (cmd_mode>MODE_RUN) {
		   if (keysym==XK_Escape) {
		      do_escape(scene);
		      break;
		   }
		   if (keysym==XK_Return || keysym==XK_exclam) {
		      do_activate(scene, keysym);
		      break;
		   } 
		   if (cmd_mode==MODE_MENU && cmd_type==M_MENU_START) {
                      if (keysym==XK_d || keysym==XK_D ||
                          keysym==XK_f || keysym==XK_F ||
                          keysym==XK_h || keysym==XK_H ||
                          keysym==XK_c || keysym==XK_C ||
                          keysym==XK_o || keysym==XK_O ||
                          keysym==XK_q || keysym==XK_Q ||
			  keysym==XK_s || keysym==XK_S) {
			 if (keysym==XK_h || keysym==XK_H) {
			   cmd_mode = MODE_HELP;
			   cmd_type = M_HELP_START;
			   drawWinCmd(scene, 1);
			   break;
			 }
			 cmd_mode = MODE_RUN;
			 cmd_type = M_RUN_START;
			 goto startkey;
		      }
		      break;
		   }
		   if (keysym==XK_Up) {
		      if (!click_input_mode()) break;
		      l = line_clicked;
		      if (l==-1) 
			 l = box_numitems;
		      else
		      if (l==1) {
                         l = -1;
			 line_clicked = -1;
			 *cmd_string = '\0';
			 caret_pos = 0;
		      } else
			 --l;
                      goto pre_do_click;
		   }
		   if (keysym==XK_Down) {
		      if (!click_input_mode()) break;
		      l = line_clicked;
		      if (l==-1) 
                         l = 1;
		      else {
			 ++l;
			 if (l>box_numitems) {
			    l = -1;
			    line_clicked = -1;
			    *cmd_string = '\0';
			    caret_pos = 0;
			 }
		      }
		   pre_do_click:
		      if (cmd_mode==MODE_MENU && cmd_type==M_MENU_LIST) {
			 if (l==1 && start_list>0) {
			    line_clicked = -10;
			    start_list -= (num_listed-3);
                            if (start_list<0) start_list = 0;
                            l = num_listed-1;
			 } else
                         if (l==num_listed &&
                             start_list+num_listed<num_found) {
                            start_list += (num_listed-3);
			    line_clicked = -10;
                            if (start_list+num_listed>num_found)
                                     start_list = num_found-num_listed;
			    l = 2;
			 }
		      }
                      goto do_click;
		   }
		   if (keysym==XK_Tab) {
                      if (cmd_mode==MODE_FILE) 
                         cmd_type = (cmd_type%M_FILE_SAVE)+1;
		      else
                      if (cmd_mode==MODE_OPTION)
                         cmd_type = (cmd_type%M_OPTION_CAT)+1;
		      else
                      if (cmd_mode==MODE_COLOR) 
                         cmd_type = (cmd_type%M_COLOR_CIL)+1;
                      if (cmd_mode==MODE_HELP) 
                         cmd_type = (cmd_type%M_HELP_MANUAL)+1;
		      line_clicked = -1;
	              drawWinCmd(scene, 0);
		      break;
		   }
		   if (keysym==XK_Page_Up) {
		      int i = fonth[0]+6;
		      XPutImage(dpy, mainwin, gc,
                          scene->image, 0, i, 0, i, 
                          scene->image->width, scene->image->height-i);
		      if (cmd_type) {
			 if (cmd_mode==MODE_MENU && cmd_type==M_MENU_LIST) {
			    cmd_type = M_MENU_SEARCH;
	                    drawWinCmd(scene, 0);
			 } else {
                            cmd_type = 0;
	                    drawWinCmd(scene, 1);
			 }
		      }
		      else {
			 cmd_mode = MODE_MENU;
			 cmd_type = M_MENU_START;
	                 drawWinCmd(scene, 0);
		      }
		      break;
		   }
                   if (text_input_mode()) {
		      if (control_pressed) {
                         if (keysym==XK_a || keysym==XK_A)
			    caret_pos = 0;
			 else
                         if (keysym==XK_b || keysym==XK_B ||
                             keysym==XK_p || keysym==XK_P) {
                            if (caret_pos>0) --caret_pos;
			 }
			 else
                         if (keysym==XK_f || keysym==XK_F ||
                             keysym==XK_n || keysym==XK_N) {
                            if (caret_pos<strlen(cmd_string)) ++caret_pos;
			 }
		         else
			 if (keysym==XK_d || keysym==XK_D) {
		            int l, m;
                            l = strlen(cmd_string);
		            for (m=caret_pos; m<l; m++)
			       cmd_string[m] = cmd_string[m+1];
		         }
		         else
			 if (keysym==XK_h || keysym==XK_H) {
		            int l, m;
                            l = strlen(cmd_string);
		            if (caret_pos) {
		               for (m=caret_pos-1; m<l; m++)
			          cmd_string[m] = cmd_string[m+1];
			       --caret_pos;
		            }
			 }
			 else
			 if (keysym==XK_k || keysym==XK_K) {
			    cmd_killed_length = strlen(cmd_string);
			    cmd_killed_pos = caret_pos;
			    cmd_killed_char = cmd_string[caret_pos];
		            cmd_string[caret_pos] = '\0';
		         }
			 else
			 if (keysym==XK_y || keysym==XK_Y) {
			    if (cmd_killed_char) {
			       caret_pos = cmd_killed_pos;
			       cmd_string[cmd_killed_pos] = cmd_killed_char;
			       cmd_string[cmd_killed_length] = '\0';
			    }
		         }
			 else
                         if (keysym==XK_e || keysym==XK_E)
			    caret_pos = strlen(cmd_string);
		      } 
                      else
		      if (keysym==XK_Right) {
		         if (caret_pos<strlen(cmd_string)) ++caret_pos;
		      } 
                      else		      
		      if (keysym==XK_Left) {
		         if (caret_pos>0) --caret_pos;
		      } 
                      else		      		      
                      if (keysym==XK_Delete) {
		         int l, m;
                         l = strlen(cmd_string);
		         if (caret_pos) {
		            for (m=caret_pos-1; m<l; m++)
			       cmd_string[m] = cmd_string[m+1];
			    --caret_pos;
		         }
		      }
		      else
		      if (keysym>=32 && keysym<=255) {
		         int l, m;
		         if (cmd_string) 
                            l = strlen(cmd_string);
		         else
			    l = 0;
                         cmd_string = (char *)realloc(cmd_string, l+2);
                         cmd_killed_char = '\0';
		         for (m=l; m>caret_pos; m--)
			     cmd_string[m] = cmd_string[m-1];
                         cmd_string[caret_pos] = keysym;
		         cmd_string[l+1] = '\0';
                         ++caret_pos;
		      }
		      drawWinString(scene,cmd_string, BOTTOM);
		   }
		   break;
		}
                else
                if (keysym==XK_less) {
		   parse_locfile(scene);
                   generate(scene);
		}
                else
                if (keysym==XK_Delete) {
		   if (scene->num_positions>=2) {
		      XYZPosition pos;
		      i = scene->num_positions-1;
		      pos = scene->position[i];
		      for (j=i; j>0; j--)
                          scene->position[j] =scene->position[j-1];
		      scene->position[0] = pos;
		      scene->gopt.xrot = scene->position[i].xrot;
		      scene->gopt.yrot = scene->position[i].yrot;
		      scene->gopt.zrot = scene->position[i].zrot;
                      generate(scene);
                      if (scene->position[i].loc>=0) {
			 city_found = scene->position[i].loc;
			 show_datawin(scene, 1);
		      }
		      break;
		   }
		}
		else
                if (keysym==XK_q || keysym==XK_Q) 
                   exit(0);
		else
		if (keysym==XK_d || keysym==XK_D || keysym==XK_exclam) {
		   commands(scene);
                   break;
		} 
		else
                if (keysym==XK_k || keysym==XK_K) {
		   close_datawin(scene);
		   break;
		}
		else
		if (keysym==XK_f || keysym==XK_F) {
                   filemenu(scene);
	           drawWinCmd(scene, 0);
		   break;
		}
		else
		if (keysym==XK_o || keysym==XK_O) {
                   if (city_shown) putPicture(scene);
		   cmd_mode = MODE_OPTION;
		   cmd_type = M_OPTION_START;
		   show_title(scene, mainwin);
		   drawWinCmd(scene, 0);
		}
		else
		if (keysym==XK_c || keysym==XK_C) {
                   if (city_shown) putPicture(scene);
		   cmd_mode = MODE_COLOR;
		   cmd_type = M_COLOR_START;
		   show_title(scene, mainwin);
		   drawWinCmd(scene, 0);
		}
		else
		if (keysym==XK_s || keysym==XK_S) {
                   if (city_shown) putPicture(scene);
		   cmd_mode = MODE_MENU;
		   cmd_type = M_MENU_SEARCH;
		   free(cmd_string);
		   cmd_string = strdup(search_string);
		   caret_pos = strlen(search_string);
		   show_title(scene, mainwin);
		   drawWinCmd(scene, 0);
		}
                else
		if (keysym==XK_m || keysym==XK_M || keysym==XK_Tab) {
                   if (city_shown) putPicture(scene);
		   cmd_mode = MODE_MENU;
		   cmd_type = M_MENU_START;
		   show_title(scene, mainwin);
		   drawWinCmd(scene, 0);
		}
                else
                if (keysym==XK_i || keysym==XK_I) {
		   if (scene->color[C_CITY_CIRCLE].bool==ON &&
                       scene->color[C_CITY_TEXT].bool==ON) {
                       scene->color[C_CITY_CIRCLE].bool = OFF; 
                       scene->color[C_CITY_TEXT].bool = OFF; 
	               drawWinString(scene, 
                         msg[GENERATING_WITHOUT_CITIES],
                         BOTTOM);
		   } else {
		       if (scene->color[C_CITY_CIRCLE].bool==OFF) {
                          scene->color[C_CITY_CIRCLE].bool = ON; 
 	                  drawWinString(scene, 
			    msg[GENERATING_WITH_SPOTS_SHOWN],
                            BOTTOM);
		       } else {
                          scene->color[C_CITY_TEXT].bool = ON; 
 	                  drawWinString(scene, 
                            msg[GENERATING_WITH_SPOTS_AND_NAME_SHOWN],
                            BOTTOM);
		       }
		   }
		   generate(scene);
                }
                else
                if (keysym==XK_p || keysym==XK_P) {
		   if (scene->color[C_PEAK_TRIANGLE].bool==ON &&
                       scene->color[C_PEAK_TEXT].bool==ON) {
                       scene->color[C_PEAK_TRIANGLE].bool = OFF; 
                       scene->color[C_PEAK_TEXT].bool = OFF; 
	               drawWinString(scene, 
                         msg[GENERATING_WITHOUT_PEAKS],
                         BOTTOM);
		   } else {
		       if (scene->color[C_PEAK_TRIANGLE].bool==OFF) {
                          scene->color[C_PEAK_TRIANGLE].bool = ON; 
 	                  drawWinString(scene, 
			    msg[GENERATING_WITH_PEAKS_SHOWN],
                            BOTTOM);
		       } else {
                          scene->color[C_PEAK_TEXT].bool = ON; 
 	                  drawWinString(scene, 
                            msg[GENERATING_WITH_PEAKS_AND_NAME_SHOWN],
                            BOTTOM);
		       }
		   }
		   generate(scene);
                }
		else
                if (keysym==XK_l || keysym==XK_L) {
		   if (scene->color[C_LOC_LAND].bool==ON || 
                       scene->color[C_LOC_WATER].bool==ON) {
                      scene->color[C_LOC_LAND].bool = OFF;
                      scene->color[C_LOC_WATER].bool = OFF;
	              drawWinString(scene, 
                         msg[GENERATING_WITHOUT_LOCATIONS],
                         BOTTOM);
		   } else {
                      scene->color[C_LOC_LAND].bool = ON;
                      scene->color[C_LOC_WATER].bool = ON;
	              drawWinString(scene, 
                         msg[GENERATING_WITH_LOCATIONS_SHOWN],
                         BOTTOM);
		   }
		   generate(scene);
                }
		else
                if (keysym==XK_e || keysym==XK_E) {
		   if (scene->projection == SPHERICAL) {
		      if (scene->color[C_EARTH_CONTOUR].bool==ON) {
                         scene->color[C_EARTH_CONTOUR].bool = OFF;
	                 drawWinString(scene, 
                           msg[GENERATING_WITH_NO_EARTH_CONTOUR],
                           BOTTOM);
		      } else {
                         scene->color[C_EARTH_CONTOUR].bool = ON;
	                 drawWinString(scene, 
                           msg[GENERATING_WITH_EARTH_CONTOUR],
                           BOTTOM);
		      }
		      generate(scene);
		   } else
                      putPicture(scene);
		} 
		else
                if (keysym==XK_g || keysym==XK_G) {
		   if (scene->color[C_LAT_GRID].bool==ON ||
		       scene->color[C_LON_GRID].bool==ON) {
                      scene->color[C_MAIN_LINES].bool = OFF;
                      scene->color[C_LAT_GRID].bool = OFF; 
                      scene->color[C_LON_GRID].bool = OFF; 
	              drawWinString(scene, 
                         msg[GENERATING_WITHOUT_GRIDLINES],
                         BOTTOM);
		   } else {
                      scene->color[C_MAIN_LINES].bool = ON;
                      scene->color[C_LAT_GRID].bool = ON; 
                      scene->color[C_LON_GRID].bool = ON; 
 	              drawWinString(scene, 
                         msg[GENERATING_WITH_GRIDLINES],
                         BOTTOM);
		   }
		   generate(scene);
                }
		else
		if (keysym==XK_t || keysym==XK_T) {
                   scene->color[C_MAIN_LINES].bool
                      = 1 - scene->color[C_MAIN_LINES].bool;
                   if (scene->color[C_MAIN_LINES].bool)
 	              drawWinString(scene, 
                         msg[GENERATING_WITH_TROPICS], BOTTOM);
                   else
 	              drawWinString(scene, 
                         msg[GENERATING_WITHOUT_TROPICS], BOTTOM);
		   generate(scene);
		}
		else
		if (keysym==XK_r || keysym==XK_R) {
	           reverse = 1 - reverse;
		   get_all_colors(scene);
		   drawWinString(scene, msg[REVERSING_COLORS], BOTTOM);
		   generate(scene);
                }
		else
		if (keysym==XK_plus) {
	           drawWinString(scene, msg[ZOOMING_BY_SQRT2], BOTTOM);
                   scene->gopt.zoom *= sqrt(2);
		   generate(scene);
		}
                else
		if (keysym==XK_minus) {
	           drawWinString(scene, msg[REDUCING_BY_SQRT2], BOTTOM);
                   scene->gopt.zoom /= sqrt(2);
		   generate(scene);
		}
		else
		if (keysym==XK_Right) {
                   scene->gopt.yrot += angle_inc(scene->gopt.zoom);
	           sprintf(str, msg[MOVING_TO_LONGITUDE], scene->gopt.yrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
		if (keysym==XK_Left) {
                   scene->gopt.yrot -= angle_inc(scene->gopt.zoom);
	           sprintf(str, msg[MOVING_TO_LONGITUDE], scene->gopt.yrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
		if (keysym==XK_Up) {
                   scene->gopt.xrot += angle_inc(scene->gopt.zoom);
	           sprintf(str, msg[MOVING_TO_LATITUDE], scene->gopt.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
		if (keysym==XK_Down) {
                   scene->gopt.xrot -= angle_inc(scene->gopt.zoom);
	           sprintf(str, msg[MOVING_TO_LATITUDE], scene->gopt.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
                if (keysym==XK_Home) {
                   scene->gopt.yrot -= 15.0;
	           sprintf(str, msg[MOVING_TO_LONGITUDE], scene->gopt.yrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
                if (keysym==XK_End) {
                   scene->gopt.yrot += 15.0;
	           sprintf(str, msg[MOVING_TO_LONGITUDE], scene->gopt.yrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
		else
                if (keysym==XK_Page_Up) {
                   scene->gopt.xrot += 15.0;
	           sprintf(str, msg[MOVING_TO_LATITUDE], scene->gopt.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
                if (keysym==XK_Page_Down) {
                   scene->gopt.xrot -= 15.0;
	           sprintf(str, msg[MOVING_TO_LATITUDE], scene->gopt.xrot);
		   drawWinString(scene, str, BOTTOM);
		   generate(scene);
		}
                else
		if (keysym==XK_Return || keysym==XK_Escape || keysym==XK_space)
		   putPicture(scene);
		else
		if (keysym == XK_h || keysym == XK_H)
		      show_manual();
		else
		if (keysym == XK_v || keysym == XK_V) {
		   if (city_shown) putPicture(scene);
		   show_version(scene);
		}
		else
		if (keysym == XK_question) {
		   if (city_shown) putPicture(scene);
                   sprintf(str, msg[ZOOM_AND_POSITION],
		      scene->gopt.zoom, scene->aspect,
                      scene->gopt.xrot, scene->gopt.yrot, scene->gopt.zrot);
                   drawWinString(scene, str, TOP);
		   list_key_shortcuts(str);
                   drawWinString(scene, str, BOTTOM);
		}
                else
		if (keysym != XK_Shift_L && 
                    keysym != XK_Shift_R ) {
		   if (city_shown) putPicture(scene);
                   drawRectCmd(scene, msg+BEGIN_LIST_SHORTCUTS, 
                       END_LIST_SHORTCUTS-BEGIN_LIST_SHORTCUTS, TOP-1, -1);
                   show_version(scene);
                   city_shown = -1;
		   line_clicked = -1;
		}
	     }
	     break;

	     /* User probably has changed window size */
	     case ConfigureNotify : 
	        {
	           int x, y;
		   unsigned int w, h;
		   if (getWindowPlacement(mainwin, &x, &y, &w, &h) &&
                       (w!=scene->width || h!=scene->height) ) {
                      XResizeWindow(dpy, mainwin, w, h);
		      XFlush(dpy);
		      create_textpix(scene->width);
	              sprintf(str, msg[RESIZING_WINDOW], w, h);
		      scene->width = w;
		      scene->height = h;
		      drawWinString(scene, str, BOTTOM);
                      set_radius(scene);
		      discard_events(mainwin);
		      if (cmd_mode==MODE_RUN)
		         generate(scene);
		      else
			 redisplay(scene);
		      if (datawin_on) show_datawin(scene, 0);
		      break;
		   }
	        }

	     /* User has moved the mouse */
	     case MotionNotify :
		if (cmd_mode && ev.xbutton.y<=fonth[0]+5) break;
		if (cmd_mode && ev.xbutton.y>=scene->height-(fonth[0]+6)) break;
                if ((but_pressed==1 || but_pressed==3) && cmd_mode<=MODE_CMD)
		   show_point(scene, ev.xbutton.x, ev.xbutton.y);
	        break;

	     case ButtonPress :
		line_pos = -1;
		if (cmd_mode && ev.xbutton.y<=fonth[0]+5) break;
		if (cmd_mode==MODE_MENU) break;
		if (cmd_mode && ev.xbutton.y>=scene->height-fonth[0]-6) {
		   if (text_input_mode())
                      do_caret(scene, ev.xbutton.x);
		   break;
		}
		if (cmd_mode==MODE_RUN && city_shown) {
		   city_shown = 0;
		   putPicture(scene);  
		}
		if (cmd_mode==MODE_RUN && ev.xbutton.button==3) {
	           show_point(scene, ev.xbutton.x, ev.xbutton.y);
		   break;
		}
		if (cmd_mode<=MODE_CMD && ev.xbutton.button!=2)
	           show_point(scene, ev.xbutton.x, ev.xbutton.y);
		break;

             case ButtonRelease :
                /* We are in MODE_RUN */
		if (cmd_mode==MODE_RUN) {
		   /* Button 1 released, erase any text on strips */
		   if (ev.xbutton.button==1) {
		      if (!city_shown)
                         putPicture(scene);
		      else
			 XMapRaised(dpy, datawin);
		      break;
		   }
		   /* Button 2 released while MODE_RUN, --> MODE_MENU */
		   if (ev.xbutton.button == 2) {
		      cmd_mode = MODE_MENU;
		      cmd_type = M_MENU_START;
		      show_title(scene, mainwin);
		      drawWinCmd(scene, 0);
		      break;
		   }
		   /* Button 3 released, recenter the map */
                   if (ev.xbutton.button==3) {
		      double x, y;
                      if (inverse_coord(scene, 
                         ev.xbutton.x, ev.xbutton.y, &x, &y)) {
		         scene->gopt.xrot = x;
		         scene->gopt.yrot = y;
		         drawWinString(scene,msg[RECENTERING_TO_POINT],BOTTOM);
		         generate(scene);
		      } else
		         putPicture(scene);
		   }
		}

		/* Click on upper command strip, while != MODE_RUN */
		if (cmd_mode && ev.xbutton.y<=fonth[0]+5) {
		   if (ev.xbutton.x>=scene->width-action_width)
		      exit(0);
		   if (ev.xbutton.x>=scene->width-2*action_width) {
		      do_escape(scene);
		      break;
		   }
		   if (ev.xbutton.x>=scene->width-3*action_width) {
                      do_activate(scene, XK_exclam);
		      break;
		   }

		   cmd_mode = MODE_MENU;
		   for (i=M_MENU_HELP; i>=1; i--) {
		      if (ev.xbutton.x>=
                             XTextWidth(xfont[0], labels_generalmenu, 
                                        spacing_generalmenu*(i-1)-1))
			 break;
		   }
	           if (i==M_MENU_HELP) {
		      cmd_mode = MODE_HELP;
		      cmd_type = M_HELP_START;
		   }
		   if (i==M_MENU_FILE) {
		      cmd_mode = MODE_FILE;
		      cmd_type = M_FILE_START;
                   }
		   if (i==M_MENU_CMD) {
		      cmd_mode = MODE_CMD;
		      cmd_type = M_CMD_START;
                   }
                   if (i==M_MENU_SEARCH) {
		      cmd_type = M_MENU_SEARCH;
		      free(cmd_string);
		      cmd_string = strdup(search_string);
		      caret_pos = strlen(search_string);
                      show_title(scene, mainwin);
	           }	    
		   if (i==M_MENU_OPTION) {
		      cmd_mode = MODE_OPTION;
		      cmd_type = M_OPTION_START;
		   }
		   if (i==M_MENU_COLOR) {
		      cmd_mode = MODE_COLOR;
                      cmd_type = M_COLOR_START;
		   }
		   if (i==M_MENU_HELP) {
		      cmd_mode = MODE_HELP;
		      cmd_type = M_HELP_START;
                   }
                   drawWinCmd(scene, 0);
		   break;
		}

		if (cmd_mode && 
                    ev.xbutton.y>=fonth[0]+16 &&
                    ev.xbutton.y<=2*fonth[0]+21) {

                   line_clicked = -1;
		   *cmd_string = '\0';
		   caret_pos = 0;

		   if (cmd_mode==MODE_CMD) {
		      if (point_shown) {
		         drawWinCmd(scene, 0);
		         point_shown = 0;
		      }
		      break;
		   }

		   if (cmd_mode==MODE_FILE) {
		      cmd_type = M_FILE_SAVE;
		      for (i=M_FILE_SAVE; i>=1; i--) {
		         if (ev.xbutton.x<
                             XTextWidth(xfont[0], labels_filemenu, 
                                                  spacing_filemenu*i-1))
			 cmd_type = i;
		      }
		      if (cmd_type == M_FILE_SAVE)
			 overwrite = 0;
		   }

                   if (cmd_mode==MODE_OPTION) {
		      cmd_type = M_OPTION_CAT;
		      for (i=M_OPTION_CAT; i>=1; i--) {
		         if (ev.xbutton.x<XTextWidth(xfont[0],
                             labels_optionmenu, spacing_optionmenu*i-1))
			 cmd_type = i;
		      }
		   }

                   if (cmd_mode==MODE_COLOR) {
		      int i;
		      cmd_type = M_COLOR_CIL;
		      for (i=M_COLOR_CIL; i>=1; i--) {
		      if (ev.xbutton.x<XTextWidth(xfont[0],
                         labels_colormenu, spacing_colormenu*i-1))
			 cmd_type = i;
		      }
		   }
                   if (cmd_mode==MODE_HELP) {
		      cmd_type = M_HELP_MANUAL;
		      for (i=M_HELP_MANUAL; i>=1; i--) {
		      if (ev.xbutton.x<XTextWidth(xfont[0],
                         labels_helpmenu, spacing_helpmenu*i-1))
			 cmd_type = i;
		      }
		   }
		   drawWinCmd(scene, 1);
		   break;
		}

                /* Click released on the bottom strip */
		if (cmd_mode && ev.xbutton.y>=scene->height-fonth[0]-6) {
		   if (point_shown)
		      drawWinCmd(scene, 0);
		   else
   		   if (text_input_mode())
		      do_caret(scene, ev.xbutton.x);
		   break;
		}
		/* We are in MODE_CMD, just refresh window */
		if (cmd_mode==MODE_CMD) {
                   if (point_shown)
                      drawWinCmd(scene, 0);
		   break;
		}
                /* Click released somewhere not on the top/bottom strips */
                if (click_input_mode()) {
                   l = -1;
		   if (ev.xbutton.x>box_x && ev.xbutton.x<box_x+box_width) {
		      j = ((int)ev.xbutton.y - box_y);
		      m = 0;
		      while (j<0) { j += box_spacing; m += 1; }
		      j = j/box_spacing - m;
		      if (ev.xbutton.y>box_y+j*box_spacing &&
			  ev.xbutton.y<box_y+j*box_spacing+box_height)
			 l = j;
		   }
		do_click:
		   if (l>=1 && l<=box_numitems) {
		      if (line_clicked == l || 
			  (cmd_mode==MODE_MENU && cmd_type==M_MENU_LIST)) {
			 int j=0;
                         if (cmd_mode==MODE_FILE) {
			    if (cmd_type==M_FILE_PRINTCFG) {
			       if (l==4) {
				  if (ps_rot==0) ps_rot = 90;
				  else
				  if (ps_rot==90) ps_rot = -90;
				  else
				  if (ps_rot==-90) ps_rot = 0;
				  cmd_string = (char *) realloc(cmd_string, 4);
	                          cmd_killed_char = '\0';
				  sprintf(cmd_string, "%d", ps_rot);
				  caret_pos = strlen(cmd_string);
			       }
			       if (l==5) {
			          ps_grayscale = 1-ps_grayscale;
			          cmd_string = (char *) realloc(cmd_string, 2);
			          cmd_string[0] = (ps_grayscale)? '+':'-';
			          cmd_string[1] = '\0';
	                          cmd_killed_char = '\0';
			       }
			       if (l==6) {
			          im_compress = 1-im_compress;
			          cmd_string = (char *) realloc(cmd_string, 2);
			          cmd_string[0] = (im_compress)? '+':'-';
			          cmd_string[1] = '\0';
	                          cmd_killed_char = '\0';
			       }
			    }
			    if (cmd_type==M_FILE_SAVE) {
			       do_activate(scene, XK_Return);
			       break;
			    }
			 }
                         if (cmd_mode==MODE_MENU) {
			    if (cmd_type==M_MENU_SEARCH) {
			       search_by = l;
			       do_activate(scene, XK_Return);
			       break;
			    }
			    if (cmd_type==M_MENU_LIST) {
			       int level;
			       if (l==1 && start_list>0) {
				  start_list -= (num_listed-3);
				  if (start_list<0) start_list = 0;
				  level = 1;
			       } else
			       if (l==num_listed && 
                                   start_list+num_listed<num_found) {
				  start_list += (num_listed-3);
				  if (start_list+num_listed>num_found)
				     start_list = num_found-num_listed;
				  level = 1;
			       } else {
				  if (line_clicked==-10)
				     level = 1;
				  else
				     level = 2;
                                  line_clicked = l;
			          city_found = loc_pointer[start_list+l-1];
			          search_index = city_found;
	                          scene->gopt.xrot = locations[city_found]->lat;
	                          scene->gopt.yrot = locations[city_found]->lon;
				  /*
				  free(search_string);
				  search_string = 
                                     strdup(locations[city_found]->name);
				  */
			          show_datawin(scene, 1);
			       }
			       drawWinCmd(scene, level);
			       break;
			    }
			 }
			 if (cmd_mode==MODE_OPTION) {
			    if (cmd_type==M_OPTION_PARAM) {
			       if (l==1) {
                                  scene->projection = 
                                     (scene->projection+1)%NUMPROJ;
				  cmd_string = (char *) realloc(cmd_string, 3);
	                          cmd_killed_char = '\0';
				  sprintf(cmd_string, "%d", scene->projection);
			       }
			    }
			    if (cmd_type==M_OPTION_DISPLAY) {
                               if (l==NUM_DISPLAY-2) {
				  scene->ropt.placetext = 1-scene->ropt.placetext;
				  cmd_string = (char *) realloc(cmd_string, 2);
			          cmd_string[0] = (scene->ropt.placetext)? '+':'-';
				  cmd_string[1] = '\0';
	                          cmd_killed_char = '\0';
			       }
                               if (l==NUM_DISPLAY-1) {
				  scene->gopt.transparent = 1-scene->gopt.transparent;
				  cmd_string = (char *) realloc(cmd_string, 2);
				  cmd_string[0] = 
                                     (scene->gopt.transparent)? '+' : '-';
				  cmd_string[1] = '\0';
			       }
                               if (l==NUM_DISPLAY) {
                                  char *ptr = strstr (list_lang, language)+3;
                                  if (!ptr || ptr>=list_lang+strlen(list_lang))
				     ptr = list_lang;
				  cmd_string = (char *) realloc(cmd_string, 4);
				  strncpy(language, ptr, 2);
				  language[2] = '\0';
				  strcpy(cmd_string, language);
                         	  read_i18n_file(scene);
	                          fix_features_description(scene);
				  drawWinCmd(scene, 0);
				  break;
			       }
			    }
			    if (cmd_type==M_OPTION_MARKS) {
	                       switch(l) {
	      case 1: if (scene->color[C_CITY_CIRCLE].bool == OFF) {
			 scene->color[C_CITY_CIRCLE].bool = ON;
			 scene->color[C_CITY_TEXT].bool = OFF;
		      } else
                      if (scene->color[C_CITY_CIRCLE].bool == ON &&
			  scene->color[C_CITY_TEXT].bool == OFF) 
			 scene->color[C_CITY_TEXT].bool = ON;
		      else
                      if (scene->color[C_CITY_CIRCLE].bool == ON &&
			  scene->color[C_CITY_TEXT].bool == ON) {
			 scene->color[C_CITY_CIRCLE].bool = OFF;
			 scene->color[C_CITY_TEXT].bool = OFF;
		      }
                      break;
			       
	      case 2: if (scene->color[C_PEAK_TRIANGLE].bool == OFF) {
			 scene->color[C_PEAK_TRIANGLE].bool = ON;
			 scene->color[C_PEAK_TEXT].bool = OFF;
		      } else
                      if (scene->color[C_PEAK_TRIANGLE].bool == ON &&
			  scene->color[C_PEAK_TEXT].bool == OFF) 
			 scene->color[C_PEAK_TEXT].bool = ON;
		      else
                      if (scene->color[C_PEAK_TRIANGLE].bool == ON &&
			  scene->color[C_PEAK_TEXT].bool == ON) {
			 scene->color[C_PEAK_TRIANGLE].bool = OFF;
			 scene->color[C_PEAK_TEXT].bool = OFF;
		      }
                      break;
			       
              case 3: scene->color[C_LOC_LAND].bool = 
                          !scene->color[C_LOC_LAND].bool;
                      break;

              case 4: scene->color[C_LOC_WATER].bool = 
                          !scene->color[C_LOC_WATER].bool;
                      break;

              case 5: scene->color[C_LON_GRID].bool = 
                          !scene->color[C_LON_GRID].bool;
                      break;
			       
              case 6: scene->color[C_LAT_GRID].bool =
                          !scene->color[C_LAT_GRID].bool;
                      break;
			    
              case 7: scene->color[C_MAIN_LINES].bool = 
			  !scene->color[C_MAIN_LINES].bool;
                      break;

              case 8: scene->color[C_EARTH_CONTOUR].bool =
                          !scene->color[C_EARTH_CONTOUR].bool;
                      break;

              case 9: scene->color[C_BG_SKY].bool =
                          !scene->color[C_BG_SKY].bool;
                      break;

              case 10: scene->color[C_FG_STARS].bool =
                          !scene->color[C_FG_STARS].bool;
                       break;
			    
	      default: break;
			       }
			       cmd_string = realloc(cmd_string, 3);
			       value2str(scene, l, cmd_string, 0);
			    }
			    if (cmd_type>=M_OPTION_CONTIN) {
                               j = 1<<(l-1);
			    }
			    if (cmd_type==M_OPTION_CONTIN) {
			       if (scene->gopt.desired_continents & j) {
				  scene->gopt.desired_continents &= ~j;
				  *cmd_string = '-';
			       } else {
				  scene->gopt.desired_continents |= j;
			          *cmd_string = '+';
			       } 
			    }
			    if (cmd_type==M_OPTION_CAT) {
			       if (scene->gopt.desired_categories & j) {
				  scene->gopt.desired_categories &= ~j;
				  *cmd_string = '-';
			       } else {
			          scene->gopt.desired_categories |= j;
				  *cmd_string = '+';
			       } 
			    }
			 }
		         if (cmd_mode==MODE_COLOR) {
			    if (cmd_type>=M_COLOR_LABELS+(l==1)) {
			       j = feature_index[cmd_type-1]+l-1;
			       scene->color[j].bool = 1-scene->color[j].bool;
			       *cmd_string = (scene->color[j].bool)? '+' : '-';
			    }
			 }
		      } else {
			if (cmd_mode==MODE_FILE && cmd_type==M_FILE_SAVE &&
                            line_clicked==1 && *cmd_string)
			   strcpy(scene->outfile, cmd_string);
		        line_clicked = l;
	                if (cmd_mode==MODE_MENU && cmd_type==M_MENU_SEARCH) {
 	                   search_by = l;
		           drawWinCmd(scene, 2);			   
			   break;
	                }
			value2str(scene, line_clicked, str, 0);
		        l = strlen(str);
		        cmd_string = (char *)realloc(cmd_string, l+2);
		        strcpy(cmd_string, str);
	                cmd_killed_char = '\0';
			if (cmd_mode==MODE_OPTION && 
                            cmd_type==M_OPTION_DISPLAY) {
			   if (line_clicked==1)
			      cmd_string[5] = '\0';
			   if (line_clicked==NUM_DISPLAY)
			      cmd_string[2] = '\0';
			   l = strlen(cmd_string);
			}
		        caret_pos = l;
		      }
		   }
		   drawWinCmd(scene, 2);
		   break;
		}
		if (cmd_mode==MODE_MENU) break;
		if (cmd_mode && ev.xbutton.y>=scene->height-fonth[0]-6) {
                   do_caret(scene, ev.xbutton.x);
	           break;
		}
	        break;

	     /* Expose events */
	     case Expose :
		exposed |= MAIN_WINDOW;
                break;
	     }
	}
	return 0;
}	
